/*
 * A V4L2 driver for samsung s5k4ec cameras.
 *
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/videodev2.h>
#include <linux/clk.h>
#include <media/v4l2-device.h>
#include <media/v4l2-chip-ident.h>
#include <media/v4l2-mediabus.h>//linux-3.0
#include <linux/io.h>
#include <linux/gpio.h>
#include <mach/sys_config.h>
#include <linux/regulator/consumer.h>
#include <mach/system.h>
#include "../include/sunxi_csi_core.h"
#include "../include/sunxi_dev_csi.h"

MODULE_AUTHOR("raymonxiu");
MODULE_DESCRIPTION("A low-level driver for samsung s5k4ec sensors");
MODULE_LICENSE("GPL");

//for internel driver debug
#define DEV_DBG_EN   		0 
#if(DEV_DBG_EN == 1)		
#define csi_dev_dbg(x,arg...) printk("[CSI_DEBUG][s5k4ec]"x,##arg)
#else
#define csi_dev_dbg(x,arg...) 
#endif
#define csi_dev_err(x,arg...) printk("[CSI_ERR][s5k4ec]"x,##arg)
#define csi_dev_print(x,arg...) printk("[CSI][s5k4ec]"x,##arg)

#define MCLK (24*1000*1000)
#define VREF_POL	CSI_HIGH
#define HREF_POL	CSI_HIGH
#define CLK_POL		CSI_RISING
#define IO_CFG		0						//0 for csi0
#define V4L2_IDENT_SENSOR 0x4ec

//define the voltage level of control signal
#define CSI_STBY_ON			0
#define CSI_STBY_OFF 		1
#define CSI_RST_ON			0
#define CSI_RST_OFF			1
#define CSI_PWR_ON			1
#define CSI_PWR_OFF			0
#define CSI_AF_PWR_ON		0
#define CSI_AF_PWR_OFF	1

#define REG_TERM 0xff
#define VAL_TERM 0xff


#define REG_ADDR_STEP 2
#define REG_DATA_STEP 2
#define REG_STEP 			(REG_ADDR_STEP+REG_DATA_STEP)

#define CONTINUEOUS_AF

/*
 * Basic window sizes.  These probably belong somewhere more globally
 * useful.
 */
#define QSXGA_WIDTH		2560
#define QSXGA_HEIGHT	1920
#define QXGA_WIDTH 		2048
#define QXGA_HEIGHT		1536
#define HD1080_WIDTH	1920
#define HD1080_HEIGHT	1080
#define UXGA_WIDTH		1600
#define UXGA_HEIGHT		1200
#define SXGA_WIDTH		1280
#define SXGA_HEIGHT		960
#define HD720_WIDTH 	1280
#define HD720_HEIGHT	720
#define XGA_WIDTH			1024
#define XGA_HEIGHT 		768
#define SVGA_WIDTH		800
#define SVGA_HEIGHT 	600
#define VGA_WIDTH			640
#define VGA_HEIGHT		480
#define QVGA_WIDTH		320
#define QVGA_HEIGHT		240
#define CIF_WIDTH			352
#define CIF_HEIGHT		288
#define QCIF_WIDTH		176
#define	QCIF_HEIGHT		144



/*
 * Our nominal (default) frame rate.
 */
#define SENSOR_FRAME_RATE 30

/*
 * The s5k4ec sits on i2c with ID 0x5a
 */
#define I2C_ADDR 0x5a

/* Registers */


static int sensor_s_band_filter(struct v4l2_subdev *sd, 
		enum v4l2_power_line_frequency value);

		
/*
 * Information we maintain about a known sensor.
 */
struct sensor_format_struct;  /* coming later */
struct snesor_colorfx_struct; /* coming later */
__csi_subdev_info_t ccm_info_con = 
{
	.mclk 	= MCLK,
	.vref 	= VREF_POL,
	.href 	= HREF_POL,
	.clock	= CLK_POL,
	.iocfg	= IO_CFG,
};

struct sensor_info {
	struct v4l2_subdev sd;
	struct sensor_format_struct *fmt;  /* Current format */
	__csi_subdev_info_t *ccm_info;
	int	width;
	int	height;
	unsigned int capture_mode;		//0:video 1:capture
	unsigned int coarse_af_pd;		//0:busy 1:ok 2:failed
	unsigned int focus_status;		//0:idle 1:busy
	int brightness;
	int	contrast;
	int saturation;
	int hue;
	int hflip;
	int vflip;
	int gain;
	int autogain;
	int exp;
	enum v4l2_exposure_auto_type autoexp;
	int autowb;
	enum v4l2_whiteblance wb;
	enum v4l2_colorfx clrfx;
	enum v4l2_flash_mode flash_mode;
	enum v4l2_power_line_frequency band_filter;
	enum v4l2_autofocus_mode af_mode;
	enum v4l2_autofocus_ctrl af_ctrl;
	struct v4l2_fract tpf;			
};

static inline struct sensor_info *to_state(struct v4l2_subdev *sd)
{
	return container_of(sd, struct sensor_info, sd);
}



/*
 * The default register settings
 *
 */

struct regval_list {
	unsigned char reg_num[REG_ADDR_STEP];
	unsigned char value[REG_DATA_STEP];
};


static struct regval_list sensor_default_regs[] = {
//==================================================================================
// 00.History
//==================================================================================
//2010 : EVT1.1
//20110429 : LSI CSE Standard
//20110728 : Sequence Changed
//20110728 : ESD Check Register Address Changed
//20110829 : TnP Changed by S.Y.Lee
//20120104 : init Parm Update sequence changed
//20120201 : Flash ֺ Green Noise  setting 
//20120228 : Add Brightness Block
//==================================================================================
  
//==================================================================================
// 01.Start Setting
//==================================================================================

{{0xFC,0xFC},{0xD0,0x00}},	
{{0x00,0x10},{0x00,0x01}},	//S/W Reset
{{0x10,0x30},{0x00,0x00}},	//contint_host_int
{{0x00,0x14},{0x00,0x01}},	//sw_load_complete - Release CORE (Arm) from reset state

{{0xff,0xff},{0x00,10}}, //Delay 10ms

//==================================================================================                
// 02.ETC Setting                                                                                    
//==================================================================================                
                                                                                                    
{{0x00,0x28},{0xD0,0x00}},	//Driving Current                                                                   
{{0x00,0x2A},{0x10,0x82}},	
{{0x0f,0x12},{0x00,0x00}}, 	//cregs_d0_d4_cd10 //D4[9:8], D3[7:6], D2[5:4], D1[3:2], D0[1:0]                    
{{0x00,0x2A},{0x10,0x84}},                                                                                           
{{0x0f,0x12},{0x00,0x00}}, 	//cregs_d5_d9_cd10 //D9[9:8], D8[7:6], D7[5:4], D6[3:2], D5[1:0]                    
{{0x00,0x2A},{0x10,0x88}},                                                                                           
{{0x0f,0x12},{0x00,0x00}},	//cregs_clks_output_cd10 //SDA[11:10], SCL[9:8], PCLK[7:6], VSYNC[3:2], HSYNC[1:0]  
	                                                                                            
//==================================================================================
// 03.Analog Setting1 & ASP Control
//==================================================================================
//This register is for FACTORY ONLY. 
//If you change it without prior notification
//YOU are RESPONSIBLE for the FAILURE that will happen in the future

{{0x00,0x28},{0xD0,0x00}},
{{0x00,0x2A},{0x00,0x7A}},
{{0x0f,0x12},{0x00,0x00}},	
{{0x00,0x2A},{0xE4,0x06}},	//[7]f_ladlc_en [6:5]f max [4]fadlc_en [3:2]L max [1]ladlc_en [0]adlc_ch_sel
{{0x0F,0x12},{0x00,0x92}},	
{{0x00,0x2A},{0xE4,0x10}},	
{{0x0F,0x12},{0x38,0x04}},	//[15:8]fadlc_filter_co_b, [7:0]fadlc_filter_co_a	
{{0x00,0x2A},{0xE4,0x1A}},	
{{0x0F,0x12},{0x00,0x10}},	 	
{{0x00,0x2A},{0xE4,0x20}},	
{{0x0F,0x12},{0x00,0x03}},	//adlc_fadlc_filter_refresh 	
{{0x0F,0x12},{0x00,0x60}},	//adlc_filter_level_diff_threshold
{{0x00,0x2A},{0xE4,0x2E}},	
{{0x0F,0x12},{0x00,0x04}},	//dithered l-ADLC(4bit)
{{0x00,0x2A},{0xF4,0x00}},	
{{0x0F,0x12},{0x5A,0x3C}},	//[15:8]stx_width, [7:0]dstx_width 
{{0x0F,0x12},{0x00,0x23}},	//[14]binning_test [13]gain_mode [11:12]row_id [10]cfpn_test [9]pd_pix [8]teg_en, [7]adc_res, [6]smp_en, [5]ldb_en, [4]ld_en, [3]clp_en [2]srx_en, [1]dshut_en, [0]dcds_en 
{{0x0F,0x12},{0x80,0x80}},	//CDS option 
{{0x0F,0x12},{0x03,0xAF}},	//[11:6]rst_mx, [5:0]sig_mx
{{0x0F,0x12},{0x00,0x0A}},	//Avg mode
{{0x0F,0x12},{0xAA,0x54}},	//x1~x1.49:No MS, x1.5~x3.99:MS2, x4~x16:MS4
{{0x0F,0x12},{0x00,0x40}},	//RMP option [6]1: RES gain
{{0x0F,0x12},{0x46,0x4E}},	//[14]msoff_en, [13:8]off_rst, [7:0]adc_sat
{{0x0F,0x12},{0x02,0x40}},	//bist_sig_width_e
{{0x0F,0x12},{0x02,0x40}},	//bist_sig_width_o
{{0x0F,0x12},{0x00,0x40}},	//[9]dbs_bist_en, [8:0]bist_rst_width
{{0x0F,0x12},{0x10,0x00}},	//[15]aac_en, [14]GCLK_DIV2_EN, [13:10]dl_cont [9:8]dbs_mode, [7:0]dbs_option
{{0x0F,0x12},{0x55,0xFF}},	//bias [15:12]pix, [11:8]pix_bst [7:4]comp2, [3:0]comp1  	
{{0x0F,0x12},{0xD0,0x00}},	//[15:8]clp_lvl, [7:0]ref_option, [5]pix_bst_en
{{0x0F,0x12},{0x00,0x10}},	//[7:0]monit
{{0x0F,0x12},{0x02,0x02}},	//[15:8]dbr_tune_tgsl, [7:0]dbr_tune_pix
{{0x0F,0x12},{0x04,0x01}},	//[15:8]dbr_tune_ntg, [7:0]dbr_tune_rg                                                                       
{{0x0F,0x12},{0x00,0x22}},	//[15:8]reg_option, [7:4]rosc_tune_ncp, [3:0]rosc_tune_cp                                                    
{{0x0F,0x12},{0x00,0x88}},	//PD [8]inrush_ctrl, [7]fblv, [6]reg_ntg, [5]reg_tgsl, [4]reg_rg, [3]reg_pix, [2]ncp_rosc, [1]cp_rosc, [0]cp 
{{0x0F,0x12},{0x00,0x9F}},	//[9]capa_ctrl_en, [8:7]fb_lv, [6:5]dbr_clk_sel, [4:0]cp_capa                                                
{{0x0F,0x12},{0x00,0x00}},	//[15:0]blst_en_cintr                                                                                        
{{0x0F,0x12},{0x18,0x00}},	//[11]blst_en, [10]rfpn_test, [9]sl_off, [8]tx_off, [7:0]rdv_option                                          
{{0x0F,0x12},{0x00,0x88}},	//[15:0]pmg_reg_tune                                                                                         
{{0x0F,0x12},{0x00,0x00}},	//[15:1]analog_dummy, [0]pd_reg_test 
{{0x0F,0x12},{0x24,0x28}},	//[13:11]srx_gap1, [10:8]srx_gap0, [7:0]stx_gap
{{0x0F,0x12},{0x00,0x00}},	//[0]atx_option       
{{0x0F,0x12},{0x03,0xEE}},	//aig_avg_half
{{0x0F,0x12},{0x00,0x00}},	//[0]hvs_test_reg     
{{0x0F,0x12},{0x00,0x00}},	//[0]dbus_bist_auto   
{{0x0F,0x12},{0x00,0x00}},	//[7:0]dbr_option	    
{{0x00,0x2A},{0xF5,0x52}},	
{{0x0F,0x12},{0x07,0x08}},	//[7:0]lat_st, [15:8]lat_width  
{{0x0F,0x12},{0x08,0x0C}},	//[7:0]hold_st, [15:8]hold_width

//=================================================================================
// 04.Trap and Patch
//=================================================================================
// Start of Patch data

{{0x00,0x28},{0x70,0x00}},                             
{{0x00,0x2A},{0x3A,0xF8}},                             
{{0x0F,0x12},{0xB5,0xF8}},    // 70003AF8
{{0x0F,0x12},{0x4B,0x44}},    // 70003AFA
{{0x0F,0x12},{0x49,0x44}},    // 70003AFC
{{0x0F,0x12},{0x48,0x45}},    // 70003AFE
{{0x0F,0x12},{0x22,0x00}},	// 70003B00              
{{0x0F,0x12},{0xC0,0x08}},	// 70003B02              
{{0x0F,0x12},{0x60,0x01}},	// 70003B04              
{{0x0F,0x12},{0x49,0x44}},    // 70003B06
{{0x0F,0x12},{0x48,0x44}},    // 70003B08
{{0x0F,0x12},{0x24,0x01}},	// 70003B0A              
{{0x0F,0x12},{0xF0,0x00}},	// 70003B0C              
{{0x0F,0x12},{0xFC,0xA4}},    // 70003B0E
{{0x0F,0x12},{0x49,0x43}},    // 70003B10
{{0x0F,0x12},{0x48,0x44}},    // 70003B12
{{0x0F,0x12},{0x27,0x02}},    // 70003B14
{{0x0F,0x12},{0x00,0x22}},	// 70003B16              
{{0x0F,0x12},{0xF0,0x00}},	// 70003B18              
{{0x0F,0x12},{0xFC,0x9E}},    // 70003B1A
{{0x0F,0x12},{0x02,0x60}},    // 70003B1C
{{0x0F,0x12},{0x4C,0x42}},    // 70003B1E
{{0x0F,0x12},{0x80,0x20}},    // 70003B20
{{0x0F,0x12},{0x26,0x00}},    // 70003B22
{{0x0F,0x12},{0x80,0x66}},    // 70003B24
{{0x0F,0x12},{0x49,0x41}},    // 70003B26
{{0x0F,0x12},{0x48,0x41}},    // 70003B28
{{0x0F,0x12},{0x60,0x41}},	// 70003B2A              
{{0x0F,0x12},{0x49,0x41}},    // 70003B2C
{{0x0F,0x12},{0x48,0x42}},    // 70003B2E
{{0x0F,0x12},{0x00,0x3A}},    // 70003B30
{{0x0F,0x12},{0x25,0x03}},    // 70003B32
{{0x0F,0x12},{0xF0,0x00}},	// 70003B34              
{{0x0F,0x12},{0xFC,0x90}},    // 70003B36
{{0x0F,0x12},{0x48,0x3D}},    // 70003B38
{{0x0F,0x12},{0x49,0x40}},    // 70003B3A
{{0x0F,0x12},{0x30,0xC0}},	// 70003B3C              
{{0x0F,0x12},{0x63,0xC1}},	// 70003B3E              
{{0x0F,0x12},{0x4F,0x3B}},    // 70003B40
{{0x0F,0x12},{0x48,0x3F}},    // 70003B42
{{0x0F,0x12},{0x3F,0x80}},    // 70003B44
{{0x0F,0x12},{0x64,0x38}},    // 70003B46
{{0x0F,0x12},{0x48,0x3E}},    // 70003B48
{{0x0F,0x12},{0x49,0x3F}},    // 70003B4A
{{0x0F,0x12},{0x63,0x88}},	// 70003B4C              
{{0x0F,0x12},{0x00,0x2A}},    // 70003B4E
{{0x0F,0x12},{0x49,0x3E}},    // 70003B50
{{0x0F,0x12},{0x48,0x3F}},    // 70003B52
{{0x0F,0x12},{0x25,0x04}},	// 70003B54              
{{0x0F,0x12},{0xF0,0x00}},	// 70003B56              
{{0x0F,0x12},{0xFC,0x7F}},    // 70003B58
{{0x0F,0x12},{0x00,0x2A}},    // 70003B5A
{{0x0F,0x12},{0x49,0x3D}},    // 70003B5C
{{0x0F,0x12},{0x48,0x3E}},    // 70003B5E
{{0x0F,0x12},{0x25,0x05}},    // 70003B60
{{0x0F,0x12},{0xF0,0x00}},	// 70003B62              
{{0x0F,0x12},{0xF8,0xA7}},    // 70003B64
{{0x0F,0x12},{0x48,0x3C}},    // 70003B66
{{0x0F,0x12},{0x00,0x2A}},    // 70003B68
{{0x0F,0x12},{0x49,0x3C}},    // 70003B6A
{{0x0F,0x12},{0x25,0x06}},	// 70003B6C              
{{0x0F,0x12},{0x1D,0x80}},    // 70003B6E
{{0x0F,0x12},{0xF0,0x00}},    // 70003B70
{{0x0F,0x12},{0xF8,0xA0}},    // 70003B72
{{0x0F,0x12},{0x48,0x38}},    // 70003B74
{{0x0F,0x12},{0x00,0x2A}},    // 70003B76
{{0x0F,0x12},{0x49,0x39}},    // 70003B78
{{0x0F,0x12},{0x25,0x07}},    // 70003B7A
{{0x0F,0x12},{0x30,0x0C}},    // 70003B7C
{{0x0F,0x12},{0xF0,0x00}},    // 70003B7E
{{0x0F,0x12},{0xF8,0x99}},    // 70003B80
{{0x0F,0x12},{0x48,0x35}},    // 70003B82
{{0x0F,0x12},{0x00,0x2A}},    // 70003B84
{{0x0F,0x12},{0x49,0x37}},    // 70003B86
{{0x0F,0x12},{0x25,0x08}},    // 70003B88
{{0x0F,0x12},{0x30,0x10}},    // 70003B8A
{{0x0F,0x12},{0xF0,0x00}},    // 70003B8C
{{0x0F,0x12},{0xF8,0x92}},    // 70003B8E
{{0x0F,0x12},{0x00,0x2A}},    // 70003B90
{{0x0F,0x12},{0x49,0x35}},    // 70003B92
{{0x0F,0x12},{0x48,0x35}},    // 70003B94
{{0x0F,0x12},{0x25,0x09}},    // 70003B96
{{0x0F,0x12},{0xF0,0x00}},    // 70003B98
{{0x0F,0x12},{0xFC,0x5E}},    // 70003B9A
{{0x0F,0x12},{0x00,0x2A}},    // 70003B9C
{{0x0F,0x12},{0x49,0x34}},    // 70003B9E
{{0x0F,0x12},{0x48,0x34}},    // 70003BA0
{{0x0F,0x12},{0x25,0x0A}},    // 70003BA2
{{0x0F,0x12},{0xF0,0x00}},    // 70003BA4
{{0x0F,0x12},{0xFC,0x58}},    // 70003BA6
{{0x0F,0x12},{0x00,0x2A}},    // 70003BA8
{{0x0F,0x12},{0x49,0x33}},    // 70003BAA
{{0x0F,0x12},{0x48,0x33}},    // 70003BAC
{{0x0F,0x12},{0x25,0x0B}},    // 70003BAE
{{0x0F,0x12},{0xF0,0x00}},    // 70003BB0
{{0x0F,0x12},{0xFC,0x52}},    // 70003BB2
{{0x0F,0x12},{0x00,0x2A}},    // 70003BB4
{{0x0F,0x12},{0x49,0x32}},    // 70003BB6
{{0x0F,0x12},{0x48,0x32}},    // 70003BB8
{{0x0F,0x12},{0x25,0x0C}},    // 70003BBA
{{0x0F,0x12},{0xF0,0x00}},    // 70003BBC
{{0x0F,0x12},{0xFC,0x4C}},    // 70003BBE
{{0x0F,0x12},{0x00,0x2A}},    // 70003BC0
{{0x0F,0x12},{0x49,0x31}},    // 70003BC2
{{0x0F,0x12},{0x48,0x31}},    // 70003BC4
{{0x0F,0x12},{0x25,0x0D}},    // 70003BC6
{{0x0F,0x12},{0xF0,0x00}},    // 70003BC8
{{0x0F,0x12},{0xFC,0x46}},    // 70003BCA
{{0x0F,0x12},{0x00,0x2A}},    // 70003BCC
{{0x0F,0x12},{0x49,0x30}},    // 70003BCE
{{0x0F,0x12},{0x48,0x30}},    // 70003BD0
{{0x0F,0x12},{0x25,0x0E}},    // 70003BD2
{{0x0F,0x12},{0xF0,0x00}},	// 70003BD4              
{{0x0F,0x12},{0xFC,0x40}},    // 70003BD6
{{0x0F,0x12},{0x00,0x2A}},    // 70003BD8
{{0x0F,0x12},{0x49,0x2F}},    // 70003BDA
{{0x0F,0x12},{0x48,0x2F}},    // 70003BDC
{{0x0F,0x12},{0x25,0x0F}},    // 70003BDE
{{0x0F,0x12},{0xF0,0x00}},    // 70003BE0
{{0x0F,0x12},{0xFC,0x3A}},    // 70003BE2
{{0x0F,0x12},{0x86,0x26}},    // 70003BE4
{{0x0F,0x12},{0x20,0xFF}},    // 70003BE6
{{0x0F,0x12},{0x1C,0x40}},    // 70003BE8
{{0x0F,0x12},{0x86,0x60}},    // 70003BEA
{{0x0F,0x12},{0x48,0x2C}},    // 70003BEC
{{0x0F,0x12},{0x64,0xF8}},    // 70003BEE
{{0x0F,0x12},{0x49,0x2C}},    // 70003BF0
{{0x0F,0x12},{0x48,0x2D}},    // 70003BF2
{{0x0F,0x12},{0x24,0x10}},    // 70003BF4
{{0x0F,0x12},{0x00,0x2A}},    // 70003BF6
{{0x0F,0x12},{0xF0,0x00}},    // 70003BF8
{{0x0F,0x12},{0xFC,0x2E}},    // 70003BFA
{{0x0F,0x12},{0x49,0x2B}},    // 70003BFC
{{0x0F,0x12},{0x48,0x2C}},    // 70003BFE
{{0x0F,0x12},{0x00,0x22}},    // 70003C00
{{0x0F,0x12},{0xF0,0x00}},    // 70003C02
{{0x0F,0x12},{0xFC,0x29}},    // 70003C04
{{0x0F,0x12},{0xBC,0xF8}},    // 70003C06
{{0x0F,0x12},{0xBC,0x08}},    // 70003C08
{{0x0F,0x12},{0x47,0x18}},    // 70003C0A
{{0x0F,0x12},{0x01,0x9C}},    // 70003C0C
{{0x0F,0x12},{0x4E,0xC2}},    // 70003C0E
{{0x0F,0x12},{0x73,0xFF}},    // 70003C10
{{0x0F,0x12},{0x00,0x00}},    // 70003C12
{{0x0F,0x12},{0x1F,0x90}},    // 70003C14
{{0x0F,0x12},{0x70,0x00}},	// 70003C16              
{{0x0F,0x12},{0x3C,0xCD}},    // 70003C18
{{0x0F,0x12},{0x70,0x00}},    // 70003C1A
{{0x0F,0x12},{0xE3,0x8B}},    // 70003C1C
{{0x0F,0x12},{0x00,0x00}},    // 70003C1E
{{0x0F,0x12},{0x3D,0x05}},    // 70003C20
{{0x0F,0x12},{0x70,0x00}},	// 70003C22              
{{0x0F,0x12},{0xC3,0xB1}},    // 70003C24
{{0x0F,0x12},{0x00,0x00}},	// 70003C26              
{{0x0F,0x12},{0x47,0x80}},    // 70003C28
{{0x0F,0x12},{0x70,0x00}},    // 70003C2A
{{0x0F,0x12},{0x3D,0x63}},    // 70003C2C
{{0x0F,0x12},{0x70,0x00}},    // 70003C2E
{{0x0F,0x12},{0x00,0x80}},    // 70003C30
{{0x0F,0x12},{0x70,0x00}},    // 70003C32
{{0x0F,0x12},{0x3D,0x9F}},    // 70003C34
{{0x0F,0x12},{0x70,0x00}},    // 70003C36
{{0x0F,0x12},{0xB4,0x9D}},    // 70003C38
{{0x0F,0x12},{0x00,0x00}},    // 70003C3A
{{0x0F,0x12},{0x3E,0x4B}},    // 70003C3C
{{0x0F,0x12},{0x70,0x00}},    // 70003C3E
{{0x0F,0x12},{0x3D,0xFF}},    // 70003C40
{{0x0F,0x12},{0x70,0x00}},	// 70003C42              
{{0x0F,0x12},{0xFF,0xFF}},    // 70003C44
{{0x0F,0x12},{0x00,0xFF}},    // 70003C46
{{0x0F,0x12},{0x17,0xE0}},    // 70003C48
{{0x0F,0x12},{0x70,0x00}},	// 70003C4A              
{{0x0F,0x12},{0x3F,0xC7}},    // 70003C4C
{{0x0F,0x12},{0x70,0x00}},    // 70003C4E
{{0x0F,0x12},{0x05,0x3D}},    // 70003C50
{{0x0F,0x12},{0x00,0x00}},    // 70003C52
{{0x0F,0x12},{0x00,0x00}},    // 70003C54
{{0x0F,0x12},{0x0A,0x89}},    // 70003C56
{{0x0F,0x12},{0x6C,0xD2}},    // 70003C58
{{0x0F,0x12},{0x00,0x00}},    // 70003C5A
{{0x0F,0x12},{0x02,0xC9}},    // 70003C5C
{{0x0F,0x12},{0x00,0x00}},	// 70003C5E              
{{0x0F,0x12},{0x00,0x00}},    // 70003C60
{{0x0F,0x12},{0x0A,0x9A}},    // 70003C62
{{0x0F,0x12},{0x00,0x00}},    // 70003C64
{{0x0F,0x12},{0x02,0xD2}},    // 70003C66
{{0x0F,0x12},{0x40,0x15}},    // 70003C68
{{0x0F,0x12},{0x70,0x00}},    // 70003C6A
{{0x0F,0x12},{0x9E,0x65}},    // 70003C6C
{{0x0F,0x12},{0x00,0x00}},    // 70003C6E
{{0x0F,0x12},{0x40,0x89}},    // 70003C70
{{0x0F,0x12},{0x70,0x00}},    // 70003C72
{{0x0F,0x12},{0x7C,0x49}},    // 70003C74
{{0x0F,0x12},{0x00,0x00}},    // 70003C76
{{0x0F,0x12},{0x40,0xFD}},    // 70003C78
{{0x0F,0x12},{0x70,0x00}},    // 70003C7A
{{0x0F,0x12},{0x7C,0x63}},    // 70003C7C
{{0x0F,0x12},{0x00,0x00}},    // 70003C7E
{{0x0F,0x12},{0x41,0x19}},    // 70003C80
{{0x0F,0x12},{0x70,0x00}},    // 70003C82
{{0x0F,0x12},{0x8F,0x01}},    // 70003C84
{{0x0F,0x12},{0x00,0x00}},    // 70003C86
{{0x0F,0x12},{0x41,0xBB}},    // 70003C88
{{0x0F,0x12},{0x70,0x00}},    // 70003C8A
{{0x0F,0x12},{0x7F,0x3F}},    // 70003C8C
{{0x0F,0x12},{0x00,0x00}},    // 70003C8E
{{0x0F,0x12},{0x42,0x49}},    // 70003C90
{{0x0F,0x12},{0x70,0x00}},    // 70003C92
{{0x0F,0x12},{0x98,0xC5}},    // 70003C94
{{0x0F,0x12},{0x00,0x00}},    // 70003C96
{{0x0F,0x12},{0x43,0xB5}},    // 70003C98
{{0x0F,0x12},{0x70,0x00}},    // 70003C9A
{{0x0F,0x12},{0x60,0x99}},    // 70003C9C
{{0x0F,0x12},{0x00,0x00}},    // 70003C9E
{{0x0F,0x12},{0x43,0x0F}},    // 70003CA0
{{0x0F,0x12},{0x70,0x00}},    // 70003CA2
{{0x0F,0x12},{0x43,0x65}},    // 70003CA4
{{0x0F,0x12},{0x70,0x00}},    // 70003CA6
{{0x0F,0x12},{0xA7,0x0B}},    // 70003CA8
{{0x0F,0x12},{0x00,0x00}},    // 70003CAA
{{0x0F,0x12},{0x43,0x87}},    // 70003CAC
{{0x0F,0x12},{0x70,0x00}},    // 70003CAE
{{0x0F,0x12},{0x40,0x0D}},    // 70003CB0
{{0x0F,0x12},{0x00,0x00}},    // 70003CB2
{{0x0F,0x12},{0xB5,0x70}},    // 70003CB4
{{0x0F,0x12},{0x00,0x0C}},    // 70003CB6
{{0x0F,0x12},{0x00,0x15}},    // 70003CB8
{{0x0F,0x12},{0x00,0x29}},    // 70003CBA
{{0x0F,0x12},{0xF0,0x00}},    // 70003CBC
{{0x0F,0x12},{0xFB,0xD4}},    // 70003CBE
{{0x0F,0x12},{0x49,0xF8}},    // 70003CC0
{{0x0F,0x12},{0x00,0xA8}},    // 70003CC2
{{0x0F,0x12},{0x50,0x0C}},    // 70003CC4
{{0x0F,0x12},{0xBC,0x70}},    // 70003CC6
{{0x0F,0x12},{0xBC,0x08}},    // 70003CC8
{{0x0F,0x12},{0x47,0x18}},    // 70003CCA
{{0x0F,0x12},{0x68,0x08}},    // 70003CCC
{{0x0F,0x12},{0x04,0x00}},    // 70003CCE
{{0x0F,0x12},{0x0C,0x00}},    // 70003CD0
{{0x0F,0x12},{0x68,0x49}},    // 70003CD2
{{0x0F,0x12},{0x04,0x09}},    // 70003CD4
{{0x0F,0x12},{0x0C,0x09}},    // 70003CD6
{{0x0F,0x12},{0x4A,0xF3}},    // 70003CD8
{{0x0F,0x12},{0x89,0x92}},    // 70003CDA
{{0x0F,0x12},{0x2A,0x00}},    // 70003CDC
{{0x0F,0x12},{0xD0,0x0D}},    // 70003CDE
{{0x0F,0x12},{0x23,0x00}},    // 70003CE0
{{0x0F,0x12},{0x1A,0x89}},    // 70003CE2
{{0x0F,0x12},{0xD4,0x00}},    // 70003CE4
{{0x0F,0x12},{0x00,0x0B}},    // 70003CE6
{{0x0F,0x12},{0x04,0x19}},    // 70003CE8
{{0x0F,0x12},{0x0C,0x09}},    // 70003CEA
{{0x0F,0x12},{0x23,0xFF}},    // 70003CEC
{{0x0F,0x12},{0x33,0xC1}},    // 70003CEE
{{0x0F,0x12},{0x18,0x10}},    // 70003CF0
{{0x0F,0x12},{0x42,0x98}},    // 70003CF2
{{0x0F,0x12},{0xD8,0x00}},    // 70003CF4
{{0x0F,0x12},{0x00,0x03}},    // 70003CF6
{{0x0F,0x12},{0x04,0x18}},    // 70003CF8
{{0x0F,0x12},{0x0C,0x00}},    // 70003CFA
{{0x0F,0x12},{0x4A,0xEB}},    // 70003CFC
{{0x0F,0x12},{0x81,0x50}},    // 70003CFE
{{0x0F,0x12},{0x81,0x91}},    // 70003D00
{{0x0F,0x12},{0x47,0x70}},    // 70003D02
{{0x0F,0x12},{0xB5,0xF3}},    // 70003D04
{{0x0F,0x12},{0x00,0x04}},    // 70003D06
{{0x0F,0x12},{0xB0,0x81}},    // 70003D08
{{0x0F,0x12},{0x98,0x02}},    // 70003D0A
{{0x0F,0x12},{0x68,0x00}},    // 70003D0C
{{0x0F,0x12},{0x06,0x00}},    // 70003D0E
{{0x0F,0x12},{0x0E,0x00}},    // 70003D10
{{0x0F,0x12},{0x22,0x01}},    // 70003D12
{{0x0F,0x12},{0x00,0x15}},    // 70003D14
{{0x0F,0x12},{0x00,0x21}},    // 70003D16
{{0x0F,0x12},{0x39,0x10}},    // 70003D18
{{0x0F,0x12},{0x40,0x8A}},    // 70003D1A
{{0x0F,0x12},{0x40,0xA5}},    // 70003D1C
{{0x0F,0x12},{0x4F,0xE4}},    // 70003D1E
{{0x0F,0x12},{0x00,0x16}},    // 70003D20
{{0x0F,0x12},{0x2C,0x10}},    // 70003D22
{{0x0F,0x12},{0xDA,0x03}},    // 70003D24
{{0x0F,0x12},{0x88,0x39}},    // 70003D26
{{0x0F,0x12},{0x43,0xA9}},    // 70003D28
{{0x0F,0x12},{0x80,0x39}},    // 70003D2A
{{0x0F,0x12},{0xE0,0x02}},    // 70003D2C
{{0x0F,0x12},{0x88,0x79}},    // 70003D2E
{{0x0F,0x12},{0x43,0xB1}},    // 70003D30
{{0x0F,0x12},{0x80,0x79}},    // 70003D32
{{0x0F,0x12},{0xF0,0x00}},    // 70003D34
{{0x0F,0x12},{0xFB,0xA0}},    // 70003D36
{{0x0F,0x12},{0x2C,0x10}},    // 70003D38
{{0x0F,0x12},{0xDA,0x03}},    // 70003D3A
{{0x0F,0x12},{0x88,0x39}},    // 70003D3C
{{0x0F,0x12},{0x43,0x29}},    // 70003D3E
{{0x0F,0x12},{0x80,0x39}},    // 70003D40
{{0x0F,0x12},{0xE0,0x02}},    // 70003D42
{{0x0F,0x12},{0x88,0x79}},    // 70003D44
{{0x0F,0x12},{0x43,0x31}},    // 70003D46
{{0x0F,0x12},{0x80,0x79}},    // 70003D48
{{0x0F,0x12},{0x49,0xDA}},    // 70003D4A
{{0x0F,0x12},{0x88,0x09}},    // 70003D4C
{{0x0F,0x12},{0x29,0x00}},    // 70003D4E
{{0x0F,0x12},{0xD1,0x02}},    // 70003D50
{{0x0F,0x12},{0xF0,0x00}},    // 70003D52
{{0x0F,0x12},{0xFB,0x99}},    // 70003D54
{{0x0F,0x12},{0x20,0x00}},    // 70003D56
{{0x0F,0x12},{0x99,0x02}},    // 70003D58
{{0x0F,0x12},{0x60,0x08}},    // 70003D5A
{{0x0F,0x12},{0xBC,0xFE}},    // 70003D5C
{{0x0F,0x12},{0xBC,0x08}},    // 70003D5E
{{0x0F,0x12},{0x47,0x18}},    // 70003D60
{{0x0F,0x12},{0xB5,0x38}},    // 70003D62
{{0x0F,0x12},{0x9C,0x04}},    // 70003D64
{{0x0F,0x12},{0x00,0x15}},    // 70003D66
{{0x0F,0x12},{0x00,0x2A}},    // 70003D68
{{0x0F,0x12},{0x94,0x00}},    // 70003D6A
{{0x0F,0x12},{0xF0,0x00}},    // 70003D6C
{{0x0F,0x12},{0xFB,0x94}},    // 70003D6E
{{0x0F,0x12},{0x4A,0xD1}},    // 70003D70
{{0x0F,0x12},{0x88,0x11}},    // 70003D72
{{0x0F,0x12},{0x29,0x00}},    // 70003D74
{{0x0F,0x12},{0xD0,0x0F}},    // 70003D76
{{0x0F,0x12},{0x88,0x20}},    // 70003D78
{{0x0F,0x12},{0x42,0x81}},    // 70003D7A
{{0x0F,0x12},{0xD2,0x0C}},    // 70003D7C
{{0x0F,0x12},{0x88,0x61}},    // 70003D7E
{{0x0F,0x12},{0x88,0x53}},    // 70003D80
{{0x0F,0x12},{0x42,0x99}},    // 70003D82
{{0x0F,0x12},{0xD2,0x00}},    // 70003D84
{{0x0F,0x12},{0x1E,0x40}},    // 70003D86
{{0x0F,0x12},{0x04,0x00}},    // 70003D88
{{0x0F,0x12},{0x0C,0x00}},    // 70003D8A
{{0x0F,0x12},{0x80,0x20}},    // 70003D8C
{{0x0F,0x12},{0x88,0x51}},    // 70003D8E
{{0x0F,0x12},{0x80,0x61}},    // 70003D90
{{0x0F,0x12},{0x43,0x68}},    // 70003D92
{{0x0F,0x12},{0x18,0x40}},    // 70003D94
{{0x0F,0x12},{0x60,0x60}},    // 70003D96
{{0x0F,0x12},{0xBC,0x38}},    // 70003D98
{{0x0F,0x12},{0xBC,0x08}},    // 70003D9A
{{0x0F,0x12},{0x47,0x18}},    // 70003D9C
{{0x0F,0x12},{0xB5,0xF8}},    // 70003D9E
{{0x0F,0x12},{0x00,0x04}},    // 70003DA0
{{0x0F,0x12},{0x68,0x08}},    // 70003DA2
{{0x0F,0x12},{0x04,0x00}},    // 70003DA4
{{0x0F,0x12},{0x0C,0x00}},    // 70003DA6
{{0x0F,0x12},{0x22,0x01}},    // 70003DA8
{{0x0F,0x12},{0x00,0x15}},    // 70003DAA
{{0x0F,0x12},{0x00,0x21}},    // 70003DAC
{{0x0F,0x12},{0x39,0x10}},    // 70003DAE
{{0x0F,0x12},{0x40,0x8A}},    // 70003DB0
{{0x0F,0x12},{0x40,0xA5}},    // 70003DB2
{{0x0F,0x12},{0x4F,0xBE}},    // 70003DB4
{{0x0F,0x12},{0x00,0x16}},    // 70003DB6
{{0x0F,0x12},{0x2C,0x10}},    // 70003DB8
{{0x0F,0x12},{0xDA,0x03}},    // 70003DBA
{{0x0F,0x12},{0x88,0x39}},    // 70003DBC
{{0x0F,0x12},{0x43,0xA9}},    // 70003DBE
{{0x0F,0x12},{0x80,0x39}},    // 70003DC0
{{0x0F,0x12},{0xE0,0x02}},    // 70003DC2
{{0x0F,0x12},{0x88,0x79}},    // 70003DC4
{{0x0F,0x12},{0x43,0xB1}},    // 70003DC6
{{0x0F,0x12},{0x80,0x79}},    // 70003DC8
{{0x0F,0x12},{0xF0,0x00}},    // 70003DCA
{{0x0F,0x12},{0xFB,0x6D}},    // 70003DCC
{{0x0F,0x12},{0x2C,0x10}},    // 70003DCE
{{0x0F,0x12},{0xDA,0x03}},    // 70003DD0
{{0x0F,0x12},{0x88,0x38}},    // 70003DD2
{{0x0F,0x12},{0x43,0x28}},    // 70003DD4
{{0x0F,0x12},{0x80,0x38}},    // 70003DD6
{{0x0F,0x12},{0xE0,0x02}},    // 70003DD8
{{0x0F,0x12},{0x88,0x78}},    // 70003DDA
{{0x0F,0x12},{0x43,0x30}},    // 70003DDC
{{0x0F,0x12},{0x80,0x78}},    // 70003DDE
{{0x0F,0x12},{0x48,0xB6}},    // 70003DE0
{{0x0F,0x12},{0x88,0x00}},    // 70003DE2
{{0x0F,0x12},{0x04,0x00}},    // 70003DE4
{{0x0F,0x12},{0xD5,0x07}},    // 70003DE6
{{0x0F,0x12},{0x4B,0xB5}},    // 70003DE8
{{0x0F,0x12},{0x78,0x19}},    // 70003DEA
{{0x0F,0x12},{0x4A,0xB5}},    // 70003DEC
{{0x0F,0x12},{0x78,0x10}},    // 70003DEE
{{0x0F,0x12},{0x70,0x18}},    // 70003DF0
{{0x0F,0x12},{0x70,0x11}},    // 70003DF2
{{0x0F,0x12},{0x49,0xB4}},    // 70003DF4
{{0x0F,0x12},{0x81,0x88}},    // 70003DF6
{{0x0F,0x12},{0xBC,0xF8}},    // 70003DF8
{{0x0F,0x12},{0xBC,0x08}},    // 70003DFA
{{0x0F,0x12},{0x47,0x18}},    // 70003DFC
{{0x0F,0x12},{0xB5,0x38}},    // 70003DFE
{{0x0F,0x12},{0x48,0xB2}},    // 70003E00
{{0x0F,0x12},{0x46,0x69}},    // 70003E02
{{0x0F,0x12},{0xF0,0x00}},    // 70003E04
{{0x0F,0x12},{0xFB,0x58}},    // 70003E06
{{0x0F,0x12},{0x48,0xB1}},    // 70003E08
{{0x0F,0x12},{0x49,0xB0}},    // 70003E0A
{{0x0F,0x12},{0x69,0xC2}},    // 70003E0C
{{0x0F,0x12},{0x24,0x00}},    // 70003E0E
{{0x0F,0x12},{0x31,0xA8}},    // 70003E10
{{0x0F,0x12},{0x2A,0x00}},    // 70003E12
{{0x0F,0x12},{0xD0,0x08}},    // 70003E14
{{0x0F,0x12},{0x61,0xC4}},    // 70003E16
{{0x0F,0x12},{0x68,0x4A}},    // 70003E18
{{0x0F,0x12},{0x62,0x42}},    // 70003E1A
{{0x0F,0x12},{0x62,0x82}},    // 70003E1C
{{0x0F,0x12},{0x46,0x6B}},    // 70003E1E
{{0x0F,0x12},{0x88,0x1A}},    // 70003E20
{{0x0F,0x12},{0x63,0x02}},    // 70003E22
{{0x0F,0x12},{0x88,0x5A}},    // 70003E24
{{0x0F,0x12},{0x63,0x42}},    // 70003E26
{{0x0F,0x12},{0x6A,0x02}},    // 70003E28
{{0x0F,0x12},{0x2A,0x00}},    // 70003E2A
{{0x0F,0x12},{0xD0,0x0A}},    // 70003E2C
{{0x0F,0x12},{0x62,0x04}},    // 70003E2E
{{0x0F,0x12},{0x68,0x49}},    // 70003E30
{{0x0F,0x12},{0x62,0x81}},    // 70003E32
{{0x0F,0x12},{0x46,0x6B}},    // 70003E34
{{0x0F,0x12},{0x88,0x19}},    // 70003E36
{{0x0F,0x12},{0x63,0x01}},    // 70003E38
{{0x0F,0x12},{0x88,0x59}},    // 70003E3A
{{0x0F,0x12},{0x63,0x41}},    // 70003E3C
{{0x0F,0x12},{0x49,0xA5}},    // 70003E3E
{{0x0F,0x12},{0x88,0xC9}},    // 70003E40
{{0x0F,0x12},{0x63,0xC1}},    // 70003E42
{{0x0F,0x12},{0xF0,0x00}},	// 70003E44              
{{0x0F,0x12},{0xFB,0x40}},    // 70003E46
{{0x0F,0x12},{0xE7,0xA6}},    // 70003E48
{{0x0F,0x12},{0xB5,0xF0}},    // 70003E4A
{{0x0F,0x12},{0xB0,0x8B}},    // 70003E4C
{{0x0F,0x12},{0x20,0xFF}},    // 70003E4E
{{0x0F,0x12},{0x1C,0x40}},    // 70003E50
{{0x0F,0x12},{0x49,0xA1}},    // 70003E52
{{0x0F,0x12},{0x89,0xCC}},    // 70003E54
{{0x0F,0x12},{0x4E,0x9E}},    // 70003E56
{{0x0F,0x12},{0x6A,0xB1}},    // 70003E58
{{0x0F,0x12},{0x42,0x84}},    // 70003E5A
{{0x0F,0x12},{0xD1,0x01}},    // 70003E5C
{{0x0F,0x12},{0x48,0x9F}},    // 70003E5E
{{0x0F,0x12},{0x60,0x81}},    // 70003E60
{{0x0F,0x12},{0x6A,0x70}},    // 70003E62
{{0x0F,0x12},{0x02,0x00}},    // 70003E64
{{0x0F,0x12},{0xF0,0x00}},    // 70003E66
{{0x0F,0x12},{0xFB,0x37}},    // 70003E68
{{0x0F,0x12},{0x04,0x00}},    // 70003E6A
{{0x0F,0x12},{0x0C,0x00}},    // 70003E6C
{{0x0F,0x12},{0x4A,0x96}},    // 70003E6E
{{0x0F,0x12},{0x8A,0x11}},    // 70003E70
{{0x0F,0x12},{0x91,0x09}},    // 70003E72
{{0x0F,0x12},{0x21,0x01}},    // 70003E74
{{0x0F,0x12},{0x03,0x49}},    // 70003E76
{{0x0F,0x12},{0x42,0x88}},    // 70003E78
{{0x0F,0x12},{0xD2,0x00}},    // 70003E7A
{{0x0F,0x12},{0x00,0x01}},    // 70003E7C
{{0x0F,0x12},{0x4A,0x92}},    // 70003E7E
{{0x0F,0x12},{0x82,0x11}},    // 70003E80
{{0x0F,0x12},{0x4D,0x97}},    // 70003E82
{{0x0F,0x12},{0x88,0x29}},    // 70003E84
{{0x0F,0x12},{0x91,0x08}},    // 70003E86
{{0x0F,0x12},{0x4A,0x8B}},    // 70003E88
{{0x0F,0x12},{0x23,0x03}},    // 70003E8A
{{0x0F,0x12},{0x32,0x22}},    // 70003E8C
{{0x0F,0x12},{0x1F,0x91}},    // 70003E8E
{{0x0F,0x12},{0xF0,0x00}},    // 70003E90
{{0x0F,0x12},{0xFB,0x28}},    // 70003E92
{{0x0F,0x12},{0x80,0x28}},    // 70003E94
{{0x0F,0x12},{0x48,0x8E}},    // 70003E96
{{0x0F,0x12},{0x49,0x87}},    // 70003E98
{{0x0F,0x12},{0x6B,0xC2}},    // 70003E9A
{{0x0F,0x12},{0x6A,0xC0}},    // 70003E9C
{{0x0F,0x12},{0x42,0x82}},    // 70003E9E
{{0x0F,0x12},{0xD2,0x01}},    // 70003EA0
{{0x0F,0x12},{0x8C,0xC8}},    // 70003EA2
{{0x0F,0x12},{0x80,0x28}},    // 70003EA4
{{0x0F,0x12},{0x88,0xE8}},    // 70003EA6
{{0x0F,0x12},{0x90,0x07}},    // 70003EA8
{{0x0F,0x12},{0x22,0x40}},    // 70003EAA
{{0x0F,0x12},{0x43,0x10}},    // 70003EAC
{{0x0F,0x12},{0x80,0xE8}},    // 70003EAE
{{0x0F,0x12},{0x20,0x00}},    // 70003EB0
{{0x0F,0x12},{0x00,0x41}},    // 70003EB2
{{0x0F,0x12},{0x19,0x4B}},    // 70003EB4
{{0x0F,0x12},{0x00,0x1E}},    // 70003EB6
{{0x0F,0x12},{0x36,0x80}},    // 70003EB8
{{0x0F,0x12},{0x8B,0xB2}},    // 70003EBA
{{0x0F,0x12},{0xAF,0x04}},    // 70003EBC
{{0x0F,0x12},{0x52,0x7A}},    // 70003EBE
{{0x0F,0x12},{0x4A,0x7D}},    // 70003EC0
{{0x0F,0x12},{0x18,0x8A}},    // 70003EC2
{{0x0F,0x12},{0x88,0x97}},    // 70003EC4
{{0x0F,0x12},{0x83,0xB7}},    // 70003EC6
{{0x0F,0x12},{0x33,0xA0}},    // 70003EC8
{{0x0F,0x12},{0x89,0x1F}},    // 70003ECA
{{0x0F,0x12},{0xAE,0x01}},    // 70003ECC
{{0x0F,0x12},{0x52,0x77}},    // 70003ECE
{{0x0F,0x12},{0x8A,0x11}},    // 70003ED0
{{0x0F,0x12},{0x81,0x19}},    // 70003ED2
{{0x0F,0x12},{0x1C,0x40}},    // 70003ED4
{{0x0F,0x12},{0x04,0x00}},    // 70003ED6
{{0x0F,0x12},{0x0C,0x00}},    // 70003ED8
{{0x0F,0x12},{0x28,0x06}},    // 70003EDA
{{0x0F,0x12},{0xD3,0xE9}},    // 70003EDC
{{0x0F,0x12},{0xF0,0x00}},    // 70003EDE
{{0x0F,0x12},{0xFB,0x09}},    // 70003EE0
{{0x0F,0x12},{0xF0,0x00}},    // 70003EE2
{{0x0F,0x12},{0xFB,0x0F}},    // 70003EE4
{{0x0F,0x12},{0x4F,0x79}},    // 70003EE6
{{0x0F,0x12},{0x37,0xA8}},    // 70003EE8
{{0x0F,0x12},{0x28,0x00}},    // 70003EEA
{{0x0F,0x12},{0xD1,0x0A}},    // 70003EEC
{{0x0F,0x12},{0x1F,0xE0}},    // 70003EEE
{{0x0F,0x12},{0x38,0xFD}},    // 70003EF0
{{0x0F,0x12},{0xD0,0x01}},    // 70003EF2
{{0x0F,0x12},{0x1C,0xC0}},    // 70003EF4
{{0x0F,0x12},{0xD1,0x05}},    // 70003EF6
{{0x0F,0x12},{0x48,0x74}},    // 70003EF8
{{0x0F,0x12},{0x88,0x29}},    // 70003EFA
{{0x0F,0x12},{0x38,0x18}},    // 70003EFC
{{0x0F,0x12},{0x68,0x40}},    // 70003EFE
{{0x0F,0x12},{0x43,0x48}},    // 70003F00
{{0x0F,0x12},{0x60,0x78}},    // 70003F02
{{0x0F,0x12},{0x49,0x72}},    // 70003F04
{{0x0F,0x12},{0x68,0x78}},    // 70003F06
{{0x0F,0x12},{0x6B,0x89}},    // 70003F08
{{0x0F,0x12},{0x42,0x88}},    // 70003F0A
{{0x0F,0x12},{0xD3,0x00}},    // 70003F0C
{{0x0F,0x12},{0x00,0x08}},    // 70003F0E
{{0x0F,0x12},{0x60,0x78}},    // 70003F10
{{0x0F,0x12},{0x20,0x00}},    // 70003F12
{{0x0F,0x12},{0x00,0x41}},    // 70003F14
{{0x0F,0x12},{0xAA,0x04}},    // 70003F16
{{0x0F,0x12},{0x5A,0x53}},    // 70003F18
{{0x0F,0x12},{0x19,0x4A}},    // 70003F1A
{{0x0F,0x12},{0x26,0x9C}},    // 70003F1C
{{0x0F,0x12},{0x52,0xB3}},    // 70003F1E
{{0x0F,0x12},{0xAB,0x01}},    // 70003F20
{{0x0F,0x12},{0x5A,0x59}},    // 70003F22
{{0x0F,0x12},{0x32,0xA0}},    // 70003F24
{{0x0F,0x12},{0x81,0x11}},    // 70003F26
{{0x0F,0x12},{0x1C,0x40}},    // 70003F28
{{0x0F,0x12},{0x04,0x00}},    // 70003F2A
{{0x0F,0x12},{0x0C,0x00}},    // 70003F2C
{{0x0F,0x12},{0x28,0x06}},    // 70003F2E
{{0x0F,0x12},{0xD3,0xF0}},    // 70003F30
{{0x0F,0x12},{0x49,0x65}},    // 70003F32
{{0x0F,0x12},{0x98,0x09}},    // 70003F34
{{0x0F,0x12},{0x82,0x08}},    // 70003F36
{{0x0F,0x12},{0x98,0x08}},    // 70003F38
{{0x0F,0x12},{0x80,0x28}},    // 70003F3A
{{0x0F,0x12},{0x98,0x07}},    // 70003F3C
{{0x0F,0x12},{0x80,0xE8}},    // 70003F3E
{{0x0F,0x12},{0x1F,0xE0}},    // 70003F40
{{0x0F,0x12},{0x38,0xFD}},    // 70003F42
{{0x0F,0x12},{0xD1,0x3B}},    // 70003F44
{{0x0F,0x12},{0x4D,0x64}},    // 70003F46
{{0x0F,0x12},{0x89,0xE8}},    // 70003F48
{{0x0F,0x12},{0x1F,0xC1}},    // 70003F4A
{{0x0F,0x12},{0x39,0xFF}},    // 70003F4C
{{0x0F,0x12},{0xD1,0x36}},    // 70003F4E
{{0x0F,0x12},{0x4C,0x5F}},    // 70003F50
{{0x0F,0x12},{0x8A,0xE0}},    // 70003F52
{{0x0F,0x12},{0xF0,0x00}},    // 70003F54
{{0x0F,0x12},{0xFA,0xDE}},    // 70003F56
{{0x0F,0x12},{0x00,0x06}},    // 70003F58
{{0x0F,0x12},{0x8B,0x20}},    // 70003F5A
{{0x0F,0x12},{0xF0,0x00}},    // 70003F5C
{{0x0F,0x12},{0xFA,0xE2}},    // 70003F5E
{{0x0F,0x12},{0x90,0x00}},    // 70003F60
{{0x0F,0x12},{0x6A,0xA1}},    // 70003F62
{{0x0F,0x12},{0x68,0x78}},    // 70003F64
{{0x0F,0x12},{0x18,0x09}},    // 70003F66
{{0x0F,0x12},{0x02,0x00}},    // 70003F68
{{0x0F,0x12},{0xF0,0x00}},    // 70003F6A
{{0x0F,0x12},{0xFA,0xB5}},    // 70003F6C
{{0x0F,0x12},{0x04,0x00}},    // 70003F6E
{{0x0F,0x12},{0x0C,0x00}},    // 70003F70
{{0x0F,0x12},{0x00,0x22}},    // 70003F72
{{0x0F,0x12},{0x32,0x46}},    // 70003F74
{{0x0F,0x12},{0x00,0x11}},    // 70003F76
{{0x0F,0x12},{0x31,0x0A}},    // 70003F78
{{0x0F,0x12},{0x23,0x05}},    // 70003F7A
{{0x0F,0x12},{0xF0,0x00}},    // 70003F7C
{{0x0F,0x12},{0xFA,0xB2}},    // 70003F7E
{{0x0F,0x12},{0x66,0xE8}},    // 70003F80
{{0x0F,0x12},{0x6B,0x23}},    // 70003F82
{{0x0F,0x12},{0x00,0x02}},    // 70003F84
{{0x0F,0x12},{0x00,0x31}},    // 70003F86
{{0x0F,0x12},{0x00,0x18}},    // 70003F88
{{0x0F,0x12},{0xF0,0x00}},    // 70003F8A
{{0x0F,0x12},{0xFA,0xD3}},    // 70003F8C
{{0x0F,0x12},{0x46,0x6B}},    // 70003F8E
{{0x0F,0x12},{0x85,0x18}},    // 70003F90
{{0x0F,0x12},{0x6E,0xEA}},    // 70003F92
{{0x0F,0x12},{0x6B,0x60}},    // 70003F94
{{0x0F,0x12},{0x99,0x00}},    // 70003F96
{{0x0F,0x12},{0xF0,0x00}},    // 70003F98
{{0x0F,0x12},{0xFA,0xCC}},    // 70003F9A
{{0x0F,0x12},{0x46,0x6B}},    // 70003F9C
{{0x0F,0x12},{0x85,0x58}},    // 70003F9E
{{0x0F,0x12},{0x00,0x29}},    // 70003FA0
{{0x0F,0x12},{0x98,0x0A}},    // 70003FA2
{{0x0F,0x12},{0x31,0x70}},    // 70003FA4
{{0x0F,0x12},{0xF0,0x00}},    // 70003FA6
{{0x0F,0x12},{0xFA,0xCD}},    // 70003FA8
{{0x0F,0x12},{0x00,0x28}},    // 70003FAA
{{0x0F,0x12},{0x30,0x60}},    // 70003FAC
{{0x0F,0x12},{0x8A,0x02}},    // 70003FAE
{{0x0F,0x12},{0x49,0x46}},    // 70003FB0
{{0x0F,0x12},{0x31,0x28}},    // 70003FB2
{{0x0F,0x12},{0x80,0x8A}},    // 70003FB4
{{0x0F,0x12},{0x8A,0x42}},    // 70003FB6
{{0x0F,0x12},{0x80,0xCA}},    // 70003FB8
{{0x0F,0x12},{0x8A,0x80}},    // 70003FBA
{{0x0F,0x12},{0x81,0x08}},    // 70003FBC
{{0x0F,0x12},{0xB0,0x0B}},    // 70003FBE
{{0x0F,0x12},{0xBC,0xF0}},    // 70003FC0
{{0x0F,0x12},{0xBC,0x08}},    // 70003FC2
{{0x0F,0x12},{0x47,0x18}},    // 70003FC4
{{0x0F,0x12},{0xB5,0x70}},    // 70003FC6
{{0x0F,0x12},{0x24,0x00}},    // 70003FC8
{{0x0F,0x12},{0x4D,0x46}},    // 70003FCA
{{0x0F,0x12},{0x48,0x46}},    // 70003FCC
{{0x0F,0x12},{0x88,0x81}},    // 70003FCE
{{0x0F,0x12},{0x48,0x46}},    // 70003FD0
{{0x0F,0x12},{0x80,0x41}},    // 70003FD2
{{0x0F,0x12},{0x21,0x01}},    // 70003FD4
{{0x0F,0x12},{0x80,0x01}},    // 70003FD6
{{0x0F,0x12},{0xF0,0x00}},    // 70003FD8
{{0x0F,0x12},{0xFA,0xBC}},    // 70003FDA
{{0x0F,0x12},{0x48,0x42}},    // 70003FDC
{{0x0F,0x12},{0x38,0x20}},    // 70003FDE
{{0x0F,0x12},{0x8B,0xC0}},    // 70003FE0
{{0x0F,0x12},{0xF0,0x00}},    // 70003FE2
{{0x0F,0x12},{0xFA,0xBF}},    // 70003FE4
{{0x0F,0x12},{0x4B,0x42}},    // 70003FE6
{{0x0F,0x12},{0x22,0x0D}},    // 70003FE8
{{0x0F,0x12},{0x07,0x12}},    // 70003FEA
{{0x0F,0x12},{0x18,0xA8}},    // 70003FEC
{{0x0F,0x12},{0x88,0x06}},    // 70003FEE
{{0x0F,0x12},{0x00,0xE1}},    // 70003FF0
{{0x0F,0x12},{0x18,0xC9}},    // 70003FF2
{{0x0F,0x12},{0x81,0xCE}},    // 70003FF4
{{0x0F,0x12},{0x88,0x46}},    // 70003FF6
{{0x0F,0x12},{0x81,0x8E}},    // 70003FF8
{{0x0F,0x12},{0x88,0x86}},    // 70003FFA
{{0x0F,0x12},{0x82,0x4E}},    // 70003FFC
{{0x0F,0x12},{0x88,0xC0}},    // 70003FFE
{{0x0F,0x12},{0x82,0x08}},    // 70004000
{{0x0F,0x12},{0x35,0x08}},    // 70004002
{{0x0F,0x12},{0x04,0x2D}},    // 70004004
{{0x0F,0x12},{0x0C,0x2D}},    // 70004006
{{0x0F,0x12},{0x1C,0x64}},    // 70004008
{{0x0F,0x12},{0x04,0x24}},    // 7000400A
{{0x0F,0x12},{0x0C,0x24}},    // 7000400C
{{0x0F,0x12},{0x2C,0x07}},    // 7000400E
{{0x0F,0x12},{0xD3,0xEC}},    // 70004010
{{0x0F,0x12},{0xE6,0x58}},    // 70004012
{{0x0F,0x12},{0xB5,0x10}},    // 70004014
{{0x0F,0x12},{0x48,0x34}},    // 70004016
{{0x0F,0x12},{0x4C,0x34}},    // 70004018
{{0x0F,0x12},{0x88,0xC0}},    // 7000401A
{{0x0F,0x12},{0x80,0x60}},    // 7000401C
{{0x0F,0x12},{0x20,0x01}},    // 7000401E
{{0x0F,0x12},{0x80,0x20}},    // 70004020
{{0x0F,0x12},{0x48,0x31}},    // 70004022
{{0x0F,0x12},{0x38,0x20}},    // 70004024
{{0x0F,0x12},{0x8B,0xC0}},    // 70004026
{{0x0F,0x12},{0xF0,0x00}},    // 70004028
{{0x0F,0x12},{0xFA,0x9C}},    // 7000402A
{{0x0F,0x12},{0x88,0xE0}},    // 7000402C
{{0x0F,0x12},{0x4A,0x31}},    // 7000402E
{{0x0F,0x12},{0x28,0x00}},    // 70004030
{{0x0F,0x12},{0xD0,0x03}},    // 70004032
{{0x0F,0x12},{0x49,0x30}},    // 70004034
{{0x0F,0x12},{0x88,0x49}},    // 70004036
{{0x0F,0x12},{0x29,0x00}},    // 70004038
{{0x0F,0x12},{0xD0,0x09}},    // 7000403A
{{0x0F,0x12},{0x20,0x01}},    // 7000403C
{{0x0F,0x12},{0x03,0xC0}},    // 7000403E
{{0x0F,0x12},{0x80,0x50}},    // 70004040
{{0x0F,0x12},{0x80,0xD0}},    // 70004042
{{0x0F,0x12},{0x20,0x00}},    // 70004044
{{0x0F,0x12},{0x80,0x90}},    // 70004046
{{0x0F,0x12},{0x81,0x10}},    // 70004048
{{0x0F,0x12},{0xBC,0x10}},    // 7000404A
{{0x0F,0x12},{0xBC,0x08}},    // 7000404C
{{0x0F,0x12},{0x47,0x18}},    // 7000404E
{{0x0F,0x12},{0x80,0x50}},    // 70004050
{{0x0F,0x12},{0x89,0x20}},    // 70004052
{{0x0F,0x12},{0x80,0xD0}},    // 70004054
{{0x0F,0x12},{0x89,0x60}},    // 70004056
{{0x0F,0x12},{0x04,0x00}},    // 70004058
{{0x0F,0x12},{0x14,0x00}},    // 7000405A
{{0x0F,0x12},{0x80,0x90}},    // 7000405C
{{0x0F,0x12},{0x89,0xA1}},    // 7000405E
{{0x0F,0x12},{0x04,0x09}},    // 70004060
{{0x0F,0x12},{0x14,0x09}},    // 70004062
{{0x0F,0x12},{0x81,0x11}},    // 70004064
{{0x0F,0x12},{0x89,0xE3}},    // 70004066
{{0x0F,0x12},{0x8A,0x24}},    // 70004068
{{0x0F,0x12},{0x2B,0x00}},    // 7000406A
{{0x0F,0x12},{0xD1,0x04}},    // 7000406C
{{0x0F,0x12},{0x17,0xC3}},    // 7000406E
{{0x0F,0x12},{0x0F,0x5B}},    // 70004070
{{0x0F,0x12},{0x18,0x18}},    // 70004072
{{0x0F,0x12},{0x10,0xC0}},    // 70004074
{{0x0F,0x12},{0x80,0x90}},    // 70004076
{{0x0F,0x12},{0x2C,0x00}},    // 70004078
{{0x0F,0x12},{0xD1,0xE6}},    // 7000407A
{{0x0F,0x12},{0x17,0xC8}},    // 7000407C
{{0x0F,0x12},{0x0F,0x40}},    // 7000407E
{{0x0F,0x12},{0x18,0x40}},    // 70004080
{{0x0F,0x12},{0x10,0xC0}},    // 70004082
{{0x0F,0x12},{0x81,0x10}},    // 70004084
{{0x0F,0x12},{0xE7,0xE0}},    // 70004086
{{0x0F,0x12},{0xB5,0x10}},    // 70004088
{{0x0F,0x12},{0x00,0x0C}},    // 7000408A
{{0x0F,0x12},{0x49,0x19}},    // 7000408C
{{0x0F,0x12},{0x22,0x04}},    // 7000408E
{{0x0F,0x12},{0x68,0x20}},    // 70004090
{{0x0F,0x12},{0x5E,0x8A}},    // 70004092
{{0x0F,0x12},{0x01,0x40}},    // 70004094
{{0x0F,0x12},{0x1A,0x80}},    // 70004096
{{0x0F,0x12},{0x02,0x80}},    // 70004098
{{0x0F,0x12},{0x88,0x49}},    // 7000409A
{{0x0F,0x12},{0xF0,0x00}},    // 7000409C
{{0x0F,0x12},{0xFA,0x6A}},    // 7000409E
{{0x0F,0x12},{0x60,0x20}},    // 700040A0
{{0x0F,0x12},{0xE7,0xD2}},    // 700040A2
{{0x0F,0x12},{0x38,0xD4}},    // 700040A4
{{0x0F,0x12},{0x70,0x00}},	// 700040A6              
{{0x0F,0x12},{0x17,0xD0}},    // 700040A8
{{0x0F,0x12},{0x70,0x00}},	// 700040AA              
{{0x0F,0x12},{0x50,0x00}},    // 700040AC
{{0x0F,0x12},{0xD0,0x00}},    // 700040AE
{{0x0F,0x12},{0x11,0x00}},    // 700040B0
{{0x0F,0x12},{0xD0,0x00}},    // 700040B2
{{0x0F,0x12},{0x17,0x1A}},    // 700040B4
{{0x0F,0x12},{0x70,0x00}},    // 700040B6
{{0x0F,0x12},{0x47,0x80}},    // 700040B8
{{0x0F,0x12},{0x70,0x00}},    // 700040BA
{{0x0F,0x12},{0x2F,0xCA}},    // 700040BC
{{0x0F,0x12},{0x70,0x00}},    // 700040BE
{{0x0F,0x12},{0x2F,0xC5}},    // 700040C0
{{0x0F,0x12},{0x70,0x00}},    // 700040C2
{{0x0F,0x12},{0x2F,0xC6}},    // 700040C4
{{0x0F,0x12},{0x70,0x00}},    // 700040C6
{{0x0F,0x12},{0x2E,0xD8}},    // 700040C8
{{0x0F,0x12},{0x70,0x00}},    // 700040CA
{{0x0F,0x12},{0x2B,0xD0}},    // 700040CC
{{0x0F,0x12},{0x70,0x00}},    // 700040CE
{{0x0F,0x12},{0x17,0xE0}},    // 700040D0
{{0x0F,0x12},{0x70,0x00}},    // 700040D2
{{0x0F,0x12},{0x2D,0xE8}},    // 700040D4
{{0x0F,0x12},{0x70,0x00}},    // 700040D6
{{0x0F,0x12},{0x37,0xE0}},    // 700040D8
{{0x0F,0x12},{0x70,0x00}},    // 700040DA
{{0x0F,0x12},{0x21,0x0C}},    // 700040DC
{{0x0F,0x12},{0x70,0x00}},    // 700040DE
{{0x0F,0x12},{0x14,0x84}},    // 700040E0
{{0x0F,0x12},{0x70,0x00}},    // 700040E2
{{0x0F,0x12},{0xA0,0x06}},    // 700040E4
{{0x0F,0x12},{0x00,0x00}},    // 700040E6
{{0x0F,0x12},{0x07,0x24}},    // 700040E8
{{0x0F,0x12},{0x70,0x00}},    // 700040EA
{{0x0F,0x12},{0xA0,0x00}},    // 700040EC
{{0x0F,0x12},{0xD0,0x00}},    // 700040EE
{{0x0F,0x12},{0x22,0x70}},    // 700040F0
{{0x0F,0x12},{0x70,0x00}},    // 700040F2
{{0x0F,0x12},{0x25,0x58}},    // 700040F4
{{0x0F,0x12},{0x70,0x00}},    // 700040F6
{{0x0F,0x12},{0x14,0x6C}},    // 700040F8
{{0x0F,0x12},{0x70,0x00}},    // 700040FA
{{0x0F,0x12},{0xB5,0x10}},    // 700040FC
{{0x0F,0x12},{0x00,0x0C}},    // 700040FE
{{0x0F,0x12},{0x49,0xC7}},    // 70004100
{{0x0F,0x12},{0x22,0x08}},    // 70004102
{{0x0F,0x12},{0x68,0x20}},    // 70004104
{{0x0F,0x12},{0x5E,0x8A}},    // 70004106
{{0x0F,0x12},{0x01,0x40}},    // 70004108
{{0x0F,0x12},{0x1A,0x80}},    // 7000410A
{{0x0F,0x12},{0x02,0x80}},    // 7000410C
{{0x0F,0x12},{0x88,0xC9}},    // 7000410E
{{0x0F,0x12},{0xF0,0x00}},    // 70004110
{{0x0F,0x12},{0xFA,0x30}},    // 70004112
{{0x0F,0x12},{0x60,0x20}},    // 70004114
{{0x0F,0x12},{0xE7,0x98}},    // 70004116
{{0x0F,0x12},{0xB5,0xFE}},    // 70004118
{{0x0F,0x12},{0x00,0x0C}},    // 7000411A
{{0x0F,0x12},{0x68,0x25}},    // 7000411C
{{0x0F,0x12},{0x68,0x66}},    // 7000411E
{{0x0F,0x12},{0x68,0xA0}},    // 70004120
{{0x0F,0x12},{0x90,0x01}},    // 70004122
{{0x0F,0x12},{0x68,0xE7}},    // 70004124
{{0x0F,0x12},{0x1B,0xA8}},    // 70004126
{{0x0F,0x12},{0x42,0xB5}},    // 70004128
{{0x0F,0x12},{0xDA,0x00}},    // 7000412A
{{0x0F,0x12},{0x1B,0x70}},    // 7000412C
{{0x0F,0x12},{0x90,0x00}},    // 7000412E
{{0x0F,0x12},{0x49,0xBB}},    // 70004130
{{0x0F,0x12},{0x48,0xBC}},    // 70004132
{{0x0F,0x12},{0x88,0x4A}},    // 70004134
{{0x0F,0x12},{0x88,0x43}},    // 70004136
{{0x0F,0x12},{0x43,0x5A}},    // 70004138
{{0x0F,0x12},{0x23,0x04}},    // 7000413A
{{0x0F,0x12},{0x5E,0xCB}},    // 7000413C
{{0x0F,0x12},{0x0A,0x92}},    // 7000413E
{{0x0F,0x12},{0x18,0xD2}},    // 70004140
{{0x0F,0x12},{0x02,0xD2}},    // 70004142
{{0x0F,0x12},{0x0C,0x12}},    // 70004144
{{0x0F,0x12},{0x88,0xCB}},    // 70004146
{{0x0F,0x12},{0x88,0x80}},    // 70004148
{{0x0F,0x12},{0x43,0x43}},    // 7000414A
{{0x0F,0x12},{0x0A,0x98}},    // 7000414C
{{0x0F,0x12},{0x23,0x08}},    // 7000414E
{{0x0F,0x12},{0x5E,0xCB}},    // 70004150
{{0x0F,0x12},{0x18,0xC0}},    // 70004152
{{0x0F,0x12},{0x02,0xC0}},    // 70004154
{{0x0F,0x12},{0x0C,0x00}},    // 70004156
{{0x0F,0x12},{0x04,0x11}},    // 70004158
{{0x0F,0x12},{0x04,0x00}},    // 7000415A
{{0x0F,0x12},{0x14,0x09}},    // 7000415C
{{0x0F,0x12},{0x14,0x00}},    // 7000415E
{{0x0F,0x12},{0x1A,0x08}},    // 70004160
{{0x0F,0x12},{0x49,0xB0}},    // 70004162
{{0x0F,0x12},{0x39,0xE0}},    // 70004164
{{0x0F,0x12},{0x61,0x48}},    // 70004166
{{0x0F,0x12},{0x98,0x01}},    // 70004168
{{0x0F,0x12},{0x30,0x40}},    // 7000416A
{{0x0F,0x12},{0x78,0x80}},    // 7000416C
{{0x0F,0x12},{0x28,0x00}},    // 7000416E
{{0x0F,0x12},{0xD1,0x03}},    // 70004170
{{0x0F,0x12},{0x98,0x01}},    // 70004172
{{0x0F,0x12},{0x00,0x29}},    // 70004174
{{0x0F,0x12},{0xF0,0x00}},	// 70004176              
{{0x0F,0x12},{0xFA,0x03}},    // 70004178
{{0x0F,0x12},{0x88,0x39}},    // 7000417A
{{0x0F,0x12},{0x98,0x00}},    // 7000417C
{{0x0F,0x12},{0x42,0x81}},    // 7000417E
{{0x0F,0x12},{0xD8,0x14}},    // 70004180
{{0x0F,0x12},{0x88,0x79}},    // 70004182
{{0x0F,0x12},{0x98,0x00}},    // 70004184
{{0x0F,0x12},{0x42,0x81}},    // 70004186
{{0x0F,0x12},{0xD2,0x0C}},    // 70004188
{{0x0F,0x12},{0x98,0x01}},    // 7000418A
{{0x0F,0x12},{0x00,0x29}},    // 7000418C
{{0x0F,0x12},{0xF0,0x00}},    // 7000418E
{{0x0F,0x12},{0xF9,0xFF}},    // 70004190
{{0x0F,0x12},{0x98,0x01}},    // 70004192
{{0x0F,0x12},{0x00,0x29}},    // 70004194
{{0x0F,0x12},{0xF0,0x00}},    // 70004196
{{0x0F,0x12},{0xF9,0xFB}},    // 70004198
{{0x0F,0x12},{0x98,0x01}},    // 7000419A
{{0x0F,0x12},{0x00,0x29}},    // 7000419C
{{0x0F,0x12},{0xF0,0x00}},    // 7000419E
{{0x0F,0x12},{0xF9,0xF7}},    // 700041A0
{{0x0F,0x12},{0xE0,0x03}},    // 700041A2
{{0x0F,0x12},{0x98,0x01}},    // 700041A4
{{0x0F,0x12},{0x00,0x29}},    // 700041A6
{{0x0F,0x12},{0xF0,0x00}},    // 700041A8
{{0x0F,0x12},{0xF9,0xF2}},    // 700041AA
{{0x0F,0x12},{0x98,0x01}},    // 700041AC
{{0x0F,0x12},{0x00,0x32}},    // 700041AE
{{0x0F,0x12},{0x00,0x39}},    // 700041B0
{{0x0F,0x12},{0xF0,0x00}},    // 700041B2
{{0x0F,0x12},{0xF9,0xF5}},    // 700041B4
{{0x0F,0x12},{0x60,0x20}},    // 700041B6
{{0x0F,0x12},{0xE5,0xD0}},    // 700041B8
{{0x0F,0x12},{0xB5,0x7C}},    // 700041BA
{{0x0F,0x12},{0x48,0x9A}},    // 700041BC
{{0x0F,0x12},{0xA9,0x01}},    // 700041BE
{{0x0F,0x12},{0x00,0x04}},    // 700041C0
{{0x0F,0x12},{0xF0,0x00}},    // 700041C2
{{0x0F,0x12},{0xF9,0x79}},    // 700041C4
{{0x0F,0x12},{0x46,0x6B}},    // 700041C6
{{0x0F,0x12},{0x88,0xD9}},    // 700041C8
{{0x0F,0x12},{0x88,0x98}},    // 700041CA
{{0x0F,0x12},{0x4B,0x95}},    // 700041CC
{{0x0F,0x12},{0x33,0x46}},    // 700041CE
{{0x0F,0x12},{0x1E,0x9A}},    // 700041D0
{{0x0F,0x12},{0xF0,0x00}},    // 700041D2
{{0x0F,0x12},{0xF9,0xED}},    // 700041D4
{{0x0F,0x12},{0x48,0x94}},    // 700041D6
{{0x0F,0x12},{0x49,0x92}},    // 700041D8
{{0x0F,0x12},{0x38,0x12}},    // 700041DA
{{0x0F,0x12},{0x31,0x40}},    // 700041DC
{{0x0F,0x12},{0x8A,0x42}},    // 700041DE
{{0x0F,0x12},{0x88,0x8B}},    // 700041E0
{{0x0F,0x12},{0x18,0xD2}},    // 700041E2
{{0x0F,0x12},{0x82,0x42}},    // 700041E4
{{0x0F,0x12},{0x8A,0xC2}},    // 700041E6
{{0x0F,0x12},{0x88,0xC9}},    // 700041E8
{{0x0F,0x12},{0x18,0x51}},    // 700041EA
{{0x0F,0x12},{0x82,0xC1}},    // 700041EC
{{0x0F,0x12},{0x00,0x20}},    // 700041EE
{{0x0F,0x12},{0x46,0x69}},    // 700041F0
{{0x0F,0x12},{0xF0,0x00}},    // 700041F2
{{0x0F,0x12},{0xF9,0x61}},    // 700041F4
{{0x0F,0x12},{0x48,0x8D}},    // 700041F6
{{0x0F,0x12},{0x21,0x4D}},    // 700041F8
{{0x0F,0x12},{0x83,0x01}},    // 700041FA
{{0x0F,0x12},{0x21,0x96}},    // 700041FC
{{0x0F,0x12},{0x83,0x81}},    // 700041FE
{{0x0F,0x12},{0x21,0x1D}},    // 70004200
{{0x0F,0x12},{0x30,0x20}},    // 70004202
{{0x0F,0x12},{0x80,0x01}},    // 70004204
{{0x0F,0x12},{0xF0,0x00}},    // 70004206
{{0x0F,0x12},{0xF9,0xDB}},    // 70004208
{{0x0F,0x12},{0xF0,0x00}},    // 7000420A
{{0x0F,0x12},{0xF9,0xE1}},    // 7000420C
{{0x0F,0x12},{0x48,0x88}},    // 7000420E
{{0x0F,0x12},{0x4C,0x88}},    // 70004210
{{0x0F,0x12},{0x6E,0x00}},    // 70004212
{{0x0F,0x12},{0x60,0xE0}},    // 70004214
{{0x0F,0x12},{0x46,0x6B}},    // 70004216
{{0x0F,0x12},{0x88,0x18}},    // 70004218
{{0x0F,0x12},{0x88,0x59}},    // 7000421A
{{0x0F,0x12},{0x00,0x25}},    // 7000421C
{{0x0F,0x12},{0x1A,0x40}},    // 7000421E
{{0x0F,0x12},{0x35,0x40}},    // 70004220
{{0x0F,0x12},{0x61,0xA8}},    // 70004222
{{0x0F,0x12},{0x48,0x7F}},    // 70004224
{{0x0F,0x12},{0x99,0x00}},    // 70004226
{{0x0F,0x12},{0x30,0x60}},    // 70004228
{{0x0F,0x12},{0xF0,0x00}},    // 7000422A
{{0x0F,0x12},{0xF9,0xD9}},    // 7000422C
{{0x0F,0x12},{0x46,0x6B}},    // 7000422E
{{0x0F,0x12},{0x88,0x19}},    // 70004230
{{0x0F,0x12},{0x1D,0xE0}},    // 70004232
{{0x0F,0x12},{0x30,0xF9}},    // 70004234
{{0x0F,0x12},{0x87,0x41}},    // 70004236
{{0x0F,0x12},{0x88,0x59}},    // 70004238
{{0x0F,0x12},{0x87,0x81}},    // 7000423A
{{0x0F,0x12},{0x20,0x00}},    // 7000423C
{{0x0F,0x12},{0x71,0xA0}},    // 7000423E
{{0x0F,0x12},{0x74,0xA8}},    // 70004240
{{0x0F,0x12},{0xBC,0x7C}},    // 70004242
{{0x0F,0x12},{0xBC,0x08}},    // 70004244
{{0x0F,0x12},{0x47,0x18}},    // 70004246
{{0x0F,0x12},{0xB5,0xF8}},    // 70004248
{{0x0F,0x12},{0x00,0x05}},    // 7000424A
{{0x0F,0x12},{0x68,0x08}},    // 7000424C
{{0x0F,0x12},{0x04,0x00}},    // 7000424E
{{0x0F,0x12},{0x0C,0x00}},    // 70004250
{{0x0F,0x12},{0x68,0x4A}},    // 70004252
{{0x0F,0x12},{0x04,0x12}},    // 70004254
{{0x0F,0x12},{0x0C,0x12}},    // 70004256
{{0x0F,0x12},{0x68,0x8E}},    // 70004258
{{0x0F,0x12},{0x68,0xCC}},    // 7000425A
{{0x0F,0x12},{0x49,0x70}},    // 7000425C
{{0x0F,0x12},{0x88,0x4B}},    // 7000425E
{{0x0F,0x12},{0x43,0x43}},    // 70004260
{{0x0F,0x12},{0x0A,0x98}},    // 70004262
{{0x0F,0x12},{0x23,0x04}},    // 70004264
{{0x0F,0x12},{0x5E,0xCB}},    // 70004266
{{0x0F,0x12},{0x18,0xC0}},    // 70004268
{{0x0F,0x12},{0x02,0xC0}},    // 7000426A
{{0x0F,0x12},{0x0C,0x00}},    // 7000426C
{{0x0F,0x12},{0x88,0xCB}},    // 7000426E
{{0x0F,0x12},{0x43,0x53}},    // 70004270
{{0x0F,0x12},{0x0A,0x9A}},    // 70004272
{{0x0F,0x12},{0x23,0x08}},    // 70004274
{{0x0F,0x12},{0x5E,0xCB}},    // 70004276
{{0x0F,0x12},{0x18,0xD1}},    // 70004278
{{0x0F,0x12},{0x02,0xC9}},    // 7000427A
{{0x0F,0x12},{0x0C,0x09}},    // 7000427C
{{0x0F,0x12},{0x27,0x01}},    // 7000427E
{{0x0F,0x12},{0x00,0x3A}},    // 70004280
{{0x0F,0x12},{0x40,0xAA}},    // 70004282
{{0x0F,0x12},{0x92,0x00}},    // 70004284
{{0x0F,0x12},{0x00,0x2A}},    // 70004286
{{0x0F,0x12},{0x3A,0x10}},    // 70004288
{{0x0F,0x12},{0x40,0x97}},    // 7000428A
{{0x0F,0x12},{0x2D,0x10}},    // 7000428C
{{0x0F,0x12},{0xDA,0x06}},    // 7000428E
{{0x0F,0x12},{0x4A,0x69}},    // 70004290
{{0x0F,0x12},{0x9B,0x00}},    // 70004292
{{0x0F,0x12},{0x88,0x12}},    // 70004294
{{0x0F,0x12},{0x43,0x9A}},    // 70004296
{{0x0F,0x12},{0x4B,0x67}},    // 70004298
{{0x0F,0x12},{0x80,0x1A}},    // 7000429A
{{0x0F,0x12},{0xE0,0x03}},    // 7000429C
{{0x0F,0x12},{0x4B,0x66}},    // 7000429E
{{0x0F,0x12},{0x88,0x5A}},    // 700042A0
{{0x0F,0x12},{0x43,0xBA}},    // 700042A2
{{0x0F,0x12},{0x80,0x5A}},    // 700042A4
{{0x0F,0x12},{0x00,0x23}},    // 700042A6
{{0x0F,0x12},{0x00,0x32}},    // 700042A8
{{0x0F,0x12},{0xF0,0x00}},    // 700042AA
{{0x0F,0x12},{0xF9,0x81}},    // 700042AC
{{0x0F,0x12},{0x2D,0x10}},    // 700042AE
{{0x0F,0x12},{0xDA,0x05}},    // 700042B0
{{0x0F,0x12},{0x49,0x61}},    // 700042B2
{{0x0F,0x12},{0x9A,0x00}},    // 700042B4
{{0x0F,0x12},{0x88,0x08}},    // 700042B6
{{0x0F,0x12},{0x43,0x10}},    // 700042B8
{{0x0F,0x12},{0x80,0x08}},    // 700042BA
{{0x0F,0x12},{0xE0,0x03}},    // 700042BC
{{0x0F,0x12},{0x48,0x5E}},    // 700042BE
{{0x0F,0x12},{0x88,0x41}},    // 700042C0
{{0x0F,0x12},{0x43,0x39}},    // 700042C2
{{0x0F,0x12},{0x80,0x41}},    // 700042C4
{{0x0F,0x12},{0x4D,0x5B}},    // 700042C6
{{0x0F,0x12},{0x20,0x00}},    // 700042C8
{{0x0F,0x12},{0x35,0x80}},    // 700042CA
{{0x0F,0x12},{0x88,0xAA}},    // 700042CC
{{0x0F,0x12},{0x5E,0x30}},    // 700042CE
{{0x0F,0x12},{0x21,0x00}},    // 700042D0
{{0x0F,0x12},{0xF0,0x00}},    // 700042D2
{{0x0F,0x12},{0xF9,0x8D}},    // 700042D4
{{0x0F,0x12},{0x80,0x30}},    // 700042D6
{{0x0F,0x12},{0x20,0x00}},    // 700042D8
{{0x0F,0x12},{0x88,0xAA}},    // 700042DA
{{0x0F,0x12},{0x5E,0x20}},    // 700042DC
{{0x0F,0x12},{0x21,0x00}},    // 700042DE
{{0x0F,0x12},{0xF0,0x00}},    // 700042E0
{{0x0F,0x12},{0xF9,0x86}},    // 700042E2
{{0x0F,0x12},{0x80,0x20}},    // 700042E4
{{0x0F,0x12},{0xE5,0x87}},    // 700042E6
{{0x0F,0x12},{0xB5,0x10}},    // 700042E8
{{0x0F,0x12},{0xF0,0x00}},    // 700042EA
{{0x0F,0x12},{0xF9,0x89}},    // 700042EC
{{0x0F,0x12},{0x4A,0x53}},    // 700042EE
{{0x0F,0x12},{0x8D,0x50}},    // 700042F0
{{0x0F,0x12},{0x28,0x00}},    // 700042F2
{{0x0F,0x12},{0xD0,0x07}},    // 700042F4
{{0x0F,0x12},{0x49,0x4E}},    // 700042F6
{{0x0F,0x12},{0x31,0xC0}},    // 700042F8
{{0x0F,0x12},{0x68,0x4B}},    // 700042FA
{{0x0F,0x12},{0x49,0x50}},    // 700042FC
{{0x0F,0x12},{0x42,0x83}},    // 700042FE
{{0x0F,0x12},{0xD2,0x02}},    // 70004300
{{0x0F,0x12},{0x8D,0x90}},    // 70004302
{{0x0F,0x12},{0x81,0xC8}},    // 70004304
{{0x0F,0x12},{0xE6,0xA0}},    // 70004306
{{0x0F,0x12},{0x8D,0xD0}},    // 70004308
{{0x0F,0x12},{0x81,0xC8}},    // 7000430A
{{0x0F,0x12},{0xE6,0x9D}},    // 7000430C
{{0x0F,0x12},{0xB5,0xF8}},    // 7000430E
{{0x0F,0x12},{0xF0,0x00}},    // 70004310
{{0x0F,0x12},{0xF9,0x7E}},    // 70004312
{{0x0F,0x12},{0x4D,0x49}},    // 70004314
{{0x0F,0x12},{0x8E,0x28}},    // 70004316
{{0x0F,0x12},{0x28,0x00}},    // 70004318
{{0x0F,0x12},{0xD0,0x1F}},    // 7000431A
{{0x0F,0x12},{0x4E,0x49}},    // 7000431C
{{0x0F,0x12},{0x48,0x44}},    // 7000431E
{{0x0F,0x12},{0x68,0xB4}},    // 70004320
{{0x0F,0x12},{0x68,0x00}},    // 70004322
{{0x0F,0x12},{0x42,0x84}},    // 70004324
{{0x0F,0x12},{0xD9,0x03}},    // 70004326
{{0x0F,0x12},{0x1A,0x21}},    // 70004328
{{0x0F,0x12},{0x08,0x49}},    // 7000432A
{{0x0F,0x12},{0x18,0x47}},    // 7000432C
{{0x0F,0x12},{0xE0,0x06}},    // 7000432E
{{0x0F,0x12},{0x42,0x84}},    // 70004330
{{0x0F,0x12},{0xD2,0x03}},    // 70004332
{{0x0F,0x12},{0x1B,0x01}},    // 70004334
{{0x0F,0x12},{0x08,0x49}},    // 70004336
{{0x0F,0x12},{0x1A,0x47}},    // 70004338
{{0x0F,0x12},{0xE0,0x00}},    // 7000433A
{{0x0F,0x12},{0x00,0x27}},    // 7000433C
{{0x0F,0x12},{0x00,0x20}},    // 7000433E
{{0x0F,0x12},{0x49,0x3B}},    // 70004340
{{0x0F,0x12},{0x31,0x20}},    // 70004342
{{0x0F,0x12},{0x7A,0x0C}},    // 70004344
{{0x0F,0x12},{0x2C,0x00}},    // 70004346
{{0x0F,0x12},{0xD0,0x04}},    // 70004348
{{0x0F,0x12},{0x02,0x00}},    // 7000434A
{{0x0F,0x12},{0x00,0x39}},    // 7000434C
{{0x0F,0x12},{0xF0,0x00}},    // 7000434E
{{0x0F,0x12},{0xF8,0xC3}},    // 70004350
{{0x0F,0x12},{0x86,0x68}},    // 70004352
{{0x0F,0x12},{0x2C,0x00}},    // 70004354
{{0x0F,0x12},{0xD0,0x00}},    // 70004356
{{0x0F,0x12},{0x60,0xB7}},    // 70004358
{{0x0F,0x12},{0xE5,0x4D}},    // 7000435A
{{0x0F,0x12},{0x20,0xFF}},    // 7000435C
{{0x0F,0x12},{0x1C,0x40}},    // 7000435E
{{0x0F,0x12},{0x86,0x68}},    // 70004360
{{0x0F,0x12},{0xE5,0x49}},    // 70004362
{{0x0F,0x12},{0xB5,0x10}},    // 70004364
{{0x0F,0x12},{0x00,0x0C}},    // 70004366
{{0x0F,0x12},{0x68,0x20}},    // 70004368
{{0x0F,0x12},{0x04,0x00}},    // 7000436A
{{0x0F,0x12},{0x0C,0x00}},    // 7000436C
{{0x0F,0x12},{0x49,0x33}},    // 7000436E
{{0x0F,0x12},{0x8E,0x0A}},    // 70004370
{{0x0F,0x12},{0x2A,0x00}},    // 70004372
{{0x0F,0x12},{0xD0,0x03}},    // 70004374
{{0x0F,0x12},{0x8E,0x49}},    // 70004376
{{0x0F,0x12},{0x02,0x00}},    // 70004378
{{0x0F,0x12},{0xF0,0x00}},    // 7000437A
{{0x0F,0x12},{0xF8,0xAD}},    // 7000437C
{{0x0F,0x12},{0x60,0x20}},    // 7000437E
{{0x0F,0x12},{0x04,0x00}},    // 70004380
{{0x0F,0x12},{0x0C,0x00}},    // 70004382
{{0x0F,0x12},{0xE6,0x61}},    // 70004384
{{0x0F,0x12},{0xB5,0x70}},    // 70004386
{{0x0F,0x12},{0x68,0x0C}},    // 70004388
{{0x0F,0x12},{0x4D,0x2F}},    // 7000438A
{{0x0F,0x12},{0x00,0x20}},    // 7000438C
{{0x0F,0x12},{0x6F,0x29}},    // 7000438E
{{0x0F,0x12},{0xF0,0x00}},    // 70004390
{{0x0F,0x12},{0xF9,0x46}},    // 70004392
{{0x0F,0x12},{0x6F,0x69}},    // 70004394
{{0x0F,0x12},{0x1D,0x20}},    // 70004396
{{0x0F,0x12},{0xF0,0x00}},    // 70004398
{{0x0F,0x12},{0xF9,0x42}},    // 7000439A
{{0x0F,0x12},{0x48,0x27}},    // 7000439C
{{0x0F,0x12},{0x8E,0x00}},    // 7000439E
{{0x0F,0x12},{0x28,0x00}},    // 700043A0
{{0x0F,0x12},{0xD0,0x06}},    // 700043A2
{{0x0F,0x12},{0x49,0x22}},    // 700043A4
{{0x0F,0x12},{0x22,0x14}},    // 700043A6
{{0x0F,0x12},{0x31,0x68}},    // 700043A8
{{0x0F,0x12},{0x00,0x08}},    // 700043AA
{{0x0F,0x12},{0x38,0x3C}},    // 700043AC
{{0x0F,0x12},{0xF0,0x00}},    // 700043AE
{{0x0F,0x12},{0xF9,0x3F}},    // 700043B0
{{0x0F,0x12},{0xE4,0x88}},    // 700043B2
{{0x0F,0x12},{0xB5,0xF8}},    // 700043B4
{{0x0F,0x12},{0x00,0x04}},    // 700043B6
{{0x0F,0x12},{0x4D,0x24}},    // 700043B8
{{0x0F,0x12},{0x8B,0x68}},    // 700043BA
{{0x0F,0x12},{0x28,0x00}},    // 700043BC
{{0x0F,0x12},{0xD0,0x12}},    // 700043BE
{{0x0F,0x12},{0x48,0x23}},    // 700043C0
{{0x0F,0x12},{0x8A,0x00}},    // 700043C2
{{0x0F,0x12},{0x06,0xC0}},    // 700043C4
{{0x0F,0x12},{0xD5,0x0E}},    // 700043C6
{{0x0F,0x12},{0x48,0x22}},    // 700043C8
{{0x0F,0x12},{0x78,0x00}},    // 700043CA
{{0x0F,0x12},{0x28,0x00}},    // 700043CC
{{0x0F,0x12},{0xD0,0x0A}},    // 700043CE
{{0x0F,0x12},{0x48,0x1D}},    // 700043D0
{{0x0F,0x12},{0x6F,0xC1}},    // 700043D2
{{0x0F,0x12},{0x20,0x00}},    // 700043D4
{{0x0F,0x12},{0xF0,0x00}},    // 700043D6
{{0x0F,0x12},{0xF9,0x23}},    // 700043D8
{{0x0F,0x12},{0x8B,0x28}},    // 700043DA
{{0x0F,0x12},{0x22,0x01}},    // 700043DC
{{0x0F,0x12},{0x21,0x80}},    // 700043DE
{{0x0F,0x12},{0xF0,0x00}},    // 700043E0
{{0x0F,0x12},{0xF9,0x2C}},    // 700043E2
{{0x0F,0x12},{0x83,0x28}},    // 700043E4
{{0x0F,0x12},{0x21,0x01}},    // 700043E6
{{0x0F,0x12},{0x00,0x0D}},    // 700043E8
{{0x0F,0x12},{0x00,0x20}},    // 700043EA
{{0x0F,0x12},{0x38,0x10}},    // 700043EC
{{0x0F,0x12},{0x40,0x81}},    // 700043EE
{{0x0F,0x12},{0x40,0xA5}},    // 700043F0
{{0x0F,0x12},{0x4F,0x11}},    // 700043F2
{{0x0F,0x12},{0x00,0x0E}},    // 700043F4
{{0x0F,0x12},{0x2C,0x10}},    // 700043F6
{{0x0F,0x12},{0xDA,0x03}},    // 700043F8
{{0x0F,0x12},{0x88,0x38}},    // 700043FA
{{0x0F,0x12},{0x43,0xA8}},    // 700043FC
{{0x0F,0x12},{0x80,0x38}},    // 700043FE
{{0x0F,0x12},{0xE0,0x02}},    // 70004400
{{0x0F,0x12},{0x88,0x78}},    // 70004402
{{0x0F,0x12},{0x43,0xB0}},    // 70004404
{{0x0F,0x12},{0x80,0x78}},    // 70004406
{{0x0F,0x12},{0xF0,0x00}},    // 70004408
{{0x0F,0x12},{0xF9,0x20}},    // 7000440A
{{0x0F,0x12},{0x2C,0x10}},    // 7000440C
{{0x0F,0x12},{0xDA,0x03}},    // 7000440E
{{0x0F,0x12},{0x88,0x38}},    // 70004410
{{0x0F,0x12},{0x43,0x28}},    // 70004412
{{0x0F,0x12},{0x80,0x38}},    // 70004414
{{0x0F,0x12},{0xE4,0xEF}},    // 70004416
{{0x0F,0x12},{0x88,0x78}},    // 70004418
{{0x0F,0x12},{0x43,0x30}},    // 7000441A
{{0x0F,0x12},{0x80,0x78}},    // 7000441C
{{0x0F,0x12},{0xE4,0xEB}},    // 7000441E
{{0x0F,0x12},{0x25,0x58}},    // 70004420
{{0x0F,0x12},{0x70,0x00}},    // 70004422
{{0x0F,0x12},{0x2A,0xB8}},    // 70004424
{{0x0F,0x12},{0x70,0x00}},    // 70004426
{{0x0F,0x12},{0x14,0x5E}},    // 70004428
{{0x0F,0x12},{0x70,0x00}},    // 7000442A
{{0x0F,0x12},{0x26,0x98}},    // 7000442C
{{0x0F,0x12},{0x70,0x00}},    // 7000442E
{{0x0F,0x12},{0x2B,0xB8}},    // 70004430
{{0x0F,0x12},{0x70,0x00}},    // 70004432
{{0x0F,0x12},{0x29,0x98}},    // 70004434
{{0x0F,0x12},{0x70,0x00}},    // 70004436
{{0x0F,0x12},{0x11,0x00}},    // 70004438
{{0x0F,0x12},{0xD0,0x00}},    // 7000443A
{{0x0F,0x12},{0x47,0x80}},    // 7000443C
{{0x0F,0x12},{0x70,0x00}},    // 7000443E
{{0x0F,0x12},{0xE2,0x00}},    // 70004440
{{0x0F,0x12},{0xD0,0x00}},    // 70004442
{{0x0F,0x12},{0x21,0x0C}},    // 70004444
{{0x0F,0x12},{0x70,0x00}},    // 70004446
{{0x0F,0x12},{0x00,0x00}},    // 70004448
{{0x0F,0x12},{0x70,0x00}},    // 7000444A
{{0x0F,0x12},{0x30,0x8C}},    // 7000444C
{{0x0F,0x12},{0x70,0x00}},    // 7000444E
{{0x0F,0x12},{0xB0,0x40}},    // 70004450
{{0x0F,0x12},{0xD0,0x00}},    // 70004452
{{0x0F,0x12},{0x38,0x58}},    // 70004454
{{0x0F,0x12},{0x70,0x00}},    // 70004456
{{0x0F,0x12},{0x47,0x78}},    // 70004458
{{0x0F,0x12},{0x46,0xC0}},    // 7000445A
{{0x0F,0x12},{0xC0,0x00}},    // 7000445C
{{0x0F,0x12},{0xE5,0x9F}},    // 7000445E
{{0x0F,0x12},{0xFF,0x1C}},    // 70004460
{{0x0F,0x12},{0xE1,0x2F}},    // 70004462
{{0x0F,0x12},{0x17,0x89}},    // 70004464
{{0x0F,0x12},{0x00,0x01}},    // 70004466
{{0x0F,0x12},{0x47,0x78}},    // 70004468
{{0x0F,0x12},{0x46,0xC0}},    // 7000446A
{{0x0F,0x12},{0xC0,0x00}},    // 7000446C
{{0x0F,0x12},{0xE5,0x9F}},    // 7000446E
{{0x0F,0x12},{0xFF,0x1C}},    // 70004470
{{0x0F,0x12},{0xE1,0x2F}},    // 70004472
{{0x0F,0x12},{0x16,0xF1}},    // 70004474
{{0x0F,0x12},{0x00,0x01}},    // 70004476
{{0x0F,0x12},{0x47,0x78}},    // 70004478
{{0x0F,0x12},{0x46,0xC0}},    // 7000447A
{{0x0F,0x12},{0xC0,0x00}},    // 7000447C
{{0x0F,0x12},{0xE5,0x9F}},    // 7000447E
{{0x0F,0x12},{0xFF,0x1C}},    // 70004480
{{0x0F,0x12},{0xE1,0x2F}},    // 70004482
{{0x0F,0x12},{0xC3,0xB1}},    // 70004484
{{0x0F,0x12},{0x00,0x00}},    // 70004486
{{0x0F,0x12},{0x47,0x78}},    // 70004488
{{0x0F,0x12},{0x46,0xC0}},    // 7000448A
{{0x0F,0x12},{0xC0,0x00}},    // 7000448C
{{0x0F,0x12},{0xE5,0x9F}},    // 7000448E
{{0x0F,0x12},{0xFF,0x1C}},    // 70004490
{{0x0F,0x12},{0xE1,0x2F}},    // 70004492
{{0x0F,0x12},{0xC3,0x6D}},    // 70004494
{{0x0F,0x12},{0x00,0x00}},    // 70004496
{{0x0F,0x12},{0x47,0x78}},    // 70004498
{{0x0F,0x12},{0x46,0xC0}},    // 7000449A
{{0x0F,0x12},{0xC0,0x00}},    // 7000449C
{{0x0F,0x12},{0xE5,0x9F}},    // 7000449E
{{0x0F,0x12},{0xFF,0x1C}},    // 700044A0
{{0x0F,0x12},{0xE1,0x2F}},    // 700044A2
{{0x0F,0x12},{0xF6,0xD7}},    // 700044A4
{{0x0F,0x12},{0x00,0x00}},    // 700044A6
{{0x0F,0x12},{0x47,0x78}},    // 700044A8
{{0x0F,0x12},{0x46,0xC0}},    // 700044AA
{{0x0F,0x12},{0xC0,0x00}},    // 700044AC
{{0x0F,0x12},{0xE5,0x9F}},    // 700044AE
{{0x0F,0x12},{0xFF,0x1C}},    // 700044B0
{{0x0F,0x12},{0xE1,0x2F}},    // 700044B2
{{0x0F,0x12},{0xB4,0x9D}},    // 700044B4
{{0x0F,0x12},{0x00,0x00}},    // 700044B6
{{0x0F,0x12},{0x47,0x78}},    // 700044B8
{{0x0F,0x12},{0x46,0xC0}},    // 700044BA
{{0x0F,0x12},{0xC0,0x00}},    // 700044BC
{{0x0F,0x12},{0xE5,0x9F}},    // 700044BE
{{0x0F,0x12},{0xFF,0x1C}},    // 700044C0
{{0x0F,0x12},{0xE1,0x2F}},    // 700044C2
{{0x0F,0x12},{0x7E,0xDF}},    // 700044C4
{{0x0F,0x12},{0x00,0x00}},    // 700044C6
{{0x0F,0x12},{0x47,0x78}},    // 700044C8
{{0x0F,0x12},{0x46,0xC0}},    // 700044CA
{{0x0F,0x12},{0xC0,0x00}},    // 700044CC
{{0x0F,0x12},{0xE5,0x9F}},    // 700044CE
{{0x0F,0x12},{0xFF,0x1C}},    // 700044D0
{{0x0F,0x12},{0xE1,0x2F}},    // 700044D2
{{0x0F,0x12},{0x44,0x8D}},    // 700044D4
{{0x0F,0x12},{0x00,0x00}},    // 700044D6
{{0x0F,0x12},{0x47,0x78}},    // 700044D8
{{0x0F,0x12},{0x46,0xC0}},    // 700044DA
{{0x0F,0x12},{0xF0,0x04}},    // 700044DC
{{0x0F,0x12},{0xE5,0x1F}},    // 700044DE
{{0x0F,0x12},{0x29,0xEC}},    // 700044E0
{{0x0F,0x12},{0x00,0x01}},    // 700044E2
{{0x0F,0x12},{0x47,0x78}},    // 700044E4
{{0x0F,0x12},{0x46,0xC0}},    // 700044E6
{{0x0F,0x12},{0xC0,0x00}},    // 700044E8
{{0x0F,0x12},{0xE5,0x9F}},    // 700044EA
{{0x0F,0x12},{0xFF,0x1C}},    // 700044EC
{{0x0F,0x12},{0xE1,0x2F}},    // 700044EE
{{0x0F,0x12},{0x2E,0xF1}},    // 700044F0
{{0x0F,0x12},{0x00,0x00}},    // 700044F2
{{0x0F,0x12},{0x47,0x78}},    // 700044F4
{{0x0F,0x12},{0x46,0xC0}},    // 700044F6
{{0x0F,0x12},{0xC0,0x00}},    // 700044F8
{{0x0F,0x12},{0xE5,0x9F}},    // 700044FA
{{0x0F,0x12},{0xFF,0x1C}},    // 700044FC
{{0x0F,0x12},{0xE1,0x2F}},    // 700044FE
{{0x0F,0x12},{0xEE,0x03}},    // 70004500
{{0x0F,0x12},{0x00,0x00}},    // 70004502
{{0x0F,0x12},{0x47,0x78}},    // 70004504
{{0x0F,0x12},{0x46,0xC0}},    // 70004506
{{0x0F,0x12},{0xC0,0x00}},    // 70004508
{{0x0F,0x12},{0xE5,0x9F}},    // 7000450A
{{0x0F,0x12},{0xFF,0x1C}},    // 7000450C
{{0x0F,0x12},{0xE1,0x2F}},    // 7000450E
{{0x0F,0x12},{0xA5,0x8B}},    // 70004510
{{0x0F,0x12},{0x00,0x00}},    // 70004512
{{0x0F,0x12},{0x47,0x78}},    // 70004514
{{0x0F,0x12},{0x46,0xC0}},    // 70004516
{{0x0F,0x12},{0xC0,0x00}},    // 70004518
{{0x0F,0x12},{0xE5,0x9F}},    // 7000451A
{{0x0F,0x12},{0xFF,0x1C}},    // 7000451C
{{0x0F,0x12},{0xE1,0x2F}},    // 7000451E
{{0x0F,0x12},{0x7C,0x49}},    // 70004520
{{0x0F,0x12},{0x00,0x00}},    // 70004522
{{0x0F,0x12},{0x47,0x78}},    // 70004524
{{0x0F,0x12},{0x46,0xC0}},    // 70004526
{{0x0F,0x12},{0xC0,0x00}},    // 70004528
{{0x0F,0x12},{0xE5,0x9F}},    // 7000452A
{{0x0F,0x12},{0xFF,0x1C}},    // 7000452C
{{0x0F,0x12},{0xE1,0x2F}},    // 7000452E
{{0x0F,0x12},{0x7C,0x63}},    // 70004530
{{0x0F,0x12},{0x00,0x00}},    // 70004532
{{0x0F,0x12},{0x47,0x78}},    // 70004534
{{0x0F,0x12},{0x46,0xC0}},    // 70004536
{{0x0F,0x12},{0xC0,0x00}},    // 70004538
{{0x0F,0x12},{0xE5,0x9F}},    // 7000453A
{{0x0F,0x12},{0xFF,0x1C}},    // 7000453C
{{0x0F,0x12},{0xE1,0x2F}},    // 7000453E
{{0x0F,0x12},{0x2D,0xB7}},    // 70004540
{{0x0F,0x12},{0x00,0x00}},    // 70004542
{{0x0F,0x12},{0x47,0x78}},    // 70004544
{{0x0F,0x12},{0x46,0xC0}},    // 70004546
{{0x0F,0x12},{0xC0,0x00}},    // 70004548
{{0x0F,0x12},{0xE5,0x9F}},    // 7000454A
{{0x0F,0x12},{0xFF,0x1C}},    // 7000454C
{{0x0F,0x12},{0xE1,0x2F}},    // 7000454E
{{0x0F,0x12},{0xEB,0x3D}},    // 70004550
{{0x0F,0x12},{0x00,0x00}},    // 70004552
{{0x0F,0x12},{0x47,0x78}},    // 70004554
{{0x0F,0x12},{0x46,0xC0}},    // 70004556
{{0x0F,0x12},{0xC0,0x00}},    // 70004558
{{0x0F,0x12},{0xE5,0x9F}},    // 7000455A
{{0x0F,0x12},{0xFF,0x1C}},    // 7000455C
{{0x0F,0x12},{0xE1,0x2F}},    // 7000455E
{{0x0F,0x12},{0xF0,0x61}},    // 70004560
{{0x0F,0x12},{0x00,0x00}},    // 70004562
{{0x0F,0x12},{0x47,0x78}},    // 70004564
{{0x0F,0x12},{0x46,0xC0}},    // 70004566
{{0x0F,0x12},{0xC0,0x00}},    // 70004568
{{0x0F,0x12},{0xE5,0x9F}},    // 7000456A
{{0x0F,0x12},{0xFF,0x1C}},    // 7000456C
{{0x0F,0x12},{0xE1,0x2F}},    // 7000456E
{{0x0F,0x12},{0xF0,0xEF}},    // 70004570
{{0x0F,0x12},{0x00,0x00}},    // 70004572
{{0x0F,0x12},{0x47,0x78}},    // 70004574
{{0x0F,0x12},{0x46,0xC0}},    // 70004576
{{0x0F,0x12},{0xF0,0x04}},    // 70004578
{{0x0F,0x12},{0xE5,0x1F}},    // 7000457A
{{0x0F,0x12},{0x28,0x24}},    // 7000457C
{{0x0F,0x12},{0x00,0x01}},    // 7000457E
{{0x0F,0x12},{0x47,0x78}},    // 70004580
{{0x0F,0x12},{0x46,0xC0}},    // 70004582
{{0x0F,0x12},{0xC0,0x00}},    // 70004584
{{0x0F,0x12},{0xE5,0x9F}},    // 70004586
{{0x0F,0x12},{0xFF,0x1C}},    // 70004588
{{0x0F,0x12},{0xE1,0x2F}},    // 7000458A
{{0x0F,0x12},{0x8E,0xDD}},    // 7000458C
{{0x0F,0x12},{0x00,0x00}},    // 7000458E
{{0x0F,0x12},{0x47,0x78}},    // 70004590
{{0x0F,0x12},{0x46,0xC0}},    // 70004592
{{0x0F,0x12},{0xC0,0x00}},    // 70004594
{{0x0F,0x12},{0xE5,0x9F}},    // 70004596
{{0x0F,0x12},{0xFF,0x1C}},    // 70004598
{{0x0F,0x12},{0xE1,0x2F}},    // 7000459A
{{0x0F,0x12},{0x8D,0xCB}},    // 7000459C
{{0x0F,0x12},{0x00,0x00}},    // 7000459E
{{0x0F,0x12},{0x47,0x78}},    // 700045A0
{{0x0F,0x12},{0x46,0xC0}},    // 700045A2
{{0x0F,0x12},{0xC0,0x00}},    // 700045A4
{{0x0F,0x12},{0xE5,0x9F}},    // 700045A6
{{0x0F,0x12},{0xFF,0x1C}},    // 700045A8
{{0x0F,0x12},{0xE1,0x2F}},    // 700045AA
{{0x0F,0x12},{0x8E,0x17}},    // 700045AC
{{0x0F,0x12},{0x00,0x00}},    // 700045AE
{{0x0F,0x12},{0x47,0x78}},    // 700045B0
{{0x0F,0x12},{0x46,0xC0}},    // 700045B2
{{0x0F,0x12},{0xC0,0x00}},    // 700045B4
{{0x0F,0x12},{0xE5,0x9F}},    // 700045B6
{{0x0F,0x12},{0xFF,0x1C}},    // 700045B8
{{0x0F,0x12},{0xE1,0x2F}},    // 700045BA
{{0x0F,0x12},{0x98,0xC5}},    // 700045BC
{{0x0F,0x12},{0x00,0x00}},    // 700045BE
{{0x0F,0x12},{0x47,0x78}},    // 700045C0
{{0x0F,0x12},{0x46,0xC0}},    // 700045C2
{{0x0F,0x12},{0xC0,0x00}},    // 700045C4
{{0x0F,0x12},{0xE5,0x9F}},    // 700045C6
{{0x0F,0x12},{0xFF,0x1C}},    // 700045C8
{{0x0F,0x12},{0xE1,0x2F}},    // 700045CA
{{0x0F,0x12},{0x7C,0x7D}},    // 700045CC
{{0x0F,0x12},{0x00,0x00}},    // 700045CE
{{0x0F,0x12},{0x47,0x78}},    // 700045D0
{{0x0F,0x12},{0x46,0xC0}},    // 700045D2
{{0x0F,0x12},{0xC0,0x00}},    // 700045D4
{{0x0F,0x12},{0xE5,0x9F}},    // 700045D6
{{0x0F,0x12},{0xFF,0x1C}},    // 700045D8
{{0x0F,0x12},{0xE1,0x2F}},    // 700045DA
{{0x0F,0x12},{0x7E,0x31}},    // 700045DC
{{0x0F,0x12},{0x00,0x00}},    // 700045DE
{{0x0F,0x12},{0x47,0x78}},    // 700045E0
{{0x0F,0x12},{0x46,0xC0}},    // 700045E2
{{0x0F,0x12},{0xC0,0x00}},    // 700045E4
{{0x0F,0x12},{0xE5,0x9F}},    // 700045E6
{{0x0F,0x12},{0xFF,0x1C}},    // 700045E8
{{0x0F,0x12},{0xE1,0x2F}},    // 700045EA
{{0x0F,0x12},{0x7E,0xAB}},    // 700045EC
{{0x0F,0x12},{0x00,0x00}},    // 700045EE
{{0x0F,0x12},{0x47,0x78}},    // 700045F0
{{0x0F,0x12},{0x46,0xC0}},    // 700045F2
{{0x0F,0x12},{0xC0,0x00}},    // 700045F4
{{0x0F,0x12},{0xE5,0x9F}},    // 700045F6
{{0x0F,0x12},{0xFF,0x1C}},    // 700045F8
{{0x0F,0x12},{0xE1,0x2F}},    // 700045FA
{{0x0F,0x12},{0x75,0x01}},    // 700045FC
{{0x0F,0x12},{0x00,0x00}},    // 700045FE
{{0x0F,0x12},{0x47,0x78}},    // 70004600
{{0x0F,0x12},{0x46,0xC0}},    // 70004602
{{0x0F,0x12},{0xC0,0x00}},    // 70004604
{{0x0F,0x12},{0xE5,0x9F}},    // 70004606
{{0x0F,0x12},{0xFF,0x1C}},    // 70004608
{{0x0F,0x12},{0xE1,0x2F}},    // 7000460A
{{0x0F,0x12},{0xF6,0x3F}},    // 7000460C
{{0x0F,0x12},{0x00,0x00}},    // 7000460E
{{0x0F,0x12},{0x47,0x78}},    // 70004610
{{0x0F,0x12},{0x46,0xC0}},    // 70004612
{{0x0F,0x12},{0xC0,0x00}},    // 70004614
{{0x0F,0x12},{0xE5,0x9F}},    // 70004616
{{0x0F,0x12},{0xFF,0x1C}},    // 70004618
{{0x0F,0x12},{0xE1,0x2F}},    // 7000461A
{{0x0F,0x12},{0x3D,0x0B}},    // 7000461C
{{0x0F,0x12},{0x00,0x00}},    // 7000461E
{{0x0F,0x12},{0x47,0x78}},    // 70004620
{{0x0F,0x12},{0x46,0xC0}},    // 70004622
{{0x0F,0x12},{0xC0,0x00}},    // 70004624
{{0x0F,0x12},{0xE5,0x9F}},    // 70004626
{{0x0F,0x12},{0xFF,0x1C}},    // 70004628
{{0x0F,0x12},{0xE1,0x2F}},    // 7000462A
{{0x0F,0x12},{0x29,0xBF}},    // 7000462C
{{0x0F,0x12},{0x00,0x01}},    // 7000462E
{{0x0F,0x12},{0x47,0x78}},    // 70004630
{{0x0F,0x12},{0x46,0xC0}},    // 70004632
{{0x0F,0x12},{0xF0,0x04}},    // 70004634
{{0x0F,0x12},{0xE5,0x1F}},    // 70004636
{{0x0F,0x12},{0x26,0xD8}},    // 70004638
{{0x0F,0x12},{0x00,0x01}},    // 7000463A
{{0x0F,0x12},{0x47,0x78}},    // 7000463C
{{0x0F,0x12},{0x46,0xC0}},    // 7000463E
{{0x0F,0x12},{0xC0,0x00}},    // 70004640
{{0x0F,0x12},{0xE5,0x9F}},    // 70004642
{{0x0F,0x12},{0xFF,0x1C}},    // 70004644
{{0x0F,0x12},{0xE1,0x2F}},    // 70004646
{{0x0F,0x12},{0x30,0x6B}},    // 70004648
{{0x0F,0x12},{0x00,0x00}},    // 7000464A
{{0x0F,0x12},{0x47,0x78}},    // 7000464C
{{0x0F,0x12},{0x46,0xC0}},    // 7000464E
{{0x0F,0x12},{0xC0,0x00}},    // 70004650
{{0x0F,0x12},{0xE5,0x9F}},    // 70004652
{{0x0F,0x12},{0xFF,0x1C}},    // 70004654
{{0x0F,0x12},{0xE1,0x2F}},    // 70004656
{{0x0F,0x12},{0x60,0x99}},    // 70004658
{{0x0F,0x12},{0x00,0x00}},    // 7000465A
// End of Patch Data(Last : 7000465Ah)
// Total Size 2916 (0x0B64)
// Addr : 3AF8 , Size : 2914(B62h)
        
// TNP_USER_MBCV_CONTROL
// TNP_4EC_MBR_TUNE
// TNP_4EC_FORBIDDEN_TUNE
// TNP_AF_FINESEARCH_DRIVEBACK
// TNP_FLASH_ALG
// TNP_GAS_ALPHA_OTP
// TNP_AWB_MODUL_COMP
// TNP_AWB_INIT_QUEUE
// TNP_AWB_GRID_LOWBR
// TNP_AWB_GRID_MODULECOMP
// TNP_ADLC_TUNE
// TNP_1FRAME_AE
// TNP_TG_OFF_CFG_CHG_IN_SPOOF_MODE

//==================================================================================
// 05.OTP Control 
//==================================================================================
          
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x07,0x22}},	
{{0x0F,0x12},{0x01,0x00}},	//skl_OTP_usWaitTime This register should be positioned in fornt of D0001000	
{{0x00,0x2A},{0x07,0x26}},	
{{0x0F,0x12},{0x00,0x01}},	//skl_bUseOTPfunc This is OTP on/off function	
{{0x00,0x2A},{0x08,0xD6}},	
{{0x0F,0x12},{0x00,0x01}},	//ash_bUseOTPData	
{{0x00,0x2A},{0x14,0x6E}},	
{{0x0F,0x12},{0x00,0x00}},	//awbb_otp_disable	
{{0x00,0x2A},{0x08,0xDC}},	
{{0x0F,0x12},{0x00,0x00}},	//ash_bUseGasAlphaOTP	          

{{0x00,0x28},{0xD0,0x00}},
{{0x00,0x2A},{0x10,0x00}},
{{0x0F,0x12},{0x00,0x01}},  

//==================================================================================
//06.Gas_Anti Shading
//==================================================================================

// Refer Mon_AWB_RotGain         	
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x08,0xB4}},	
{{0x0F,0x12},{0x00,0x01}},	//wbt_bUseOutdoorASH	
{{0x00,0x2A},{0x08,0xBC}},	
{{0x0F,0x12},{0x00,0xC0}},	//TVAR_ash_AwbAshCord_0_ 2300K	
{{0x0F,0x12},{0x00,0xDF}},	//TVAR_ash_AwbAshCord_1_ 2750K	
{{0x0F,0x12},{0x01,0x00}},	//TVAR_ash_AwbAshCord_2_ 3300K	
{{0x0F,0x12},{0x01,0x25}},	//TVAR_ash_AwbAshCord_3_ 4150K	
{{0x0F,0x12},{0x01,0x5F}},	//TVAR_ash_AwbAshCord_4_ 5250K	
{{0x0F,0x12},{0x01,0x7C}},	//TVAR_ash_AwbAshCord_5_ 6400K	
{{0x0F,0x12},{0x01,0x94}},	//TVAR_ash_AwbAshCord_6_ 7500K	
	
{{0x00,0x2A},{0x08,0xF6}},	
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_0__0_ R  // 2300K     	
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_0__1_ GR          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_0__2_ GB          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_0__3_ B    

{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_1__0_ R  // 2750K 
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_1__1_ GR          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_1__2_ GB  
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_1__3_ B   

{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_2__0_ R  // 3300K 
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_2__1_ GR          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_2__2_ GB          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_2__3_ B   

{{0x0F,0x12},{0x40,0x00}}, 	//TVAR_ash_GASAlpha_3__0_ R  // 4150K 
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_3__1_ GR          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_3__2_ GB          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_3__3_ B    

{{0x0F,0x12},{0x40,0x00}}, 	//TVAR_ash_GASAlpha_4__0_ R  // 5250K 
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_4__1_ GR          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_4__2_ GB          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_4__3_ B      

{{0x0F,0x12},{0x43,0x00}}, 	//TVAR_ash_GASAlpha_5__0_ R  // 6400K 
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_5__1_ GR          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_5__2_ GB          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_5__3_ B        

{{0x0F,0x12},{0x43,0x00}},	//TVAR_ash_GASAlpha_6__0_ R  // 7500K 
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_6__1_ GR          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_6__2_ GB          
{{0x0F,0x12},{0x40,0x00}},	//TVAR_ash_GASAlpha_6__3_ B        

//Outdoor GAS Alpha   
{{0x0F,0x12},{0x45,0x00}},  	
{{0x0F,0x12},{0x40,0x00}},	
{{0x0F,0x12},{0x40,0x00}},	
{{0x0F,0x12},{0x40,0x00}},	

{{0x00,0x2A},{0x08,0xF4}},	
{{0x0F,0x12},{0x00,0x01}},	//ash_bUseGasAlpha

//==================================================================================
// 07. Analog Setting 2 
//==================================================================================
//This register is for FACTORY ONLY. 
//If you change it without prior notification
//YOU are RESPONSIBLE for the FAILURE that will happen in the future      
//For subsampling Size	

{{0x00,0x28},{0x70,0x00}},	
{{0x00,0x2A},{0x18,0xBC}},	
{{0x0F,0x12},{0x00,0x04}},	
{{0x0F,0x12},{0x05,0xB6}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x01}},	
{{0x0F,0x12},{0x05,0xBA}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x07}},	
{{0x0F,0x12},{0x05,0xBA}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x01,0xF4}},	
{{0x0F,0x12},{0x02,0x4E}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x01,0xF4}},	
{{0x0F,0x12},{0x05,0xB6}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x01,0xF4}},	
{{0x0F,0x12},{0x05,0xBA}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x01,0xF4}},	
{{0x0F,0x12},{0x02,0x4F}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x75}},	
{{0x0F,0x12},{0x00,0xCF}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x75}},	
{{0x0F,0x12},{0x00,0xD6}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x04}},	
{{0x0F,0x12},{0x01,0xF4}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0xF0}},	
{{0x0F,0x12},{0x01,0xF4}},	
{{0x0F,0x12},{0x02,0x9E}},	
{{0x0F,0x12},{0x05,0xB2}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x01,0xF8}},	
{{0x0F,0x12},{0x02,0x28}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0x08}},	
{{0x0F,0x12},{0x02,0x38}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0x18}},	
{{0x0F,0x12},{0x02,0x38}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x01}},	
{{0x0F,0x12},{0x00,0x09}},	
{{0x0F,0x12},{0x00,0xDE}},	
{{0x0F,0x12},{0x05,0xC0}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0xDF}},	
{{0x0F,0x12},{0x00,0xE4}},	
{{0x0F,0x12},{0x01,0xF8}},	
{{0x0F,0x12},{0x01,0xFD}},	
{{0x0F,0x12},{0x05,0xB6}},	
{{0x0F,0x12},{0x05,0xBB}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x01,0xF8}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x77}},	
{{0x0F,0x12},{0x00,0x7E}},	
{{0x0F,0x12},{0x02,0x4F}},	
{{0x0F,0x12},{0x02,0x5E}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
	
 // For Capture 	
	
{{0x0F,0x12},{0x00,0x04}},	
{{0x0F,0x12},{0x09,0xD1}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x01}},	
{{0x0F,0x12},{0x09,0xD5}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x08}},	
{{0x0F,0x12},{0x09,0xD5}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xAA}},	
{{0x0F,0x12},{0x03,0x26}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xAA}},	
{{0x0F,0x12},{0x09,0xD1}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xAA}},	
{{0x0F,0x12},{0x09,0xD5}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xAA}},	
{{0x0F,0x12},{0x03,0x27}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x08}},	
{{0x0F,0x12},{0x00,0x84}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x08}},	
{{0x0F,0x12},{0x00,0x8D}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x08}},	
{{0x0F,0x12},{0x02,0xAA}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0xAA}},	
{{0x0F,0x12},{0x02,0xAA}},	
{{0x0F,0x12},{0x03,0xAD}},	
{{0x0F,0x12},{0x09,0xCD}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xAE}},	
{{0x0F,0x12},{0x02,0xDE}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xBE}},	
{{0x0F,0x12},{0x02,0xEE}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xCE}},	
{{0x0F,0x12},{0x02,0xEE}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x01}},	
{{0x0F,0x12},{0x00,0x09}},	
{{0x0F,0x12},{0x00,0x95}},	
{{0x0F,0x12},{0x09,0xDB}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x96}},	
{{0x0F,0x12},{0x00,0x9B}},	
{{0x0F,0x12},{0x02,0xAE}},	
{{0x0F,0x12},{0x02,0xB3}},	
{{0x0F,0x12},{0x09,0xD1}},	
{{0x0F,0x12},{0x09,0xD6}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x02,0xAE}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x09}},	
{{0x0F,0x12},{0x00,0x10}},	
{{0x0F,0x12},{0x03,0x27}},	
{{0x0F,0x12},{0x03,0x36}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
	
{{0x00,0x2A},{0x1A,0xF8}},	
{{0x0F,0x12},{0x5A,0x3C}},	//senHal_TuneStr_AngTuneData1_2_D000F400 register at subsampling
{{0x00,0x2A},{0x18,0x96}},	
{{0x0F,0x12},{0x00,0x02}},	//senHal_SamplingType	0002 03EE: PLA setting          
{{0x0F,0x12},{0x00,0x00}},	//senHal_SamplingMode 0 : 2 PLA / 1 : 4PLA              
{{0x0F,0x12},{0x00,0x03}},	//senHal_PLAOption	[0] VPLA enable  [1] HPLA enable

{{0x00,0x2A},{0x1B,0x00}},  //Green Noise 
{{0x0F,0x12},{0xF4,0x28}},
{{0x0F,0x12},{0xFF,0xFF}},
{{0x0F,0x12},{0x00,0x00}},

{{0x00,0x2A},{0x18,0x9E}},	
{{0x0F,0x12},{0x0F,0xB0}},	//senHal_ExpMinPixels
{{0x00,0x2A},{0x18,0xAC}},	
{{0x0F,0x12},{0x00,0x60}},	//senHal_uAddColsBin  
{{0x0F,0x12},{0x00,0x60}},	//senHal_uAddColsNoBin
{{0x0F,0x12},{0x05,0xC0}},	//senHal_uMinColsBin  
{{0x0F,0x12},{0x05,0xC0}},	//senHal_uMinColsNoBin
{{0x00,0x2A},{0x1A,0xEA}},	
{{0x0F,0x12},{0x80,0x80}},	//senHal_SubF404Tune 
{{0x0F,0x12},{0x00,0x80}},	//senHal_FullF404Tune
{{0x00,0x2A},{0x1A,0xE0}},	                     
{{0x0F,0x12},{0x00,0x00}},	//senHal_bSenAAC     
{{0x00,0x2A},{0x1A,0x72}},	
{{0x0F,0x12},{0x00,0x00}},	//senHal_bSRX	SRX off
{{0x00,0x2A},{0x18,0xA2}},	
{{0x0F,0x12},{0x00,0x04}},	//senHal_NExpLinesCheckFine extend Forbidden area line       
{{0x00,0x2A},{0x1A,0x6A}},	                                                             
{{0x0F,0x12},{0x00,0x9A}},	//senHal_usForbiddenRightOfs extend right Forbidden area line
{{0x00,0x2A},{0x38,0x5E}},	                                                             
{{0x0F,0x12},{0x02,0x4C}},	//Mon_Sen_uExpPixelsOfs                                      
{{0x00,0x2A},{0x0E,0xE6}},	
{{0x0F,0x12},{0x00,0x00}},	//setot_bUseDigitalHbin
{{0x00,0x2A},{0x1B,0x2A}},	
{{0x0F,0x12},{0x03,0x00}},	//70001B2A //senHal_TuneStr2_usAngTuneGainTh	
{{0x0F,0x12},{0x00,0xD6}},	//70001B2C //senHal_TuneStr2_AngTuneF4CA_0_ 
{{0x0F,0x12},{0x00,0x8D}},	//70001B2E //senHal_TuneStr2_AngTuneF4CA_1_ 
{{0x0F,0x12},{0x00,0xCF}},	//70001B30 //senHal_TuneStr2_AngTuneF4C2_0_ 
{{0x0F,0x12},{0x00,0x84}},	//70001B32 //senHal_TuneStr2_AngTuneF4C2_1_
                                                                                     
//==================================================================================
// 08.AF Setting
//==================================================================================

//AF interface setting	
{{0x00,0x28},{0x70,0x00}},                                                                  
{{0x00,0x2A},{0x01,0xFC}},                                                                  
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_IPRM_LedGpio                                      
{{0x00,0x2A},{0x01,0xFE}},                                                                  
{{0x0F,0x12},{0x00,0x03}},	//REG_TC_IPRM_CM_Init_AfModeType VCM IIC                   
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_IPRM_CM_Init_PwmConfig1                           
{{0x00,0x2A},{0x02,0x04}},                                                                  
{{0x0F,0x12},{0x00,0x61}},	//REG_TC_IPRM_CM_Init_GpioConfig1 AF Enable GPIO 6	
{{0x00,0x2A},{0x02,0x0C}},                                                                  
{{0x0F,0x12},{0x2F,0x0C}},	//REG_TC_IPRM_CM_Init_Mi2cBits                             
{{0x0F,0x12},{0x01,0x90}},	//REG_TC_IPRM_CM_Init_Mi2cRateKhz IIC Speed     

{{0x00,0x2A},{0x02,0x94}},                                                                  
{{0x0F,0x12},{0x01,0x00}},	//REG_TC_AF_FstWinStartX                                   
{{0x0F,0x12},{0x00,0xE3}},	//REG_TC_AF_FstWinStartY                                   
{{0x0F,0x12},{0x02,0x00}},	//REG_TC_AF_FstWinSizeX                                    
{{0x0F,0x12},{0x02,0x38}},	//REG_TC_AF_FstWinSizeY                                    
{{0x0F,0x12},{0x01,0xC6}},	//REG_TC_AF_ScndWinStartX                                  
{{0x0F,0x12},{0x01,0x66}},	//REG_TC_AF_ScndWinStartY                                  
{{0x0F,0x12},{0x00,0x74}},	//REG_TC_AF_ScndWinSizeX                                   
{{0x0F,0x12},{0x01,0x32}},	//REG_TC_AF_ScndWinSizeY                                   
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_AF_WinSizesUpdated                                
                                                                           
	
//2nd search setting	
{{0x00,0x2A},{0x07,0x0E}},                                                                  
{{0x0F,0x12},{0x00,0xFF}}, //C0 	//skl_af_StatOvlpExpFactor                                 
{{0x00,0x2A},{0x07,0x1E}},                                                                  
{{0x0F,0x12},{0x00,0x00}}, //01 00 	//skl_af_bAfStatOff                                        
{{0x00,0x2A},{0x16,0x3C}},                                                                  
{{0x0F,0x12},{0x00,0x00}},	//af_search_usAeStable                                     
{{0x00,0x2A},{0x16,0x48}},                                                                  
{{0x0F,0x12},{0x90,0x02}},	//af_search_usSingleAfFlags                                
{{0x00,0x2A},{0x16,0x52}},                                                                  
{{0x0F,0x12},{0x00,0x02}},	//af_search_usFinePeakCount                                
{{0x0F,0x12},{0x00,0x00}},	//af_search_usFineMaxScale                                 
{{0x00,0x2A},{0x15,0xE0}},                                                                  
{{0x0F,0x12},{0x09,0x02}},	//af_pos_usFineStepNumSize                                 
                                                                           
//Peak Threshold	
{{0x00,0x2A},{0x16,0x4C}},                                                                  
{{0x0F,0x12},{0x00,0x03}},	//af_search_usMinPeakSamples                               
{{0x00,0x2A},{0x16,0x3E}},                                                                  
{{0x0F,0x12},{0x00,0xE5}}, //80 //8a //A0 	//A0 af_search_usPeakThr Full search (E5 90%)                 
{{0x0F,0x12},{0x00,0x98}}, //80 //a0 //8a  //98	//98 af_search_usPeakThrLow                                   
{{0x00,0x2A},{0x47,0xA8}},	                               
{{0x0F,0x12},{0x00,0x80}},	//TNP, Macro Threshold register
	
	
//Home Pos	
{{0x00,0x2A},{0x15,0xD4}},                                                                  
{{0x0F,0x12},{0x00,0x00}},	//af_pos_usHomePos                                         
{{0x0F,0x12},{0xD0,0x00}},	//af_pos_usLowConfPos                                      
	
	
//AF statistics	
{{0x00,0x2A},{0x16,0x9A}},                                                                  
{{0x0F,0x12},{0xFF,0x95}},	//af_search_usConfCheckOrder_1_                            
{{0x00,0x2A},{0x16,0x6A}},                                                                  
{{0x0F,0x12},{0x02,0x80}},	//af_search_usConfThr_4_                                   
{{0x00,0x2A},{0x16,0x76}},	                                                           
{{0x0F,0x12},{0x03,0xA0}},	//af_search_usConfThr_10_                                  
{{0x0F,0x12},{0x03,0x20}},	//af_search_usConfThr_11_                                  
{{0x00,0x2A},{0x16,0xBC}},	                                                           
{{0x0F,0x12},{0x00,0x30}},	//af_stat_usMinStatVal                                     
{{0x00,0x2A},{0x16,0xE0}},	                                                           
{{0x0F,0x12},{0x00,0x60}},	//af_scene_usSceneLowNormBrThr                             
{{0x00,0x2A},{0x16,0xD4}},	                                                           
{{0x0F,0x12},{0x00,0x10}},	//af_stat_usBpfThresh                                      
{{0x00,0x2A},{0x16,0x56}},	                                                           
{{0x0F,0x12},{0x00,0x00}},	//af_search_usCapturePolicy                                
{{0x00,0x2A},{0x15,0xE6}},                                                                  
{{0x0F,0x12},{0x00,0x3C}},	//af_pos_usCaptureFixedPos                                 
                                                                           
{{0x0F,0x12},{0x00,0x18}},	//af_pos_usTableLastInd                                    
{{0x0F,0x12},{0x00,0x2A}},	//af_pos_usTable_0_                                        
{{0x0F,0x12},{0x00,0x30}},	//af_pos_usTable_1_                                        
{{0x0F,0x12},{0x00,0x36}},	//af_pos_usTable_2_                                        
{{0x0F,0x12},{0x00,0x3C}},	//af_pos_usTable_3_                                        
{{0x0F,0x12},{0x00,0x42}},	//af_pos_usTable_4_  
{{0x0F,0x12},{0x00,0x48}},	//af_pos_usTable_5_                                        
{{0x0F,0x12},{0x00,0x4E}},	//af_pos_usTable_6_                                        
{{0x0F,0x12},{0x00,0x54}},	//af_pos_usTable_7_                                        
{{0x0F,0x12},{0x00,0x5A}},	//af_pos_usTable_8_                                        
{{0x0F,0x12},{0x00,0x60}},	//af_pos_usTable_9_                                        
{{0x0F,0x12},{0x00,0x66}},	//af_pos_usTable_10 
{{0x0F,0x12},{0x00,0x6C}},	//af_pos_usTable_11_                                       
{{0x0F,0x12},{0x00,0x72}},	//af_pos_usTable_12_                                       
{{0x0F,0x12},{0x00,0x78}},	//af_pos_usTable_13_                                       
{{0x0F,0x12},{0x00,0x7E}},	//af_pos_usTable_14_                                       
{{0x0F,0x12},{0x00,0x84}},	//af_pos_usTable_15_ 
{{0x0F,0x12},{0x00,0x8A}},	//af_pos_usTable_16_                                       
{{0x0F,0x12},{0x00,0x90}},	//af_pos_usTable_17_                                       
{{0x0F,0x12},{0x00,0x96}},	//af_pos_usTable_18_                                       
{{0x0F,0x12},{0x00,0x9C}},	//af_pos_usTable_19_                                       
{{0x0F,0x12},{0x00,0xA2}},	//af_pos_usTable_20_                                       
{{0x0F,0x12},{0x00,0xA8}},	//af_pos_usTable_21_                                       
{{0x0F,0x12},{0x00,0xAE}},	//af_pos_usTable_22_                                       
{{0x0F,0x12},{0x00,0xB4}},	//af_pos_usTable_23_                                       
{{0x0F,0x12},{0x00,0xBA}},	//af_pos_usTable_24_  
  	
//VCM AF driver with PWM/I2C                                                                         
{{0x00,0x2A},{0x17,0x22}},                                                                  
{{0x0F,0x12},{0x80,0x00}},	//afd_usParam_0_                                           
{{0x0F,0x12},{0x00,0x06}},	//afd_usParam_1_                                           
{{0x0F,0x12},{0x3F,0xF0}},	//afd_usParam_2_                                           
{{0x0F,0x12},{0x03,0xE8}},	//afd_usParam_3_                                           
{{0x0F,0x12},{0x00,0x00}},	//afd_usParam_4_                                           
{{0x0F,0x12},{0x00,0x80}},   //10 afd_usParam_5_                                           
{{0x0F,0x12},{0x00,0x10}},	//10 afd_usParam_6_                                           
{{0x0F,0x12},{0x00,0x10}},	//08 afd_usParam_7_                                           
{{0x0F,0x12},{0x00,0x40}},	//afd_usParam_8_                                           
{{0x0F,0x12},{0x00,0x80}},	//afd_usParam_9_                                           
{{0x0F,0x12},{0x00,0xC0}},	//afd_usParam_10_                                          
{{0x0F,0x12},{0x00,0xE0}},	//afd_usParam_11_                                          
                                                                           
{{0x00,0x2A},{0x02,0x8C}},                                                                  
{{0x0F,0x12},{0x00,0x03}},	//REG_TC_AF_AfCmd         

//==================================================================================
// 09.AWB-BASIC setting
//==================================================================================

// AWB init Start point
{{0x00,0x2A},{0x14,0x5E}},
{{0x0F,0x12},{0x05,0xC0}},	//620 580 //awbb_GainsInit_0_
{{0x0F,0x12},{0x04,0x28}},	//awbb_GainsInit_1_
{{0x0F,0x12},{0x06,0xE0}},	//6F0 780 //awbb_GainsInit_2_

// AWB Convergence Speed
{{0x00,0x2A},{0x14,0x64}},
{{0x0F,0x12},{0x00,0x08}},	//awbb_WpFilterMinThr
{{0x0F,0x12},{0x00,0x60}},	//190 awbb_WpFilterMaxThr
{{0x0F,0x12},{0x01,0x00}},	//F0 awbb_WpFilterCoef
{{0x0F,0x12},{0x00,0x04}},	//awbb_WpFilterSize		
{{0x0F,0x12},{0x00,0x02}},	//awbb_GridEnable	

{{0x00,0x2A},{0x14,0x4E}},			
{{0x0F,0x12},{0x00,0x00}},	//awbb_RGainOff 		
{{0x0F,0x12},{0x00,0x00}},	//awbb_BGainOff 		
{{0x0F,0x12},{0x00,0x00}},	//awbb_GGainOff 		
{{0x0f,0x12},{0x00,0xC2}},	//awbb_Alpha_Comp_Mode		
{{0x0F,0x12},{0x00,0x02}},	//awbb_Rpl_InvalidOutDoor		
{{0x0F,0x12},{0x00,0x01}},	//awbb_UseGrThrCorr		
{{0x0F,0x12},{0x00,0x74}},	//awbb_Use_Filters     		
{{0x0F,0x12},{0x00,0x01}},	//awbb_CorrectMinNumPatches		

// White Locus
{{0x00,0x2A},{0x11,0xF0}},
{{0x0F,0x12},{0x01,0x2C}},	//awbb_IntcR
{{0x0F,0x12},{0x01,0x21}},	//awbb_IntcB
{{0x0F,0x12},{0x02,0xDF}},	//awbb_GLocusR		
{{0x0F,0x12},{0x03,0x14}},	//awbb_GLocusB	

{{0x00,0x2A},{0x12,0x0E}},
{{0x0F,0x12},{0x00,0x00}},	//awbb_MovingScale10
{{0x0F,0x12},{0x05,0xFD}},	//awbb_GamutWidthThr1
{{0x0F,0x12},{0x03,0x6B}},	//awbb_GamutHeightThr1
{{0x0F,0x12},{0x00,0x20}},	//awbb_GamutWidthThr2
{{0x0F,0x12},{0x00,0x1A}},	//awbb_GamutHeightThr2

{{0x00,0x2A},{0x12,0x78}},
{{0x0F,0x12},{0xFE,0xF7}},	//awbb_SCDetectionMap_SEC_StartR_B
{{0x0F,0x12},{0x00,0x21}},	//awbb_SCDetectionMap_SEC_StepR_B
{{0x0F,0x12},{0x07,0xD0}},	//awbb_SCDetectionMap_SEC_SunnyNB
{{0x0F,0x12},{0x07,0xD0}}, 	//awbb_SCDetectionMap_SEC_StepNB
{{0x0F,0x12},{0x01,0xC8}},	//awbb_SCDetectionMap_SEC_LowTempR_B
{{0x0F,0x12},{0x00,0x96}},	//awbb_SCDetectionMap_SEC_SunnyNBZone
{{0x0F,0x12},{0x00,0x04}},	//awbb_SCDetectionMap_SEC_LowTempR_BZone

{{0x00,0x2A},{0x12,0x24}},
{{0x0F,0x12},{0x00,0x32}},	//awbb_LowBr
{{0x0F,0x12},{0x00,0x1E}},	//awbb_LowBr_NBzone
{{0x0F,0x12},{0x00,0xE2}},	//awbb_YThreshHigh
{{0x0F,0x12},{0x00,0x10}},	//awbb_YThreshLow_Norm
{{0x0F,0x12},{0x00,0x02}},	//awbb_YThreshLow_Low

{{0x00,0x2A},{0x2B,0xA4}},
{{0x0F,0x12},{0x00,0x04}},	//Mon_AWB_ByPassMode
				
{{0x00,0x2A},{0x11,0xFC}},			
{{0x0F,0x12},{0x00,0x0C}},	//awbb_MinNumOfFinalPatches  		

{{0x00,0x2A},{0x12,0x08}},  
{{0x0F,0x12},{0x00,0x20}},	//awbb_MinNumOfChromaclassifpatches
			
// Indoor Zone
{{0x00,0x2A},{0x10,0x1C}},
{{0x0F,0x12},{0x03,0x60}}, //0360 //0360	//awbb_IndoorGrZones_m_BGrid_0__m_left
{{0x0F,0x12},{0x03,0x6C}}, //036C //036C	//awbb_IndoorGrZones_m_BGrid_0__m_right
{{0x0F,0x12},{0x03,0x20}}, //0320 //0320	//awbb_IndoorGrZones_m_BGrid_1__m_left
{{0x0F,0x12},{0x03,0x8A}}, //038A //038A	//awbb_IndoorGrZones_m_BGrid_1__m_right
{{0x0F,0x12},{0x02,0xE8}}, //02E8 //02E8	//awbb_IndoorGrZones_m_BGrid_2__m_left
{{0x0F,0x12},{0x03,0x6C}}, //036C //0380	//awbb_IndoorGrZones_m_BGrid_2__m_right
{{0x0F,0x12},{0x02,0xBE}}, //02BE //02BE	//awbb_IndoorGrZones_m_BGrid_3__m_left
{{0x0F,0x12},{0x03,0x42}}, //0342 //035A	//awbb_IndoorGrZones_m_BGrid_3__m_right
{{0x0F,0x12},{0x02,0x98}}, //0298 //0298	//awbb_IndoorGrZones_m_BGrid_4__m_left
{{0x0F,0x12},{0x03,0x1C}}, //031C //0334	//awbb_IndoorGrZones_m_BGrid_4__m_right
{{0x0F,0x12},{0x02,0x72}}, //0272 //0272	//awbb_IndoorGrZones_m_BGrid_5__m_left
{{0x0F,0x12},{0x02,0xF6}}, //02F6 //030E	//awbb_IndoorGrZones_m_BGrid_5__m_right
{{0x0F,0x12},{0x02,0x4C}}, //024C //024C	//awbb_IndoorGrZones_m_BGrid_6__m_left
{{0x0F,0x12},{0x02,0xD6}}, //02D2 //02EA	//awbb_IndoorGrZones_m_BGrid_6__m_right
{{0x0F,0x12},{0x02,0x30}}, //0230 //0230	//awbb_IndoorGrZones_m_BGrid_7__m_left
{{0x0F,0x12},{0x02,0xBA}}, //02B6 //02CC	//awbb_IndoorGrZones_m_BGrid_7__m_right
{{0x0F,0x12},{0x02,0x14}}, //0214 //0214	//awbb_IndoorGrZones_m_BGrid_8__m_left
{{0x0F,0x12},{0x02,0xA6}}, //02A6 //02B0	//awbb_IndoorGrZones_m_BGrid_8__m_right
{{0x0F,0x12},{0x01,0xF8}}, //01F8 //01F8	//awbb_IndoorGrZones_m_BGrid_9__m_left
{{0x0F,0x12},{0x02,0x92}}, //0292 //0294	//awbb_IndoorGrZones_m_BGrid_9__m_right
{{0x0F,0x12},{0x01,0xDC}}, //01DC //01DC	//awbb_IndoorGrZones_m_BGrid_10__m_left
{{0x0F,0x12},{0x02,0x78}}, //0278 //0278	//awbb_IndoorGrZones_m_BGrid_10__m_right
{{0x0F,0x12},{0x01,0xC0}}, //01C0 //01C0	//awbb_IndoorGrZones_m_BGrid_11__m_left
{{0x0F,0x12},{0x02,0x64}}, //0264 //0264	//awbb_IndoorGrZones_m_BGrid_11__m_right
{{0x0F,0x12},{0x01,0xAA}}, //01AA //01AA	//awbb_IndoorGrZones_m_BGrid_12__m_left
{{0x0F,0x12},{0x02,0x50}}, //0250 //0250	//awbb_IndoorGrZones_m_BGrid_12__m_right
{{0x0F,0x12},{0x01,0x96}}, //0196 //0196	//awbb_IndoorGrZones_m_BGrid_13__m_left
{{0x0F,0x12},{0x02,0x3C}}, //023C //023C	//awbb_IndoorGrZones_m_BGrid_13__m_right
{{0x0F,0x12},{0x01,0x80}}, //0180 //0180	//awbb_IndoorGrZones_m_BGrid_14__m_left
{{0x0F,0x12},{0x02,0x28}}, //0228 //0228	//awbb_IndoorGrZones_m_BGrid_14__m_right
{{0x0F,0x12},{0x01,0x6C}}, //016C //016C	//awbb_IndoorGrZones_m_BGrid_15__m_left
{{0x0F,0x12},{0x02,0x14}}, //0214 //0214	//awbb_IndoorGrZones_m_BGrid_15__m_right
{{0x0F,0x12},{0x01,0x68}}, //0168 //0168	//awbb_IndoorGrZones_m_BGrid_16__m_left
{{0x0F,0x12},{0x02,0x00}}, //0200 //0200	//awbb_IndoorGrZones_m_BGrid_16__m_right
{{0x0F,0x12},{0x01,0x72}}, //0172 //0172	//awbb_IndoorGrZones_m_BGrid_17__m_left
{{0x0F,0x12},{0x01,0xEC}}, //01EC //01EC	//awbb_IndoorGrZones_m_BGrid_17__m_right
{{0x0F,0x12},{0x01,0x9A}}, //019A //019A	//awbb_IndoorGrZones_m_BGrid_18__m_left
{{0x0F,0x12},{0x01,0xD8}}, //01D8 //01D8	//awbb_IndoorGrZones_m_BGrid_18__m_right
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000	//awbb_IndoorGrZones_m_BGrid_19__m_left
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000	//awbb_IndoorGrZones_m_BGrid_19__m_right

{{0x0F,0x12},{0x00,0x05}},	//awbb_IndoorGrZones_m_GridStep
{{0x00,0x2A},{0x10,0x70}},
{{0x0F,0x12},{0x00,0x13}},	//awbb_IndoorGrZones_ZInfo_m_GridSz
{{0x00,0x2A},{0x10,0x74}},
{{0x0F,0x12},{0x00,0xEC}},	//awbb_IndoorGrZones_m_Boffs

// Outdoor Zone
{{0x00,0x2A},{0x10,0x78}},
{{0x0F,0x12},{0x02,0x40}}, //0240 //0232	//awbb_OutdoorGrZones_m_BGrid_0__m_left
{{0x0F,0x12},{0x02,0x5A}}, //025A //025A	//awbb_OutdoorGrZones_m_BGrid_0__m_right
{{0x0F,0x12},{0x02,0x34}}, //022C //021E	//awbb_OutdoorGrZones_m_BGrid_1__m_left
{{0x0F,0x12},{0x02,0x74}}, //0274 //0274	//awbb_OutdoorGrZones_m_BGrid_1__m_right
{{0x0F,0x12},{0x02,0x28}}, //021E //020E	//awbb_OutdoorGrZones_m_BGrid_2__m_left
{{0x0F,0x12},{0x02,0x8E}}, //028E //028E	//awbb_OutdoorGrZones_m_BGrid_2__m_right
{{0x0F,0x12},{0x02,0x20}}, //0210 //0200	//awbb_OutdoorGrZones_m_BGrid_3__m_left
{{0x0F,0x12},{0x02,0x90}}, //0290 //0290	//awbb_OutdoorGrZones_m_BGrid_3__m_right
{{0x0F,0x12},{0x02,0x12}}, //0204 //01F4	//awbb_OutdoorGrZones_m_BGrid_4__m_left
{{0x0F,0x12},{0x02,0x86}}, //0286 //0286	//awbb_OutdoorGrZones_m_BGrid_4__m_right
{{0x0F,0x12},{0x02,0x02}}, //01FA //01E8	//awbb_OutdoorGrZones_m_BGrid_5__m_left
{{0x0F,0x12},{0x02,0x7E}}, //027E //027E	//awbb_OutdoorGrZones_m_BGrid_5__m_right
{{0x0F,0x12},{0x01,0xF6}}, //01F2 //01DE	//awbb_OutdoorGrZones_m_BGrid_6__m_left
{{0x0F,0x12},{0x02,0x74}}, //0274 //0274	//awbb_OutdoorGrZones_m_BGrid_6__m_right
{{0x0F,0x12},{0x01,0xEC}}, //01EA //01D2	//awbb_OutdoorGrZones_m_BGrid_7__m_left
{{0x0F,0x12},{0x02,0x68}}, //0268 //0268	//awbb_OutdoorGrZones_m_BGrid_7__m_right
{{0x0F,0x12},{0x01,0xE6}}, //01E6 //01D0	//awbb_OutdoorGrZones_m_BGrid_8__m_left
{{0x0F,0x12},{0x02,0x5E}}, //025E //025E	//awbb_OutdoorGrZones_m_BGrid_8__m_right
{{0x0F,0x12},{0x01,0xE4}}, //01E4 //01D6	//awbb_OutdoorGrZones_m_BGrid_9__m_left
{{0x0F,0x12},{0x02,0x52}}, //0252 //0252	//awbb_OutdoorGrZones_m_BGrid_9__m_right
{{0x0F,0x12},{0x01,0xE6}}, //01E6 //01E2	//awbb_OutdoorGrZones_m_BGrid_10__m_left
{{0x0F,0x12},{0x02,0x48}}, //0248 //0248	//awbb_OutdoorGrZones_m_BGrid_10__m_right
{{0x0F,0x12},{0x01,0xF4}}, //01F4 //01F4	//awbb_OutdoorGrZones_m_BGrid_11__m_left
{{0x0F,0x12},{0x02,0x1A}}, //021A //021A	//awbb_OutdoorGrZones_m_BGrid_11__m_right
           
{{0x0F,0x12},{0x00,0x04}},	//awbb_OutdoorGrZones_m_GridStep
{{0x00,0x2A},{0x10,0xAC}},
{{0x0F,0x12},{0x00,0x0C}},	//awbb_OutdoorGrZones_ZInfo_m_GridSz
{{0x00,0x2A},{0x10,0xB0}},
{{0x0F,0x12},{0x01,0xDA}},	//awbb_OutdoorGrZones_m_Boffs

// Low Brightness Zone
{{0x00,0x2A},{0x10,0xB4}},
{{0x0F,0x12},{0x03,0x48}},	//awbb_LowBrGrZones_m_BGrid_0__m_left
{{0x0F,0x12},{0x03,0xB6}},	//awbb_LowBrGrZones_m_BGrid_0__m_right
{{0x0F,0x12},{0x02,0xB8}},	//awbb_LowBrGrZones_m_BGrid_1__m_left
{{0x0F,0x12},{0x03,0xB6}},	//awbb_LowBrGrZones_m_BGrid_1__m_right
{{0x0F,0x12},{0x02,0x58}},	//awbb_LowBrGrZones_m_BGrid_2__m_left
{{0x0F,0x12},{0x03,0x8E}},	//awbb_LowBrGrZones_m_BGrid_2__m_right
{{0x0F,0x12},{0x02,0x12}},	//awbb_LowBrGrZones_m_BGrid_3__m_left
{{0x0F,0x12},{0x03,0x48}},	//awbb_LowBrGrZones_m_BGrid_3__m_right
{{0x0F,0x12},{0x01,0xCC}},	//awbb_LowBrGrZones_m_BGrid_4__m_left
{{0x0F,0x12},{0x03,0x0C}},	//awbb_LowBrGrZones_m_BGrid_4__m_right
{{0x0F,0x12},{0x01,0xA2}},	//awbb_LowBrGrZones_m_BGrid_5__m_left
{{0x0F,0x12},{0x02,0xD2}},	//awbb_LowBrGrZones_m_BGrid_5__m_right
{{0x0F,0x12},{0x01,0x70}},	//awbb_LowBrGrZones_m_BGrid_6__m_left
{{0x0F,0x12},{0x02,0xA6}},	//awbb_LowBrGrZones_m_BGrid_6__m_right
{{0x0F,0x12},{0x01,0x4C}},	//awbb_LowBrGrZones_m_BGrid_7__m_left
{{0x0F,0x12},{0x02,0x80}},	//awbb_LowBrGrZones_m_BGrid_7__m_right
{{0x0F,0x12},{0x01,0x28}},	//awbb_LowBrGrZones_m_BGrid_8__m_left
{{0x0F,0x12},{0x02,0x5C}},	//awbb_LowBrGrZones_m_BGrid_8__m_right
{{0x0F,0x12},{0x01,0x46}},	//awbb_LowBrGrZones_m_BGrid_9__m_left
{{0x0F,0x12},{0x02,0x36}},	//awbb_LowBrGrZones_m_BGrid_9__m_right
{{0x0F,0x12},{0x01,0x64}},	//awbb_LowBrGrZones_m_BGrid_10__m_left
{{0x0F,0x12},{0x02,0x12}},	//awbb_LowBrGrZones_m_BGrid_10__m_right
{{0x0F,0x12},{0x00,0x00}},	//awbb_LowBrGrZones_m_BGrid_11__m_left
{{0x0F,0x12},{0x00,0x00}},	//awbb_LowBrGrZones_m_BGrid_11__m_right

{{0x0F,0x12},{0x00,0x06}},	//awbb_LowBrGrZones_m_GridStep
{{0x00,0x2A},{0x10,0xE8}},
{{0x0F,0x12},{0x00,0x0B}},	//awbb_LowBrGrZones_ZInfo_m_GridSz
{{0x00,0x2A},{0x10,0xEC}},
{{0x0F,0x12},{0x00,0xD2}},	//awbb_LowBrGrZones_m_Boffs

// Low Temp. Zone
{{0x00,0x2A},{0x10,0xF0}},
{{0x0F,0x12},{0x03,0x9A}},
{{0x0F,0x12},{0x00,0x00}},	//awbb_CrclLowT_R_c
{{0x0F,0x12},{0x00,0xFE}},
{{0x0F,0x12},{0x00,0x00}},	//awbb_CrclLowT_B_c
{{0x0F,0x12},{0x22,0x84}},
{{0x0F,0x12},{0x00,0x00}},	//awbb_CrclLowT_Rad_c   	
			
//AWB - GridCorrection			
{{0x00,0x2A},{0x14,0x34}},
{{0x0F,0x12},{0x02,0xC1}},	//awbb_GridConst_1_0_
{{0x0F,0x12},{0x03,0x3A}},	//awbb_GridConst_1_1_
{{0x0F,0x12},{0x03,0x8A}},	//awbb_GridConst_1_2_
{{0x0F,0x12},{0x10,0x1A}},	//awbb_GridConst_2_0_
{{0x0F,0x12},{0x10,0x75}},	//awbb_GridConst_2_1_
{{0x0F,0x12},{0x11,0x3D}},	//awbb_GridConst_2_2_
{{0x0F,0x12},{0x11,0x3F}},	//awbb_GridConst_2_3_
{{0x0F,0x12},{0x11,0xAF}},	//awbb_GridConst_2_4_
{{0x0F,0x12},{0x11,0xF0}},	//awbb_GridConst_2_5_
{{0x0F,0x12},{0x00,0xB2}},	//awbb_GridCoeff_R_1
{{0x0F,0x12},{0x00,0xB8}},	//awbb_GridCoeff_B_1
{{0x0F,0x12},{0x00,0xCA}},	//awbb_GridCoeff_R_2
{{0x0F,0x12},{0x00,0x9D}},	//awbb_GridCoeff_B_2

// Indoor Grid Offset			
{{0x00,0x2A},{0x13,0xA4}},
{{0x0F,0x12},{0x00,0x00}},	//D65 awbb_GridCorr_R_0__0_
{{0x0F,0x12},{0xff,0xd0}},	//D65 CW awbb_GridCorr_R_0__1_
{{0x0F,0x12},{0xFF,0xE0}},	//CW awbb_GridCorr_R_0__2_
{{0x0F,0x12},{0x00,0x00}},	//FFF0 FFd0 A awbb_GridCorr_R_0__3_
{{0x0F,0x12},{0x00,0x00}},	//FFF0 FFd0 A awbb_GridCorr_R_0__4_
{{0x0F,0x12},{0x00,0x00}},	//FFF0 FFd0 H awbb_GridCorr_R_0__5_

{{0x0F,0x12},{0x00,0x00}},	//D65awbb_GridCorr_R_1__0_
{{0x0F,0x12},{0xff,0xd0}},	//D65 CW awbb_GridCorr_R_1__1_
{{0x0F,0x12},{0xFF,0xE0}},	//CW awbb_GridCorr_R_1__2_
{{0x0F,0x12},{0x00,0x00}},	//A awbb_GridCorr_R_1__3_
{{0x0F,0x12},{0x00,0x00}},	//A awbb_GridCorr_R_1__4_
{{0x0F,0x12},{0x00,0x00}},	//H awbb_GridCorr_R_1__5_

{{0x0F,0x12},{0x00,0x00}},	//D65awbb_GridCorr_R_2__0_
{{0x0F,0x12},{0xff,0xd0}},	//D65 CW awbb_GridCorr_R_2__1_
{{0x0F,0x12},{0xFF,0xE0}},	//CW awbb_GridCorr_R_2__2_
{{0x0F,0x12},{0x00,0x00}},	//A awbb_GridCorr_R_2__3_
{{0x0F,0x12},{0x00,0x00}},	//A awbb_GridCorr_R_2__4_
{{0x0F,0x12},{0x00,0x00}},	//H awbb_GridCorr_R_2__5_

{{0x0F,0x12},{0xFF,0xD0}},	//awbb_GridCorr_B_0__0_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_0__1_
{{0x0F,0x12},{0x00,0x20}},	//awbb_GridCorr_B_0__2_
{{0x0F,0x12},{0xFF,0xE0}},	//FFE0 0000 awbb_GridCorr_B_0__3_
{{0x0F,0x12},{0xFF,0xE0}},	//FFE0 0000 awbb_GridCorr_B_0__4_
{{0x0F,0x12},{0xFF,0xE0}},	//FFE0 0000 awbb_GridCorr_B_0__5_

{{0x0F,0x12},{0xFF,0xD0}},	//awbb_GridCorr_B_1__0_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_1__1_
{{0x0F,0x12},{0x00,0x20}},	//awbb_GridCorr_B_1__2_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_1__3_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_1__4_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_1__5_

{{0x0F,0x12},{0xFF,0xD0}},	//awbb_GridCorr_B_2__0_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_2__1_
{{0x0F,0x12},{0x00,0x20}},	//awbb_GridCorr_B_2__2_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_2__3_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_2__4_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_B_2__5_

// Outdoor Grid Offset
{{0x0F,0x12},{0xFF,0xD0}}, //FFB0	//FFC0 awbb_GridCorr_R_Out_0__0_
{{0x0F,0x12},{0xFF,0xE0}}, //FFC0	//FFD0 awbb_GridCorr_R_Out_0__1_
{{0x0F,0x12},{0xFF,0xE0}}, //FFC0	//FFD0 awbb_GridCorr_R_Out_0__2_
{{0x0F,0x12},{0xFF,0xE0}}, //FFC0	//FFD0 awbb_GridCorr_R_Out_0__3_
{{0x0F,0x12},{0x00,0x00}}, //0000	//0000 awbb_GridCorr_R_Out_0__4_
{{0x0F,0x12},{0x00,0x00}}, //0000	//0000 awbb_GridCorr_R_Out_0__5_

{{0x0F,0x12},{0xFF,0xD0}},	//awbb_GridCorr_R_Out_1__0_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_R_Out_1__1_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_R_Out_1__2_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_R_Out_1__3_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_R_Out_1__4_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_R_Out_1__5_

{{0x0F,0x12},{0xFF,0xD0}},	//awbb_GridCorr_R_Out_2__0_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_R_Out_2__1_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_R_Out_2__2_
{{0x0F,0x12},{0xFF,0xE0}},	//awbb_GridCorr_R_Out_2__3_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_R_Out_2__4_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_R_Out_2__5_

{{0x0F,0x12},{0x00,0x90}}, //0080	//awbb_GridCorr_B_Out_0__0_
{{0x0F,0x12},{0xFF,0xF0}}, //FFE0	//awbb_GridCorr_B_Out_0__1_
{{0x0F,0x12},{0xFF,0xF0}}, //FFE0	//awbb_GridCorr_B_Out_0__2_
{{0x0F,0x12},{0xFF,0xF0}}, //FFE0	//awbb_GridCorr_B_Out_0__3_
{{0x0F,0x12},{0x00,0x00}}, //0000	//awbb_GridCorr_B_Out_0__4_
{{0x0F,0x12},{0x00,0x00}}, //0000	//awbb_GridCorr_B_Out_0__5_

{{0x0F,0x12},{0x00,0x90}}, 	//awbb_GridCorr_B_Out_1__0_
{{0x0F,0x12},{0xFF,0xF0}},	//awbb_GridCorr_B_Out_1__1_
{{0x0F,0x12},{0xFF,0xF0}},	//awbb_GridCorr_B_Out_1__2_
{{0x0F,0x12},{0xFF,0xF0}},	//awbb_GridCorr_B_Out_1__3_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_B_Out_1__4_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_B_Out_1__5_

{{0x0F,0x12},{0x00,0x90}},	//awbb_GridCorr_B_Out_2__0_
{{0x0F,0x12},{0xFF,0xF0}},	//awbb_GridCorr_B_Out_2__1_
{{0x0F,0x12},{0xFF,0xF0}},	//awbb_GridCorr_B_Out_2__2_
{{0x0F,0x12},{0xFF,0xF0}},	//awbb_GridCorr_B_Out_2__3_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_B_Out_2__4_
{{0x0F,0x12},{0x00,0x00}},	//awbb_GridCorr_B_Out_2__5_

//==================================================================================
// 10.Clock Setting
//==================================================================================

//Input Clock (Mclk)	
{{0x00,0x2A},{0x01,0xF8}},	
{{0x0F,0x12},{0x5D,0xC0}},	//REG_TC_IPRM_InClockLSBs
{{0x00,0x2A},{0x02,0x12}},	                                     
{{0x0F,0x12},{0x00,0x02}},   //0	//REG_TC_IPRM_UseNPviClocks          
{{0x0F,0x12},{0x00,0x00}},   //2	//REG_TC_IPRM_UseNMipiClocks         
{{0x0F,0x12},{0x00,0x00}},   //2	//REG_TC_IPRM_NumberOfMipiLanes
	
//System Clock & Output clock (Pclk)		
{{0x00,0x2A},{0x02,0x1A}},	
{{0x0F,0x12},{0x34,0xBC}},   //REG_TC_IPRM_OpClk4KHz_0 
{{0x0F,0x12},{0x4F,0x1A}},   //REG_TC_IPRM_MinOutRate4KHz_0
{{0x0F,0x12},{0x4F,0x1A}},   //REG_TC_IPRM_MaxOutRate4KHz_0
{{0x0F,0x12},{0x4F,0x1A}},   //REG_TC_IPRM_OpClk4KHz_1	SCLK 	 : 81Mhz
{{0x0F,0x12},{0x4F,0x1A}},   //REG_TC_IPRM_MinOutRate4KHz_1	PCLK Min : 81Mhz
{{0x0F,0x12},{0x4F,0x1A}},   //REG_TC_IPRM_MaxOutRate4KHz_1 PCLK Max : 81Mhz

//==================================================================================
// 11.Auto Flicker Detection
//==================================================================================

{{0x00,0x2A},{0x0F,0x30}}, 	
{{0x0F,0x12},{0x00,0x01}},   	//AFC_D_ConvAccelerPower 

// Auto Flicker (60Mhz start)	
{{0x00,0x2A},{0x0F,0x2A}},	
{{0x0F,0x12},{0x00,0x00}},	//AFC_Default BIT[0] 1:60Hz 0:50Hz
{{0x00,0x2A},{0x04,0xE6}},	                                      
{{0x0F,0x12},{0x07,0x5F}},	//REG_TC_DBG 7F: 60Hz  5F:50Hz

{{0x00,0x2A},{0x04,0xE6}},
{{0x0F,0x12},{0x07,0x5F}},	//REG_TC_DBG 7F: 60Hz  5F:50Hz
{{0x00,0x2a},{0x04,0xd6}},
{{0x0F,0x12},{0x00,0x01}},	//REG_SF_USER_FlickerQuant 01h : 50hz  02h : 60hz
{{0x0F,0x12},{0x00,0x01}},	//REG_SF_USER_FlickerQuantChanged

//==================================================================================
// 12.AE Setting
//==================================================================================

//AE Target	
{{0x00,0x2A},{0x14,0x84}},	
{{0x0F,0x12},{0x00,0x3C}},   	//TVAR_ae_BrAve

//ae_StatMode bit[3] BLC has to be bypassed to prevent AE weight change especially backlight scene 
{{0x00,0x2A},{0x14,0x8A}},	
{{0x0F,0x12},{0x00,0x0F}},  	 //ae_StatMode

{{0x00,0x2A},{0x05,0x88}},
{{0x0F,0x12},{0x00,0x02}},	//lt_uInitPostToleranceCnt

//AE_state                	
{{0x00,0x2A},{0x05,0x44}},	
{{0x0F,0x12},{0x01,0x11}},  	 //lt_uLimitHigh
{{0x0F,0x12},{0x00,0xEF}},  	 //lt_uLimitLow

//AE Concept	
{{0x00,0x2A},{0x06,0x08}},                                	
{{0x0F,0x12},{0x00,0x01}}, 	//lt_ExpGain_uSubsamplingmode
{{0x0F,0x12},{0x00,0x01}},	//lt_ExpGain_uNonSubsampling
	
//Exposure	
{{0x00,0x2a},{0x06,0x10}},	//lt_ExpGain_ExpCurveGainMaxStr_0__ulExpIn_0
{{0x0F,0x12},{0x00,0x01}},
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x0A,0x3C}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x0D,0x05}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x40,0x08}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x70,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x9C,0x00}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0xAD,0x00}},	
{{0x0F,0x12},{0x00,0x01}},	
{{0x0F,0x12},{0xF1,0xD4}},	
{{0x0F,0x12},{0x00,0x02}},	
{{0x0F,0x12},{0xDC,0x00}},	
{{0x0F,0x12},{0x00,0x05}},	
{{0x0F,0x12},{0xDC,0x00}},	
{{0x0F,0x12},{0x00,0x05}},	
	
{{0x00,0x2A},{0x06,0x38}},	//lt_ExpGain_ExpCurveGainMaxStr_0__ulExpOut_0_
{{0x0F,0x12},{0x00,0x01}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x0A,0x3C}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x0D,0x05}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x34,0x08}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x34,0x08}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x68,0x10}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0x82,0x14}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0xC3,0x50}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0xC3,0x50}},	
{{0x0F,0x12},{0x00,0x00}},	
{{0x0F,0x12},{0xC3,0x50}},	
{{0x0F,0x12},{0x00,0x00}},	
	
//Gain	
{{0x00,0x2A},{0x06,0x0C}},	
{{0x0F,0x12},{0x08,0x00}},	//lt_ExpGain_ExpCurveGainMaxStr
{{0x0F,0x12},{0x01,0x00}},	//lt_ExpGain_ExpCurveGainMaxStr_0__uMaxDigGain
{{0x00,0x2A},{0x05,0xa2}},		
{{0x0F,0x12},{0x10,0x00}},	//lt_uMaxTotGain
	
// Lei Control	
{{0x00,0x2A},{0x06,0xB8}},               	
{{0x0F,0x12},{0x45,0x2C}},               	
{{0x0F,0x12},{0x00,0x05}},  	 //lt_uMaxLei

//==================================================================================
// 13.AE Weight (Normal)
//==================================================================================
	
{{0x00,0x2A},{0x14,0x92}},	
{{0x0F,0x12},{0x01,0x00}},	//ae_WeightTbl_16[0]			
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[1]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[2]
{{0x0F,0x12},{0x00,0x01}},	//ae_WeightTbl_16[3]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[4]
{{0x0F,0x12},{0x02,0x01}},	//ae_WeightTbl_16[5]
{{0x0F,0x12},{0x01,0x02}},	//ae_WeightTbl_16[6]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[7]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[8]
{{0x0F,0x12},{0x02,0x02}},	//ae_WeightTbl_16[9]
{{0x0F,0x12},{0x02,0x02}},	//ae_WeightTbl_16[10]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[11]
{{0x0F,0x12},{0x02,0x01}},	//ae_WeightTbl_16[12]
{{0x0F,0x12},{0x03,0x02}},	//ae_WeightTbl_16[13]
{{0x0F,0x12},{0x02,0x03}},	//ae_WeightTbl_16[14]
{{0x0F,0x12},{0x01,0x02}},	//ae_WeightTbl_16[15]
{{0x0F,0x12},{0x02,0x01}},	//ae_WeightTbl_16[16]
{{0x0F,0x12},{0x03,0x02}},	//ae_WeightTbl_16[17]
{{0x0F,0x12},{0x02,0x03}},	//ae_WeightTbl_16[18]
{{0x0F,0x12},{0x01,0x02}},	//ae_WeightTbl_16[19]
{{0x0F,0x12},{0x02,0x01}},	//ae_WeightTbl_16[20]
{{0x0F,0x12},{0x02,0x02}},	//ae_WeightTbl_16[21]
{{0x0F,0x12},{0x02,0x02}},	//ae_WeightTbl_16[22]
{{0x0F,0x12},{0x01,0x02}},	//ae_WeightTbl_16[23]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[24]
{{0x0F,0x12},{0x02,0x02}},	//ae_WeightTbl_16[25]
{{0x0F,0x12},{0x02,0x02}},	//ae_WeightTbl_16[26]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[27]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[28]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[29]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[30]
{{0x0F,0x12},{0x01,0x01}},	//ae_WeightTbl_16[31]

//==================================================================================
// 14.Flash Setting
//==================================================================================

{{0x00,0x2A},{0x04,0x84}},	
{{0x0F,0x12},{0x00,0x02}},	//capture flash on                         
{{0x00,0x2A},{0x18,0x3A}},	                                           
{{0x0F,0x12},{0x00,0x01}},	//one frame AE                             
{{0x00,0x2A},{0x17,0xF6}},	                                           
{{0x0F,0x12},{0x02,0x3C}},	//AWB R point                              
{{0x0F,0x12},{0x02,0x48}},	//AWB B point                              
{{0x00,0x2A},{0x18,0x40}},	                                           
{{0x0F,0x12},{0x00,0x01}},	// Fls AE tune start                       
{{0x0F,0x12},{0x01,0x00}},	// fls_afl_FlsAFIn  Rin                    
{{0x0F,0x12},{0x01,0x20}},	                                           
{{0x0F,0x12},{0x01,0x80}},	                                           
{{0x0F,0x12},{0x02,0x00}},	                                           
{{0x0F,0x12},{0x04,0x00}},	                                           
{{0x0F,0x12},{0x08,0x00}},	                                           
{{0x0F,0x12},{0x0A,0x00}},	                                           
{{0x0F,0x12},{0x10,0x00}},	                                           
{{0x0F,0x12},{0x01,0x00}},	// fls_afl_FlsAFOut  Rout                  
{{0x0F,0x12},{0x00,0xA0}},	                                           
{{0x0F,0x12},{0x00,0x90}},	                                           
{{0x0F,0x12},{0x00,0x80}},	                                           
{{0x0F,0x12},{0x00,0x70}},	                                           
{{0x0F,0x12},{0x00,0x45}},	                                           
{{0x0F,0x12},{0x00,0x30}},	                                           
{{0x0F,0x12},{0x00,0x10}},	                                           
{{0x00,0x2A},{0x18,0x84}},	                                           
{{0x0F,0x12},{0x01,0x00}},	// fls_afl_FlsNBOut  flash NB default      
{{0x0F,0x12},{0x01,0x00}},	                                           
{{0x0F,0x12},{0x01,0x00}},	                                           
{{0x0F,0x12},{0x01,0x00}},	                                           
{{0x0F,0x12},{0x01,0x00}},	                                           
{{0x0F,0x12},{0x01,0x00}},	                                           
{{0x0F,0x12},{0x01,0x00}},	                                           
{{0x0F,0x12},{0x01,0x00}},	                                           
{{0x00,0x2A},{0x18,0x26}},	                                           
{{0x0F,0x12},{0x01,0x00}},	// fls_afl_FlashWP_Weight  flash NB default
{{0x0F,0x12},{0x00,0xC0}},	                                           
{{0x0F,0x12},{0x00,0x80}},	                                           
{{0x0F,0x12},{0x00,0x0A}},	                                           
{{0x0F,0x12},{0x00,0x00}},	                                           
{{0x0F,0x12},{0x00,0x30}},	// fls_afl_FlashWP_Weight  flash NB default
{{0x0F,0x12},{0x00,0x40}},	
{{0x0F,0x12},{0x00,0x48}},	
{{0x0F,0x12},{0x00,0x50}},	
{{0x0F,0x12},{0x00,0x60}},	
{{0x00,0x2A},{0x47,0x84}},	
{{0x0F,0x12},{0x00,0xA0}},	// TNP_Regs_FlsWeightRIn  weight tune start in    
{{0x0F,0x12},{0x00,0xC0}},	     
{{0x0F,0x12},{0x00,0xD0}},	    
{{0x0F,0x12},{0x01,0x00}},	      
{{0x0F,0x12},{0x02,0x00}},	     
{{0x0F,0x12},{0x03,0x00}},	     
{{0x0F,0x12},{0x00,0x88}},	// TNP_Regs_FlsWeightROut  weight tune start out  
{{0x0F,0x12},{0x00,0xB0}},	      
{{0x0F,0x12},{0x00,0xC0}},	     
{{0x0F,0x12},{0x01,0x00}},	    
{{0x0F,0x12},{0x02,0x00}},	  
{{0x0F,0x12},{0x03,0x00}},	 
{{0x00,0x2A},{0x47,0x9C}},	   
{{0x0F,0x12},{0x01,0x20}},	//Fls  BRIn
{{0x0F,0x12},{0x01,0x50}},	      
{{0x0F,0x12},{0x02,0x00}},	     
{{0x0F,0x12},{0x00,0x3C}},	// Fls  BROut
{{0x0F,0x12},{0x00,0x3B}},	    
{{0x0F,0x12},{0x00,0x26}},	//brightness

//==================================================================================
// 15.CCM Setting
//==================================================================================

{{0x00,0x2A},{0x08,0xA6}},	
{{0x0F,0x12},{0x00,0xC0}},	//SARR_AwbCcmCord[0]
{{0x0F,0x12},{0x01,0x00}},	//SARR_AwbCcmCord[1]
{{0x0F,0x12},{0x01,0x25}},	//SARR_AwbCcmCord[2]
{{0x0F,0x12},{0x01,0x5F}},	//SARR_AwbCcmCord[3]
{{0x0F,0x12},{0x01,0x7C}},	//SARR_AwbCcmCord[4]
{{0x0F,0x12},{0x01,0x94}},	//SARR_AwbCcmCord[5]
	
{{0x00,0x2A},{0x08,0x98}},	
{{0x0F,0x12},{0x48,0x00}},	//TVAR_wbt_pBaseCcms
{{0x0F,0x12},{0x70,0x00}},	
{{0x00,0x2A},{0x08,0xA0}},	
{{0x0F,0x12},{0x48,0xD8}},	//TVAR_wbt_pOutdoorCcm
{{0x0F,0x12},{0x70,0x00}},	
	
//Horizon	
{{0x00,0x2A},{0x48,0x00}},	
{{0x0F,0x12},{0x01,0xEB}}, //0210 	//H TVAR_wbt_pBaseCcms[0]  
{{0x0F,0x12},{0xFF,0xD1}}, //FFC0 	//TVAR_wbt_pBaseCcms[1]  
{{0x0F,0x12},{0xFF,0xDC}}, //FFCC 	//TVAR_wbt_pBaseCcms[2]  
{{0x0F,0x12},{0xFF,0x2B}}, //FF34 	//TVAR_wbt_pBaseCcms[3]  
{{0x0F,0x12},{0x02,0x00}}, //01EB 	//TVAR_wbt_pBaseCcms[4]  
{{0x0F,0x12},{0xFE,0xFD}}, //FF09 	//TVAR_wbt_pBaseCcms[5]  
{{0x0F,0x12},{0x00,0x1C}}, //001C 	//TVAR_wbt_pBaseCcms[6]  
{{0x0F,0x12},{0xFF,0xE4}}, //FFE4 	//TVAR_wbt_pBaseCcms[7]  
{{0x0F,0x12},{0x01,0xCD}}, //01CD 	//TVAR_wbt_pBaseCcms[8]  
{{0x0F,0x12},{0x00,0xE3}}, //00D9 	//TVAR_wbt_pBaseCcms[9]  
{{0x0F,0x12},{0x00,0xE3}}, //00D9 	//TVAR_wbt_pBaseCcms[10] 
{{0x0F,0x12},{0xFE,0x80}}, //FE95 	//TVAR_wbt_pBaseCcms[11] 
{{0x0F,0x12},{0x02,0x08}}, //0208 	//TVAR_wbt_pBaseCcms[12] 
{{0x0F,0x12},{0xFF,0x75}}, //FF75 	//TVAR_wbt_pBaseCcms[13] 
{{0x0F,0x12},{0x01,0x99}}, //0199 	//TVAR_wbt_pBaseCcms[14] 
{{0x0F,0x12},{0xFF,0x06}}, //FF06 	//TVAR_wbt_pBaseCcms[15] 
{{0x0F,0x12},{0x01,0xBA}}, //01BA 	//TVAR_wbt_pBaseCcms[16] 
{{0x0F,0x12},{0x01,0x08}}, //0108 	//TVAR_wbt_pBaseCcms[17] 
	
// INCA A                               	
{{0x0F,0x12},{0x01,0xEB}}, //0210 	//A TVAR_wbt_pBaseCcms[18] 
{{0x0F,0x12},{0xFF,0xD1}}, //FFC0 	//TVAR_wbt_pBaseCcms[19] 
{{0x0F,0x12},{0xFF,0xDC}}, //FFCC 	//TVAR_wbt_pBaseCcms[20] 
{{0x0F,0x12},{0xFF,0x2B}}, //FF34 	//TVAR_wbt_pBaseCcms[21] 
{{0x0F,0x12},{0x02,0x00}}, //01EB 	//TVAR_wbt_pBaseCcms[22] 
{{0x0F,0x12},{0xFE,0xFD}}, //FF09 	//TVAR_wbt_pBaseCcms[23] 
{{0x0F,0x12},{0x00,0x1C}}, //001C 	//TVAR_wbt_pBaseCcms[24] 
{{0x0F,0x12},{0xFF,0xE4}}, //FFE4 	//TVAR_wbt_pBaseCcms[25] 
{{0x0F,0x12},{0x01,0xCD}}, //01CD 	//TVAR_wbt_pBaseCcms[26] 
{{0x0F,0x12},{0x00,0xE3}}, //00D9 	//TVAR_wbt_pBaseCcms[27] 
{{0x0F,0x12},{0x00,0xE3}}, //00D9 	//TVAR_wbt_pBaseCcms[28] 
{{0x0F,0x12},{0xFE,0x80}}, //FE95 	//TVAR_wbt_pBaseCcms[29] 
{{0x0F,0x12},{0x02,0x08}}, //0208 	//TVAR_wbt_pBaseCcms[30] 
{{0x0F,0x12},{0xFF,0x75}}, //FF75 	//TVAR_wbt_pBaseCcms[31] 
{{0x0F,0x12},{0x01,0x99}}, //0199 	//TVAR_wbt_pBaseCcms[32] 
{{0x0F,0x12},{0xFF,0x06}}, //FF06 	//TVAR_wbt_pBaseCcms[33] 
{{0x0F,0x12},{0x01,0xBA}}, //01BA 	//TVAR_wbt_pBaseCcms[34] 
{{0x0F,0x12},{0x01,0x08}}, //0108 	//TVAR_wbt_pBaseCcms[35] 
	
//Warm White                    	
{{0x0F,0x12},{0x02,0x2B}}, //0230 //0243 //023E //023B //0231 //0227 	//CW TVAR_wbt_pBaseCcms[36] 
{{0x0F,0x12},{0xFF,0x8C}}, //FF9E //FF8E //FF8A //FF8D //FF93 //FF95 	//TVAR_wbt_pBaseCcms[37] 
{{0x0F,0x12},{0xFF,0xD8}}, //FFC5 //FFC3 //FFCC //FFCF //FFD4 //FFE0 	//TVAR_wbt_pBaseCcms[38] 
{{0x0F,0x12},{0xFE,0xCA}}, //FECC //FECE //FED4 //FED2 	//TVAR_wbt_pBaseCcms[39] 
{{0x0F,0x12},{0x01,0x92}}, //018D //0189 //017F //0174 	//TVAR_wbt_pBaseCcms[40] 
{{0x0F,0x12},{0xFF,0x0E}}, //FF10 //FF12 //FF16 //FF1F 	//TVAR_wbt_pBaseCcms[41] 
{{0x0F,0x12},{0xFF,0xD2}}, //FFD2 //FFD2 //FFD2 //FFCB 	//TVAR_wbt_pBaseCcms[42] 
{{0x0F,0x12},{0xFF,0xD2}}, //FFD2 //FFD2 //FFD2 //FFCB 	//TVAR_wbt_pBaseCcms[43] 
{{0x0F,0x12},{0x01,0xC8}}, //01C8 //01C8 //01C8 //01D6 	//TVAR_wbt_pBaseCcms[44] 
{{0x0F,0x12},{0x00,0xF9}}, //00F9 //00F9 //00F9 //00F8 	//TVAR_wbt_pBaseCcms[45] 
{{0x0F,0x12},{0x00,0xDE}}, //00DE //00DE //00DE //00D5 	//TVAR_wbt_pBaseCcms[46] 
{{0x0F,0x12},{0xFF,0x26}}, //FF26 //FF26 //FF26 //FF34 	//TVAR_wbt_pBaseCcms[47] 
{{0x0F,0x12},{0x01,0x51}}, //0151 //0151 //0151 //0151 	//TVAR_wbt_pBaseCcms[48] 
{{0x0F,0x12},{0xFF,0x50}}, //FF50 //FF50 //FF50 //FF50 	//TVAR_wbt_pBaseCcms[49] 
{{0x0F,0x12},{0x01,0x47}}, //0147 //0147 //0147 //0147 	//TVAR_wbt_pBaseCcms[50] 
{{0x0F,0x12},{0xFF,0x86}}, //FF86 //FF86 //FF86 //FF86 //FF75 	//TVAR_wbt_pBaseCcms[51] 
{{0x0F,0x12},{0x01,0xB4}}, //01B4 //01B4 //01B4 //01B4 //0187 	//TVAR_wbt_pBaseCcms[52] 
{{0x0F,0x12},{0x01,0x7E}}, //017E //017E //017E //017E //01BF 	//TVAR_wbt_pBaseCcms[53] 
         	
//Cool White	
{{0x0F,0x12},{0x02,0x2B}}, //0230 //0243 //023E //023B //0231 //0227 	//CW TVAR_wbt_pBaseCcms[54] 
{{0x0F,0x12},{0xFF,0x8C}}, //FF9E //FF8E //FF8A //FF8D //FF93 //FF95 	//TVAR_wbt_pBaseCcms[55] 
{{0x0F,0x12},{0xFF,0xD8}}, //FFC5 //FFC3 //FFCC //FFCF //FFD4 //FFE0 	//TVAR_wbt_pBaseCcms[56] 
{{0x0F,0x12},{0xFE,0xCA}}, //FECC //FECE //FED4 //FED2 	//TVAR_wbt_pBaseCcms[57] 
{{0x0F,0x12},{0x01,0x92}}, //018D //0189 //017F //0174 	//TVAR_wbt_pBaseCcms[58] 
{{0x0F,0x12},{0xFF,0x0E}}, //FF10 //FF12 //FF16 //FF1F 	//TVAR_wbt_pBaseCcms[59] 
{{0x0F,0x12},{0xFF,0xD2}}, //FFD2 //FFD2 //FFD2 //FFCB 	//TVAR_wbt_pBaseCcms[60] 
{{0x0F,0x12},{0xFF,0xD2}}, //FFD2 //FFD2 //FFD2 //FFCB 	//TVAR_wbt_pBaseCcms[61] 
{{0x0F,0x12},{0x01,0xC8}}, //01C8 //01C8 //01C8 //01D6 	//TVAR_wbt_pBaseCcms[62] 
{{0x0F,0x12},{0x00,0xF9}}, //00F9 //00F9 //00F9 //00F8 	//TVAR_wbt_pBaseCcms[63] 
{{0x0F,0x12},{0x00,0xDE}}, //00DE //00DE //00DE //00D5 	//TVAR_wbt_pBaseCcms[64] 
{{0x0F,0x12},{0xFF,0x26}}, //FF26 //FF26 //FF26 //FF34 	//TVAR_wbt_pBaseCcms[65] 
{{0x0F,0x12},{0x01,0x51}}, //0151 //0151 //0151 //0151 	//TVAR_wbt_pBaseCcms[66] 
{{0x0F,0x12},{0xFF,0x50}}, //FF50 //FF50 //FF50 //FF50 	//TVAR_wbt_pBaseCcms[67] 
{{0x0F,0x12},{0x01,0x47}}, //0147 //0147 //0147 //0147 	//TVAR_wbt_pBaseCcms[68] 
{{0x0F,0x12},{0xFF,0x86}}, //FF86 //FF86 //FF86 //FF86 //FF75 	//TVAR_wbt_pBaseCcms[69] 
{{0x0F,0x12},{0x01,0xB4}}, //01B4 //01B4 //01B4 //01B4 //0187 	//TVAR_wbt_pBaseCcms[70] 
{{0x0F,0x12},{0x01,0x7E}}, //017E //017E //017E //017E //01BF 	//TVAR_wbt_pBaseCcms[71] 
         	
//D50	
{{0x0F,0x12},{0x01,0xEC}}, 	//TVAR_wbt_pBaseCcms[72]  
{{0x0F,0x12},{0xFF,0xB2}}, 	//TVAR_wbt_pBaseCcms[73] 
{{0x0F,0x12},{0x00,0x07}}, 	//TVAR_wbt_pBaseCcms[74] 
{{0x0F,0x12},{0xFE,0xE1}}, 	//TVAR_wbt_pBaseCcms[75] 
{{0x0F,0x12},{0x01,0x3A}}, 	//TVAR_wbt_pBaseCcms[76] 
{{0x0F,0x12},{0xFF,0x3C}}, 	//TVAR_wbt_pBaseCcms[77] 
{{0x0F,0x12},{0xFF,0xDB}}, 	//TVAR_wbt_pBaseCcms[78] 
{{0x0F,0x12},{0xFF,0xDB}}, 	//TVAR_wbt_pBaseCcms[79] 
{{0x0F,0x12},{0x01,0xBE}}, 	//TVAR_wbt_pBaseCcms[80] 
{{0x0F,0x12},{0x00,0xD3}}, 	//TVAR_wbt_pBaseCcms[81] 
{{0x0F,0x12},{0x00,0xE4}}, 	//TVAR_wbt_pBaseCcms[82] 
{{0x0F,0x12},{0xFF,0x4A}}, 	//TVAR_wbt_pBaseCcms[83] 
{{0x0F,0x12},{0x01,0x51}}, 	//TVAR_wbt_pBaseCcms[84] 
{{0x0F,0x12},{0xFF,0x50}}, 	//TVAR_wbt_pBaseCcms[85] 
{{0x0F,0x12},{0x01,0x47}}, 	//TVAR_wbt_pBaseCcms[86] 
{{0x0F,0x12},{0xFF,0x86}}, //FF74 //FF75 	//TVAR_wbt_pBaseCcms[87] 
{{0x0F,0x12},{0x01,0xB4}}, //01A2 //0187 	//TVAR_wbt_pBaseCcms[88] 
{{0x0F,0x12},{0x01,0x7E}}, //01A2 //01BF 	//TVAR_wbt_pBaseCcms[89] 
         		
//D65     	
{{0x0F,0x12},{0x01,0xEC}}, 	//d65 TVAR_wbt_pBaseCcms[90]  
{{0x0F,0x12},{0xFF,0xB2}}, 	//TVAR_wbt_pBaseCcms[91] 
{{0x0F,0x12},{0x00,0x07}}, 	//TVAR_wbt_pBaseCcms[92] 
{{0x0F,0x12},{0xFE,0xE1}}, 	//TVAR_wbt_pBaseCcms[93] 
{{0x0F,0x12},{0x01,0x3A}}, 	//TVAR_wbt_pBaseCcms[94] 
{{0x0F,0x12},{0xFF,0x3C}}, 	//TVAR_wbt_pBaseCcms[95] 
{{0x0F,0x12},{0xFF,0xDB}}, 	//TVAR_wbt_pBaseCcms[96] 
{{0x0F,0x12},{0xFF,0xDB}}, 	//TVAR_wbt_pBaseCcms[97] 
{{0x0F,0x12},{0x01,0xBE}}, 	//TVAR_wbt_pBaseCcms[98] 
{{0x0F,0x12},{0x00,0xD3}}, 	//TVAR_wbt_pBaseCcms[99] 
{{0x0F,0x12},{0x00,0xE4}}, 	//TVAR_wbt_pBaseCcms[100]
{{0x0F,0x12},{0xFF,0x4A}}, 	//TVAR_wbt_pBaseCcms[101]
{{0x0F,0x12},{0x01,0x51}}, 	//TVAR_wbt_pBaseCcms[102]
{{0x0F,0x12},{0xFF,0x50}}, 	//TVAR_wbt_pBaseCcms[103]
{{0x0F,0x12},{0x01,0x47}}, 	//TVAR_wbt_pBaseCcms[104]
{{0x0F,0x12},{0xFF,0x86}}, //FF74 //FF75 	//TVAR_wbt_pBaseCcms[105]
{{0x0F,0x12},{0x01,0xB4}}, //01A2 //0187 	//TVAR_wbt_pBaseCcms[106]
{{0x0F,0x12},{0x01,0x7E}}, //01A2 //01BF 	//TVAR_wbt_pBaseCcms[107]
	  	
//Outdoor		
{{0x00,0x2A},{0x48,0xd8}},		
{{0x0F,0x12},{0x01,0xDC}}, //01D2 //01D4 	//TVAR_wbt_pOutdoorCcm[0] 
{{0x0F,0x12},{0xFF,0x91}}, //FF97 //FF9D 	//TVAR_wbt_pOutdoorCcm[1] 
{{0x0F,0x12},{0xFF,0xE9}}, //FFED //FFE8 	//TVAR_wbt_pOutdoorCcm[2] 
{{0x0F,0x12},{0xFE,0xB5}}, //FEA9 	//TVAR_wbt_pOutdoorCcm[3] 
{{0x0F,0x12},{0x01,0x3A}}, //0143 	//TVAR_wbt_pOutdoorCcm[4] 
{{0x0F,0x12},{0xFE,0xFF}}, //FF01 	//TVAR_wbt_pOutdoorCcm[5] 
{{0x0F,0x12},{0xFF,0xB7}}, //FFB7 	//TVAR_wbt_pOutdoorCcm[6] 
{{0x0F,0x12},{0xFF,0xF5}}, //FFF5 	//TVAR_wbt_pOutdoorCcm[7] 
{{0x0F,0x12},{0x02,0x37}}, //0237 	//TVAR_wbt_pOutdoorCcm[8] 
{{0x0F,0x12},{0x00,0xD5}}, //00B9 	//TVAR_wbt_pOutdoorCcm[9] 
{{0x0F,0x12},{0x01,0x01}}, //0126 	//TVAR_wbt_pOutdoorCcm[10]
{{0x0F,0x12},{0xFF,0x39}}, //FF33 	//TVAR_wbt_pOutdoorCcm[11]
{{0x0F,0x12},{0x01,0xCE}}, //01CE 	//TVAR_wbt_pOutdoorCcm[12]
{{0x0F,0x12},{0xFF,0x83}}, //FF83 	//TVAR_wbt_pOutdoorCcm[13]
{{0x0F,0x12},{0x01,0x95}}, //0195 	//TVAR_wbt_pOutdoorCcm[14]
{{0x0F,0x12},{0xFE,0xF3}}, //FEF3 	//TVAR_wbt_pOutdoorCcm[15]
{{0x0F,0x12},{0x01,0x4F}}, //014F 	//TVAR_wbt_pOutdoorCcm[16]
{{0x0F,0x12},{0x01,0x37}}, //0137 	//TVAR_wbt_pOutdoorCcm[17]

//==================================================================================
// 16.GAMMA
//==================================================================================

{{0x00,0x2A},{0x07,0x34}},	
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000 //0000	//saRR_usDualGammaLutRGBIndoor[0][0] 
{{0x0F,0x12},{0x00,0x08}}, //0004 //0004 //000A	//saRR_usDualGammaLutRGBIndoor[0][1] 
{{0x0F,0x12},{0x00,0x10}}, //0010 //0010 //0016	//saRR_usDualGammaLutRGBIndoor[0][2] 
{{0x0F,0x12},{0x00,0x28}}, //0020 //0020 //0030	//saRR_usDualGammaLutRGBIndoor[0][3] 
{{0x0F,0x12},{0x00,0x58}}, //0050 //0060 //0066	//saRR_usDualGammaLutRGBIndoor[0][4] 
{{0x0F,0x12},{0x00,0xC8}}, //00C0 //00D5 //00D5	//saRR_usDualGammaLutRGBIndoor[0][5] 
{{0x0F,0x12},{0x01,0x38}}, //0138 //0138 //0138	//saRR_usDualGammaLutRGBIndoor[0][6] 
{{0x0F,0x12},{0x01,0x63}}, //0163 //0163 //0163	//saRR_usDualGammaLutRGBIndoor[0][7] 
{{0x0F,0x12},{0x01,0x89}}, //0189 //0189 //0189	//saRR_usDualGammaLutRGBIndoor[0][8] 
{{0x0F,0x12},{0x01,0xC6}}, //01C6 //01C6 //01C6	//saRR_usDualGammaLutRGBIndoor[0][9] 
{{0x0F,0x12},{0x01,0xF8}}, //01F8 //01F8 //01F8	//saRR_usDualGammaLutRGBIndoor[0][10]
{{0x0F,0x12},{0x02,0x22}}, //0222 //0222 //0222	//saRR_usDualGammaLutRGBIndoor[0][11]
{{0x0F,0x12},{0x02,0x47}}, //0247 //0247 //0247	//saRR_usDualGammaLutRGBIndoor[0][12]
{{0x0F,0x12},{0x02,0x88}}, //0288 //0288 //0282	//saRR_usDualGammaLutRGBIndoor[0][13]
{{0x0F,0x12},{0x02,0xC0}}, //02C0 //02C0 //02B5	//saRR_usDualGammaLutRGBIndoor[0][14]
{{0x0F,0x12},{0x03,0x18}}, //0318 //0318 //030F	//saRR_usDualGammaLutRGBIndoor[0][15]
{{0x0F,0x12},{0x03,0x5F}}, //035F //035F //035F	//saRR_usDualGammaLutRGBIndoor[0][16]
{{0x0F,0x12},{0x03,0xA2}}, //03A2 //03A2 //03A2	//saRR_usDualGammaLutRGBIndoor[0][17]
{{0x0F,0x12},{0x03,0xD8}}, //03D8 //03D8 //03D8	//saRR_usDualGammaLutRGBIndoor[0][18]
{{0x0F,0x12},{0x03,0xFF}}, //03FF //03FF //03FF	//saRR_usDualGammaLutRGBIndoor[0][19]
	
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000 //0000	//saRR_usDualGammaLutRGBIndoor[1][0] 
{{0x0F,0x12},{0x00,0x08}}, //0004 //0004 //000A	//saRR_usDualGammaLutRGBIndoor[1][1] 
{{0x0F,0x12},{0x00,0x10}}, //0010 //0010 //0016	//saRR_usDualGammaLutRGBIndoor[1][2] 
{{0x0F,0x12},{0x00,0x28}}, //0020 //0020 //0030	//saRR_usDualGammaLutRGBIndoor[1][3] 
{{0x0F,0x12},{0x00,0x58}}, //0050 //0060 //0066	//saRR_usDualGammaLutRGBIndoor[1][4] 
{{0x0F,0x12},{0x00,0xC8}}, //00C0 //00D5 //00D5	//saRR_usDualGammaLutRGBIndoor[1][5] 
{{0x0F,0x12},{0x01,0x38}}, //0138 //0138 //0138	//saRR_usDualGammaLutRGBIndoor[1][6] 
{{0x0F,0x12},{0x01,0x63}}, //0163 //0163 //0163	//saRR_usDualGammaLutRGBIndoor[1][7] 
{{0x0F,0x12},{0x01,0x89}}, //0189 //0189 //0189	//saRR_usDualGammaLutRGBIndoor[1][8] 
{{0x0F,0x12},{0x01,0xC6}}, //01C6 //01C6 //01C6	//saRR_usDualGammaLutRGBIndoor[1][9] 
{{0x0F,0x12},{0x01,0xF8}}, //01F8 //01F8 //01F8	//saRR_usDualGammaLutRGBIndoor[1][10]
{{0x0F,0x12},{0x02,0x22}}, //0222 //0222 //0222	//saRR_usDualGammaLutRGBIndoor[1][11]
{{0x0F,0x12},{0x02,0x47}}, //0247 //0247 //0247	//saRR_usDualGammaLutRGBIndoor[1][12]
{{0x0F,0x12},{0x02,0x88}}, //0288 //0288 //0282	//saRR_usDualGammaLutRGBIndoor[1][13]
{{0x0F,0x12},{0x02,0xC0}}, //02C0 //02C0 //02B5	//saRR_usDualGammaLutRGBIndoor[1][14]
{{0x0F,0x12},{0x03,0x18}}, //0318 //0318 //030F	//saRR_usDualGammaLutRGBIndoor[1][15]
{{0x0F,0x12},{0x03,0x5F}}, //035F //035F //035F	//saRR_usDualGammaLutRGBIndoor[1][16]
{{0x0F,0x12},{0x03,0xA2}}, //03A2 //03A2 //03A2	//saRR_usDualGammaLutRGBIndoor[1][17]
{{0x0F,0x12},{0x03,0xD8}}, //03D8 //03D8 //03D8	//saRR_usDualGammaLutRGBIndoor[1][18]
{{0x0F,0x12},{0x03,0xFF}}, //03FF //03FF //03FF	//saRR_usDualGammaLutRGBIndoor[1][19]
	
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000 //0000	//saRR_usDualGammaLutRGBIndoor[2][0] 
{{0x0F,0x12},{0x00,0x08}}, //0004 //0004 //000A	//saRR_usDualGammaLutRGBIndoor[2][1] 
{{0x0F,0x12},{0x00,0x10}}, //0010 //0010 //0016	//saRR_usDualGammaLutRGBIndoor[2][2] 
{{0x0F,0x12},{0x00,0x28}}, //0020 //0020 //0030	//saRR_usDualGammaLutRGBIndoor[2][3] 
{{0x0F,0x12},{0x00,0x58}}, //0050 //0060 //0066	//saRR_usDualGammaLutRGBIndoor[2][4] 
{{0x0F,0x12},{0x00,0xC8}}, //00C0 //00D5 //00D5	//saRR_usDualGammaLutRGBIndoor[2][5] 
{{0x0F,0x12},{0x01,0x38}}, //0138 //0138 //0138	//saRR_usDualGammaLutRGBIndoor[2][6] 
{{0x0F,0x12},{0x01,0x63}}, //0163 //0163 //0163	//saRR_usDualGammaLutRGBIndoor[2][7] 
{{0x0F,0x12},{0x01,0x89}}, //0189 //0189 //0189	//saRR_usDualGammaLutRGBIndoor[2][8] 
{{0x0F,0x12},{0x01,0xC6}}, //01C6 //01C6 //01C6	//saRR_usDualGammaLutRGBIndoor[2][9] 
{{0x0F,0x12},{0x01,0xF8}}, //01F8 //01F8 //01F8	//saRR_usDualGammaLutRGBIndoor[2][10]
{{0x0F,0x12},{0x02,0x22}}, //0222 //0222 //0222	//saRR_usDualGammaLutRGBIndoor[2][11]
{{0x0F,0x12},{0x02,0x47}}, //0247 //0247 //0247	//saRR_usDualGammaLutRGBIndoor[2][12]
{{0x0F,0x12},{0x02,0x88}}, //0288 //0288 //0282	//saRR_usDualGammaLutRGBIndoor[2][13]
{{0x0F,0x12},{0x02,0xC0}}, //02C0 //02C0 //02B5	//saRR_usDualGammaLutRGBIndoor[2][14]
{{0x0F,0x12},{0x03,0x18}}, //0318 //0318 //030F	//saRR_usDualGammaLutRGBIndoor[2][15]
{{0x0F,0x12},{0x03,0x5F}}, //035F //035F //035F	//saRR_usDualGammaLutRGBIndoor[2][16]
{{0x0F,0x12},{0x03,0xA2}}, //03A2 //03A2 //03A2	//saRR_usDualGammaLutRGBIndoor[2][17]
{{0x0F,0x12},{0x03,0xD8}}, //03D8 //03D8 //03D8	//saRR_usDualGammaLutRGBIndoor[2][18]
{{0x0F,0x12},{0x03,0xFF}}, //03FF //03FF //03FF	//saRR_usDualGammaLutRGBIndoor[2][19]
	
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000 //0000 //0000 //0000 //0000	//saRR_usDualGammaLutRGBOutdoor[0][0] 
{{0x0F,0x12},{0x00,0x0B}}, //0004 //0004 //0004 //000C //000B //000B	//saRR_usDualGammaLutRGBOutdoor[0][1] 
{{0x0F,0x12},{0x00,0x19}}, //0010 //0010 //0010 //0020 //0019 //0019	//saRR_usDualGammaLutRGBOutdoor[0][2] 
{{0x0F,0x12},{0x00,0x36}}, //0020 //0020 //0020 //0050 //0048 //0036	//saRR_usDualGammaLutRGBOutdoor[0][3] 
{{0x0F,0x12},{0x00,0x6F}}, //0060 //0060 //0060 //0098 //008C //006F	//saRR_usDualGammaLutRGBOutdoor[0][4] 
{{0x0F,0x12},{0x00,0xD8}}, //00E4 //00F0 //00D8 //00FC //00F0 //00D8	//saRR_usDualGammaLutRGBOutdoor[0][5] 
{{0x0F,0x12},{0x01,0x35}}, //0158 //0168 //0148 //0148 //0140 //0135	//saRR_usDualGammaLutRGBOutdoor[0][6] 
{{0x0F,0x12},{0x01,0x5F}}, //0180 //0190 //0174 //0168 //0164 //015F	//saRR_usDualGammaLutRGBOutdoor[0][7] 
{{0x0F,0x12},{0x01,0x85}}, //01A8 //01B8 //0198 //0185 //0185 //0185	//saRR_usDualGammaLutRGBOutdoor[0][8] 
{{0x0F,0x12},{0x01,0xC1}}, //01E8 //01FC //01D4 //01C1 //01C1 //01C1	//saRR_usDualGammaLutRGBOutdoor[0][9] 
{{0x0F,0x12},{0x01,0xF3}}, //021C //0230 //0208 //01F3 //01F3 //01F3	//saRR_usDualGammaLutRGBOutdoor[0][10]
{{0x0F,0x12},{0x02,0x20}}, //0248 //0260 //0234 //0220 //0220 //0220	//saRR_usDualGammaLutRGBOutdoor[0][11]
{{0x0F,0x12},{0x02,0x4A}}, //0274 //0288 //0260 //024A //024A //024A	//saRR_usDualGammaLutRGBOutdoor[0][12]
{{0x0F,0x12},{0x02,0x91}}, //02C0 //02D0 //02A8 //0291 //0291 //0291	//saRR_usDualGammaLutRGBOutdoor[0][13]
{{0x0F,0x12},{0x02,0xD0}}, //02FC //030C //02E4 //02D0 //02D0 //02D0	//saRR_usDualGammaLutRGBOutdoor[0][14]
{{0x0F,0x12},{0x03,0x2A}}, //0358 //0360 //0340 //032A //032A //032A	//saRR_usDualGammaLutRGBOutdoor[0][15]
{{0x0F,0x12},{0x03,0x6A}}, //0394 //0398 //0380 //036A //036A //036A	//saRR_usDualGammaLutRGBOutdoor[0][16]
{{0x0F,0x12},{0x03,0x9F}}, //03BC //03C0 //03B0 //039F //039F //039F	//saRR_usDualGammaLutRGBOutdoor[0][17]
{{0x0F,0x12},{0x03,0xCC}}, //03DC //03E0 //03D8 //03CC //03CC //03CC	//saRR_usDualGammaLutRGBOutdoor[0][18]
{{0x0F,0x12},{0x03,0xF9}}, //03F9 //03F9 //03F9 //03F9 //03F9 //03F9	//saRR_usDualGammaLutRGBOutdoor[0][19]
	
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000 //0000 //0000 //0000 //0000	//saRR_usDualGammaLutRGBOutdoor[1][0] 
{{0x0F,0x12},{0x00,0x0B}}, //0004 //0004 //0004 //000C //000B //000B	//saRR_usDualGammaLutRGBOutdoor[1][1] 
{{0x0F,0x12},{0x00,0x19}}, //0010 //0010 //0010 //0020 //0019 //0019	//saRR_usDualGammaLutRGBOutdoor[1][2] 
{{0x0F,0x12},{0x00,0x36}}, //0020 //0020 //0020 //0050 //0048 //0036	//saRR_usDualGammaLutRGBOutdoor[1][3] 
{{0x0F,0x12},{0x00,0x6F}}, //0060 //0060 //0060 //0098 //008C //006F	//saRR_usDualGammaLutRGBOutdoor[1][4] 
{{0x0F,0x12},{0x00,0xD8}}, //00E4 //00F0 //00D8 //00FC //00F0 //00D8	//saRR_usDualGammaLutRGBOutdoor[1][5] 
{{0x0F,0x12},{0x01,0x35}}, //0158 //0168 //0148 //0148 //0140 //0135	//saRR_usDualGammaLutRGBOutdoor[1][6] 
{{0x0F,0x12},{0x01,0x5F}}, //0180 //0190 //0174 //0168 //0164 //015F	//saRR_usDualGammaLutRGBOutdoor[1][7] 
{{0x0F,0x12},{0x01,0x85}}, //01A8 //01B8 //0198 //0185 //0185 //0185	//saRR_usDualGammaLutRGBOutdoor[1][8] 
{{0x0F,0x12},{0x01,0xC1}}, //01E8 //01FC //01D4 //01C1 //01C1 //01C1	//saRR_usDualGammaLutRGBOutdoor[1][9] 
{{0x0F,0x12},{0x01,0xF3}}, //021C //0230 //0208 //01F3 //01F3 //01F3	//saRR_usDualGammaLutRGBOutdoor[1][10]
{{0x0F,0x12},{0x02,0x20}}, //0248 //0260 //0234 //0220 //0220 //0220	//saRR_usDualGammaLutRGBOutdoor[1][11]
{{0x0F,0x12},{0x02,0x4A}}, //0274 //0288 //0260 //024A //024A //024A	//saRR_usDualGammaLutRGBOutdoor[1][12]
{{0x0F,0x12},{0x02,0x91}}, //02C0 //02D0 //02A8 //0291 //0291 //0291	//saRR_usDualGammaLutRGBOutdoor[1][13]
{{0x0F,0x12},{0x02,0xD0}}, //02FC //030C //02E4 //02D0 //02D0 //02D0	//saRR_usDualGammaLutRGBOutdoor[1][14]
{{0x0F,0x12},{0x03,0x2A}}, //0358 //0360 //0340 //032A //032A //032A	//saRR_usDualGammaLutRGBOutdoor[1][15]
{{0x0F,0x12},{0x03,0x6A}}, //0394 //0398 //0380 //036A //036A //036A	//saRR_usDualGammaLutRGBOutdoor[1][16]
{{0x0F,0x12},{0x03,0x9F}}, //03BC //03C0 //03B0 //039F //039F //039F	//saRR_usDualGammaLutRGBOutdoor[1][17]
{{0x0F,0x12},{0x03,0xCC}}, //03DC //03E0 //03D8 //03CC //03CC //03CC	//saRR_usDualGammaLutRGBOutdoor[1][18]
{{0x0F,0x12},{0x03,0xF9}}, //03F9 //03F9 //03F9 //03F9 //03F9 //03F9	//saRR_usDualGammaLutRGBOutdoor[1][19]
	
{{0x0F,0x12},{0x00,0x00}}, //0000 //0000 //0000 //0000 //0000 //0000	//saRR_usDualGammaLutRGBOutdoor[2][0] 
{{0x0F,0x12},{0x00,0x0B}}, //0004 //0004 //0004 //000C //000B //000B	//saRR_usDualGammaLutRGBOutdoor[2][1] 
{{0x0F,0x12},{0x00,0x19}}, //0010 //0010 //0010 //0020 //0019 //0019	//saRR_usDualGammaLutRGBOutdoor[2][2] 
{{0x0F,0x12},{0x00,0x36}}, //0020 //0020 //0020 //0050 //0048 //0036	//saRR_usDualGammaLutRGBOutdoor[2][3] 
{{0x0F,0x12},{0x00,0x6F}}, //0060 //0060 //0060 //0098 //008C //006F	//saRR_usDualGammaLutRGBOutdoor[2][4] 
{{0x0F,0x12},{0x00,0xD8}}, //00E4 //00F0 //00D8 //00FC //00F0 //00D8	//saRR_usDualGammaLutRGBOutdoor[2][5] 
{{0x0F,0x12},{0x01,0x35}}, //0158 //0168 //0148 //0148 //0140 //0135	//saRR_usDualGammaLutRGBOutdoor[2][6] 
{{0x0F,0x12},{0x01,0x5F}}, //0180 //0190 //0174 //0168 //0164 //015F	//saRR_usDualGammaLutRGBOutdoor[2][7] 
{{0x0F,0x12},{0x01,0x85}}, //01A8 //01B8 //0198 //0185 //0185 //0185	//saRR_usDualGammaLutRGBOutdoor[2][8] 
{{0x0F,0x12},{0x01,0xC1}}, //01E8 //01FC //01D4 //01C1 //01C1 //01C1	//saRR_usDualGammaLutRGBOutdoor[2][9] 
{{0x0F,0x12},{0x01,0xF3}}, //021C //0230 //0208 //01F3 //01F3 //01F3	//saRR_usDualGammaLutRGBOutdoor[2][10]
{{0x0F,0x12},{0x02,0x20}}, //0248 //0260 //0234 //0220 //0220 //0220	//saRR_usDualGammaLutRGBOutdoor[2][11]
{{0x0F,0x12},{0x02,0x4A}}, //0274 //0288 //0260 //024A //024A //024A	//saRR_usDualGammaLutRGBOutdoor[2][12]
{{0x0F,0x12},{0x02,0x91}}, //02C0 //02D0 //02A8 //0291 //0291 //0291	//saRR_usDualGammaLutRGBOutdoor[2][13]
{{0x0F,0x12},{0x02,0xD0}}, //02FC //030C //02E4 //02D0 //02D0 //02D0	//saRR_usDualGammaLutRGBOutdoor[2][14]
{{0x0F,0x12},{0x03,0x2A}}, //0358 //0360 //0340 //032A //032A //032A	//saRR_usDualGammaLutRGBOutdoor[2][15]
{{0x0F,0x12},{0x03,0x6A}}, //0394 //0398 //0380 //036A //036A //036A	//saRR_usDualGammaLutRGBOutdoor[2][16]
{{0x0F,0x12},{0x03,0x9F}}, //03BC //03C0 //03B0 //039F //039F //039F	//saRR_usDualGammaLutRGBOutdoor[2][17]
{{0x0F,0x12},{0x03,0xCC}}, //03DC //03E0 //03D8 //03CC //03CC //03CC	//saRR_usDualGammaLutRGBOutdoor[2][18]
{{0x0F,0x12},{0x03,0xF9}}, //03F9 //03F9 //03F9 //03F9 //03F9 //03F9	//saRR_usDualGammaLutRGBOutdoor[2][19]


//==================================================================================
// 17.AFIT
//==================================================================================
{{0x00,0x2A},{0x09,0x44}},	
{{0x0F,0x12},{0x00,0x50}},	//afit_uNoiseIndInDoor
{{0x0F,0x12},{0x00,0xB0}},	//afit_uNoiseIndInDoor
{{0x0F,0x12},{0x01,0x96}},	//afit_uNoiseIndInDoor
{{0x0F,0x12},{0x02,0x45}},	//afit_uNoiseIndInDoor
{{0x0F,0x12},{0x03,0x00}},	//afit_uNoiseIndInDoor
	
{{0x00,0x2A},{0x09,0x38}},	
{{0x0F,0x12},{0x00,0x00}},	// on/off AFIT by NB option
{{0x0F,0x12},{0x00,0x14}},	//SARR_uNormBrInDoor
{{0x0F,0x12},{0x00,0xD2}},	//SARR_uNormBrInDoor
{{0x0F,0x12},{0x03,0x84}},	//SARR_uNormBrInDoor
{{0x0F,0x12},{0x07,0xD0}},	//SARR_uNormBrInDoor
{{0x0F,0x12},{0x13,0x88}},	//SARR_uNormBrInDoor
	
{{0x00,0x2A},{0x09,0x76}},	
{{0x0F,0x12},{0x00,0x70}},	//afit_usGamutTh       
{{0x0F,0x12},{0x00,0x05}},	//afit_usNeargrayOffset
{{0x0F,0x12},{0x00,0x00}},	//afit_bUseSenBpr         
{{0x0F,0x12},{0x01,0xCC}},	//afit_usBprThr_0_        
{{0x0F,0x12},{0x01,0xCC}},	//afit_usBprThr_1_        
{{0x0F,0x12},{0x01,0xCC}},	//afit_usBprThr_2_        
{{0x0F,0x12},{0x01,0xCC}},	//afit_usBprThr_3_        
{{0x0F,0x12},{0x01,0xCC}},	//afit_usBprThr_4_        
{{0x0F,0x12},{0x01,0x80}},	//afit_NIContrastAFITValue
{{0x0F,0x12},{0x01,0x96}},	//afit_NIContrastTh       
	
{{0x00,0x2A},{0x09,0x8C}},	
{{0x0F,0x12},{0x00,0x00}},	//7000098C//AFIT16_BRIGHTNESS           
{{0x0F,0x12},{0x00,0x00}},	//7000098E//AFIT16_CONTRAST                                                        
{{0x0F,0x12},{0x00,0x00}},	//70000990//AFIT16_SATURATION                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000992//AFIT16_SHARP_BLUR                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000994//AFIT16_GLAMOUR                                                         
{{0x0F,0x12},{0x00,0xC0}},	//70000996//AFIT16_bnr_edge_high                                                   
{{0x0F,0x12},{0x00,0x64}},	//70000998//AFIT16_postdmsc_iLowBright                                             
{{0x0F,0x12},{0x03,0x84}},	//7000099A//AFIT16_postdmsc_iHighBright                                            
{{0x0F,0x12},{0x00,0x5F}},	//7000099C//AFIT16_postdmsc_iLowSat                                                
{{0x0F,0x12},{0x01,0xF4}},	//7000099E//AFIT16_postdmsc_iHighSat                                               
{{0x0F,0x12},{0x00,0x70}},	//700009A0//AFIT16_postdmsc_iTune                                                  
{{0x0F,0x12},{0x00,0x40}},	//700009A2//AFIT16_yuvemix_mNegRanges_0                                            
{{0x0F,0x12},{0x00,0xA0}},	//700009A4//AFIT16_yuvemix_mNegRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//700009A6//AFIT16_yuvemix_mNegRanges_2                                            
{{0x0F,0x12},{0x00,0x10}},	//700009A8//AFIT16_yuvemix_mPosRanges_0                                            
{{0x0F,0x12},{0x00,0x40}},	//700009AA//AFIT16_yuvemix_mPosRanges_1                                            
{{0x0F,0x12},{0x00,0xA0}},	//700009AC//AFIT16_yuvemix_mPosRanges_2                                            
{{0x0F,0x12},{0x14,0x30}},	//700009AE//AFIT8_bnr_edge_low  [7:0] AFIT8_bnr_repl_thresh                        
{{0x0F,0x12},{0x02,0x01}},	//700009B0//AFIT8_bnr_repl_force  [7:0] AFIT8_bnr_iHotThreshHigh                   
{{0x0F,0x12},{0x02,0x04}},	//700009B2//AFIT8_bnr_iHotThreshLow   [7:0] AFIT8_bnr_iColdThreshHigh              
{{0x0F,0x12},{0x36,0x04}},	//700009B4//AFIT8_bnr_iColdThreshLow   [7:0] AFIT8_bnr_DispTH_Low                  
{{0x0F,0x12},{0x03,0x2A}},	//700009B6//AFIT8_bnr_DispTH_High   [7:0] AFIT8_bnr_DISP_Limit_Low                 
{{0x0F,0x12},{0x04,0x03}},	//700009B8//AFIT8_bnr_DISP_Limit_High   [7:0] AFIT8_bnr_iDistSigmaMin              
{{0x0F,0x12},{0x1B,0x06}},	//700009BA//AFIT8_bnr_iDistSigmaMax   [7:0] AFIT8_bnr_iDiffSigmaLow                
{{0x0F,0x12},{0x60,0x15}},	//700009BC//AFIT8_bnr_iDiffSigmaHigh   [7:0] AFIT8_bnr_iNormalizedSTD_TH           
{{0x0F,0x12},{0x00,0xC0}},	//700009BE//AFIT8_bnr_iNormalizedSTD_Limit   [7:0] AFIT8_bnr_iDirNRTune            
{{0x0F,0x12},{0x60,0x80}},	//700009C0//AFIT8_bnr_iDirMinThres   [7:0] AFIT8_bnr_iDirFltDiffThresHigh          
{{0x0F,0x12},{0x40,0x80}},	//700009C2//AFIT8_bnr_iDirFltDiffThresLow   [7:0] AFIT8_bnr_iDirSmoothPowerHigh    
{{0x0F,0x12},{0x06,0x40}},	//700009C4//AFIT8_bnr_iDirSmoothPowerLow   [7:0] AFIT8_bnr_iLowMaxSlopeAllowed     
{{0x0F,0x12},{0x03,0x06}},	//700009C6//AFIT8_bnr_iHighMaxSlopeAllowed   [7:0] AFIT8_bnr_iLowSlopeThresh       
{{0x0F,0x12},{0x20,0x03}},	//700009C8//AFIT8_bnr_iHighSlopeThresh   [7:0] AFIT8_bnr_iSlopenessTH              
{{0x0F,0x12},{0xFF,0x01}},	//700009CA//AFIT8_bnr_iSlopeBlurStrength   [7:0] AFIT8_bnr_iSlopenessLimit         
{{0x0F,0x12},{0x00,0x00}},	//700009CC//AFIT8_bnr_AddNoisePower1   [7:0] AFIT8_bnr_AddNoisePower2              
{{0x0F,0x12},{0x04,0x00}},	//700009CE//AFIT8_bnr_iRadialTune   [7:0] AFIT8_bnr_iRadialPower                   
{{0x0F,0x12},{0x36,0x5A}},	//700009D0//AFIT8_bnr_iRadialLimit   [7:0] AFIT8_ee_iFSMagThLow                    
{{0x0F,0x12},{0x10,0x2A}},	//700009D2//AFIT8_ee_iFSMagThHigh   [7:0] AFIT8_ee_iFSVarThLow                     
{{0x0F,0x12},{0x00,0x0B}},	//700009D4//AFIT8_ee_iFSVarThHigh   [7:0] AFIT8_ee_iFSThLow                        
{{0x0F,0x12},{0x06,0x00}},	//700009D6//AFIT8_ee_iFSThHigh   [7:0] AFIT8_ee_iFSmagPower                        
{{0x0F,0x12},{0x5A,0x0F}},	//700009D8//AFIT8_ee_iFSVarCountTh   [7:0] AFIT8_ee_iRadialLimit                   
{{0x0F,0x12},{0x05,0x05}},	//700009DA//AFIT8_ee_iRadialPower   [7:0] AFIT8_ee_iSmoothEdgeSlope                
{{0x0F,0x12},{0x18,0x02}},	//700009DC//AFIT8_ee_iROADThres   [7:0] AFIT8_ee_iROADMaxNR                        
{{0x0F,0x12},{0x00,0x00}},	//700009DE//AFIT8_ee_iROADSubMaxNR   [7:0] AFIT8_ee_iROADSubThres                  
{{0x0F,0x12},{0x20,0x06}},	//700009E0//AFIT8_ee_iROADNeiThres   [7:0] AFIT8_ee_iROADNeiMaxNR                  
{{0x0F,0x12},{0x30,0x28}},	//700009E2//AFIT8_ee_iSmoothEdgeThres   [7:0] AFIT8_ee_iMSharpen                   
{{0x0F,0x12},{0x04,0x18}},	//700009E4//AFIT8_ee_iWSharpen   [7:0] AFIT8_ee_iMShThresh                         
{{0x0F,0x12},{0x01,0x01}},	//700009E6//AFIT8_ee_iWShThresh   [7:0] AFIT8_ee_iReduceNegative                   
{{0x0F,0x12},{0x08,0x00}},	//700009E8//AFIT8_ee_iEmbossCentAdd   [7:0] AFIT8_ee_iShDespeckle                  
{{0x0F,0x12},{0x18,0x04}},	//700009EA//AFIT8_ee_iReduceEdgeThresh   [7:0] AFIT8_dmsc_iEnhThresh               
{{0x0F,0x12},{0x40,0x08}},	//700009EC//AFIT8_dmsc_iDesatThresh   [7:0] AFIT8_dmsc_iDemBlurHigh                
{{0x0F,0x12},{0x05,0x40}},	//700009EE//AFIT8_dmsc_iDemBlurLow   [7:0] AFIT8_dmsc_iDemBlurRange                
{{0x0F,0x12},{0x80,0x06}},	//700009F0//AFIT8_dmsc_iDecisionThresh   [7:0] AFIT8_dmsc_iCentGrad                
{{0x0F,0x12},{0x00,0x20}},	//700009F2//AFIT8_dmsc_iMonochrom   [7:0] AFIT8_dmsc_iGBDenoiseVal                 
{{0x0F,0x12},{0x00,0x00}},	//700009F4//AFIT8_dmsc_iGRDenoiseVal   [7:0] AFIT8_dmsc_iEdgeDesatThrHigh          
{{0x0F,0x12},{0x18,0x00}},	//700009F6//AFIT8_dmsc_iEdgeDesatThrLow   [7:0] AFIT8_dmsc_iEdgeDesat              
{{0x0F,0x12},{0x00,0x00}},	//700009F8//AFIT8_dmsc_iNearGrayDesat   [7:0] AFIT8_dmsc_iEdgeDesatLimit           
{{0x0F,0x12},{0x1E,0x10}},	//700009FA//AFIT8_postdmsc_iBCoeff   [7:0] AFIT8_postdmsc_iGCoeff                  
{{0x0F,0x12},{0x00,0x0B}},	//700009FC//AFIT8_postdmsc_iWideMult   [7:0] AFIT8_yuvemix_mNegSlopes_0            
{{0x0F,0x12},{0x06,0x07}},	//700009FE//AFIT8_yuvemix_mNegSlopes_1   [7:0] AFIT8_yuvemix_mNegSlopes_2          
{{0x0F,0x12},{0x00,0x05}},	//70000A00//AFIT8_yuvemix_mNegSlopes_3   [7:0] AFIT8_yuvemix_mPosSlopes_0          
{{0x0F,0x12},{0x06,0x07}},	//70000A02//AFIT8_yuvemix_mPosSlopes_1   [7:0] AFIT8_yuvemix_mPosSlopes_2          
{{0x0F,0x12},{0x04,0x05}},	//70000A04//AFIT8_yuvemix_mPosSlopes_3   [7:0] AFIT8_yuviirnr_iXSupportY           
{{0x0F,0x12},{0x02,0x05}},	//70000A06//AFIT8_yuviirnr_iXSupportUV   [7:0] AFIT8_yuviirnr_iLowYNorm            
{{0x0F,0x12},{0x03,0x04}},	//70000A08//AFIT8_yuviirnr_iHighYNorm   [7:0] AFIT8_yuviirnr_iLowUVNorm            
{{0x0F,0x12},{0x04,0x09}},	//70000A0A//AFIT8_yuviirnr_iHighUVNorm   [7:0] AFIT8_yuviirnr_iYNormShift          
{{0x0F,0x12},{0x03,0x06}},	//70000A0C//AFIT8_yuviirnr_iUVNormShift   [7:0] AFIT8_yuviirnr_iVertLength_Y       
{{0x0F,0x12},{0x04,0x07}},	//70000A0E//AFIT8_yuviirnr_iVertLength_UV   [7:0] AFIT8_yuviirnr_iDiffThreshL_Y    
{{0x0F,0x12},{0x1C,0x04}},	//70000A10//AFIT8_yuviirnr_iDiffThreshH_Y   [7:0] AFIT8_yuviirnr_iDiffThreshL_UV   
{{0x0F,0x12},{0x02,0x14}},	//70000A12//AFIT8_yuviirnr_iDiffThreshH_UV   [7:0] AFIT8_yuviirnr_iMaxThreshL_Y    
{{0x0F,0x12},{0x10,0x02}},	//70000A14//AFIT8_yuviirnr_iMaxThreshH_Y   [7:0] AFIT8_yuviirnr_iMaxThreshL_UV     
{{0x0F,0x12},{0x06,0x10}},	//70000A16//AFIT8_yuviirnr_iMaxThreshH_UV   [7:0] AFIT8_yuviirnr_iYNRStrengthL     
{{0x0F,0x12},{0x1A,0x02}},	//70000A18//AFIT8_yuviirnr_iYNRStrengthH   [7:0] AFIT8_yuviirnr_iUVNRStrengthL     
{{0x0F,0x12},{0x4A,0x18}},	//70000A1A//AFIT8_yuviirnr_iUVNRStrengthH   [7:0] AFIT8_byr_gras_iShadingPower     
{{0x0F,0x12},{0x00,0x80}},	//70000A1C//AFIT8_RGBGamma2_iLinearity   [7:0] AFIT8_RGBGamma2_iDarkReduce         
{{0x0F,0x12},{0x03,0x48}},	//70000A1E//AFIT8_ccm_oscar_iSaturation   [7:0] AFIT8_RGB2YUV_iYOffset             
{{0x0F,0x12},{0x01,0x80}},	//70000A20//AFIT8_RGB2YUV_iRGBGain   [7:0] AFIT8_bnr_nClustLevel_H                 
{{0x0F,0x12},{0x0A,0x0A}},	//70000A22//AFIT8_bnr_iClustMulT_H   [7:0] AFIT8_bnr_iClustMulT_C                  
{{0x0F,0x12},{0x01,0x01}},	//70000A24//AFIT8_bnr_iClustThresh_H   [7:0] AFIT8_bnr_iClustThresh_C              
{{0x0F,0x12},{0x2A,0x36}},	//70000A26//AFIT8_bnr_iDenThreshLow   [7:0] AFIT8_bnr_iDenThreshHigh               
{{0x0F,0x12},{0x60,0x24}},	//70000A28//AFIT8_ee_iLowSharpPower   [7:0] AFIT8_ee_iHighSharpPower               
{{0x0F,0x12},{0x2A,0x36}},	//70000A2A//AFIT8_ee_iLowShDenoise   [7:0] AFIT8_ee_iHighShDenoise                 
{{0x0F,0x12},{0xFF,0xFF}},	//70000A2C//AFIT8_ee_iLowSharpClamp   [7:0] AFIT8_ee_iHighSharpClamp               
{{0x0F,0x12},{0x08,0x08}},	//70000A2E//AFIT8_ee_iReduceEdgeMinMult   [7:0] AFIT8_ee_iReduceEdgeSlope          
{{0x0F,0x12},{0x0A,0x01}},	//70000A30//AFIT8_bnr_nClustLevel_H_Bin   [7:0] AFIT8_bnr_iClustMulT_H_Bin         
{{0x0F,0x12},{0x01,0x0A}},	//70000A32//AFIT8_bnr_iClustMulT_C_Bin   [7:0] AFIT8_bnr_iClustThresh_H_Bin        
{{0x0F,0x12},{0x36,0x01}},	//70000A34//AFIT8_bnr_iClustThresh_C_Bin   [7:0] AFIT8_bnr_iDenThreshLow_Bin       
{{0x0F,0x12},{0x24,0x2A}},	//70000A36//AFIT8_bnr_iDenThreshHigh_Bin   [7:0] AFIT8_ee_iLowSharpPower_Bin       
{{0x0F,0x12},{0x36,0x60}},	//70000A38//AFIT8_ee_iHighSharpPower_Bin   [7:0] AFIT8_ee_iLowShDenoise_Bin        
{{0x0F,0x12},{0xFF,0x2A}},	//70000A3A//AFIT8_ee_iHighShDenoise_Bin   [7:0] AFIT8_ee_iLowSharpClamp_Bin        
{{0x0F,0x12},{0x08,0xFF}},	//70000A3C//AFIT8_ee_iHighSharpClamp_Bin   [7:0] AFIT8_ee_iReduceEdgeMinMult_Bin   
{{0x0F,0x12},{0x00,0x08}},	//70000A3E//AFIT8_ee_iReduceEdgeSlope_Bin [7:0]                                    
{{0x0F,0x12},{0x00,0x01}},	//70000A40//AFITB_bnr_nClustLevel_C      [0]                                       	
{{0x0F,0x12},{0x00,0x00}},	//70000A42//AFIT16_BRIGHTNESS           
{{0x0F,0x12},{0x00,0x00}},	//70000A44//AFIT16_CONTRAST                                                        
{{0x0F,0x12},{0x00,0x00}},	//70000A46//AFIT16_SATURATION                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000A48//AFIT16_SHARP_BLUR                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000A4A//AFIT16_GLAMOUR                                                         
{{0x0F,0x12},{0x00,0xC0}},	//70000A4C//AFIT16_bnr_edge_high                                                   
{{0x0F,0x12},{0x00,0x64}},	//70000A4E//AFIT16_postdmsc_iLowBright                                             
{{0x0F,0x12},{0x03,0x84}},	//70000A50//AFIT16_postdmsc_iHighBright                                            
{{0x0F,0x12},{0x00,0x51}},	//70000A52//AFIT16_postdmsc_iLowSat                                                
{{0x0F,0x12},{0x01,0xF4}},	//70000A54//AFIT16_postdmsc_iHighSat                                               
{{0x0F,0x12},{0x00,0x70}},	//70000A56//AFIT16_postdmsc_iTune                                                  
{{0x0F,0x12},{0x00,0x40}},	//70000A58//AFIT16_yuvemix_mNegRanges_0                                            
{{0x0F,0x12},{0x00,0xA0}},	//70000A5A//AFIT16_yuvemix_mNegRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000A5C//AFIT16_yuvemix_mNegRanges_2                                            
{{0x0F,0x12},{0x00,0x10}},	//70000A5E//AFIT16_yuvemix_mPosRanges_0                                            
{{0x0F,0x12},{0x00,0x60}},	//70000A60//AFIT16_yuvemix_mPosRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000A62//AFIT16_yuvemix_mPosRanges_2                                            
{{0x0F,0x12},{0x14,0x30}},	//70000A64//AFIT8_bnr_edge_low  [7:0] AFIT8_bnr_repl_thresh                        
{{0x0F,0x12},{0x02,0x01}},	//70000A66//AFIT8_bnr_repl_force  [7:0] AFIT8_bnr_iHotThreshHigh                   
{{0x0F,0x12},{0x02,0x04}},	//70000A68//AFIT8_bnr_iHotThreshLow   [7:0] AFIT8_bnr_iColdThreshHigh              
{{0x0F,0x12},{0x24,0x04}},	//70000A6A//AFIT8_bnr_iColdThreshLow   [7:0] AFIT8_bnr_DispTH_Low                  
{{0x0F,0x12},{0x03,0x1B}},	//70000A6C//AFIT8_bnr_DispTH_High   [7:0] AFIT8_bnr_DISP_Limit_Low                 
{{0x0F,0x12},{0x01,0x03}},	//70000A6E//AFIT8_bnr_DISP_Limit_High   [7:0] AFIT8_bnr_iDistSigmaMin              
{{0x0F,0x12},{0x12,0x05}},	//70000A70//AFIT8_bnr_iDistSigmaMax   [7:0] AFIT8_bnr_iDiffSigmaLow                
{{0x0F,0x12},{0x40,0x0D}},	//70000A72//AFIT8_bnr_iDiffSigmaHigh   [7:0] AFIT8_bnr_iNormalizedSTD_TH           
{{0x0F,0x12},{0x00,0x80}},	//70000A74//AFIT8_bnr_iNormalizedSTD_Limit   [7:0] AFIT8_bnr_iDirNRTune            
{{0x0F,0x12},{0x20,0x80}},	//70000A76//AFIT8_bnr_iDirMinThres   [7:0] AFIT8_bnr_iDirFltDiffThresHigh          
{{0x0F,0x12},{0x30,0x40}},	//70000A78//AFIT8_bnr_iDirFltDiffThresLow   [7:0] AFIT8_bnr_iDirSmoothPowerHigh    
{{0x0F,0x12},{0x06,0x30}},	//70000A7A//AFIT8_bnr_iDirSmoothPowerLow   [7:0] AFIT8_bnr_iLowMaxSlopeAllowed     
{{0x0F,0x12},{0x03,0x06}},	//70000A7C//AFIT8_bnr_iHighMaxSlopeAllowed   [7:0] AFIT8_bnr_iLowSlopeThresh       
{{0x0F,0x12},{0x20,0x03}},	//70000A7E//AFIT8_bnr_iHighSlopeThresh   [7:0] AFIT8_bnr_iSlopenessTH              
{{0x0F,0x12},{0xFF,0x01}},	//70000A80//AFIT8_bnr_iSlopeBlurStrength   [7:0] AFIT8_bnr_iSlopenessLimit         
{{0x0F,0x12},{0x04,0x04}},	//70000A82//AFIT8_bnr_AddNoisePower1   [7:0] AFIT8_bnr_AddNoisePower2              
{{0x0F,0x12},{0x03,0x00}},	//70000A84//AFIT8_bnr_iRadialTune   [7:0] AFIT8_bnr_iRadialPower                   
{{0x0F,0x12},{0x24,0x5A}},	//70000A86//AFIT8_bnr_iRadialLimit   [7:0] AFIT8_ee_iFSMagThLow                    
{{0x0F,0x12},{0x10,0x18}},	//70000A88//AFIT8_ee_iFSMagThHigh   [7:0] AFIT8_ee_iFSVarThLow                     
{{0x0F,0x12},{0x00,0x0B}},	//70000A8A//AFIT8_ee_iFSVarThHigh   [7:0] AFIT8_ee_iFSThLow                        
{{0x0F,0x12},{0x0B,0x00}},	//70000A8C//AFIT8_ee_iFSThHigh   [7:0] AFIT8_ee_iFSmagPower                        
{{0x0F,0x12},{0x5A,0x0F}},	//70000A8E//AFIT8_ee_iFSVarCountTh   [7:0] AFIT8_ee_iRadialLimit                   
{{0x0F,0x12},{0x05,0x05}},	//70000A90//AFIT8_ee_iRadialPower   [7:0] AFIT8_ee_iSmoothEdgeSlope                
{{0x0F,0x12},{0x18,0x02}},	//70000A92//AFIT8_ee_iROADThres   [7:0] AFIT8_ee_iROADMaxNR                        
{{0x0F,0x12},{0x00,0x00}},	//70000A94//AFIT8_ee_iROADSubMaxNR   [7:0] AFIT8_ee_iROADSubThres                  
{{0x0F,0x12},{0x20,0x06}},	//70000A96//AFIT8_ee_iROADNeiThres   [7:0] AFIT8_ee_iROADNeiMaxNR                  
{{0x0F,0x12},{0x34,0x28}},	//70000A98//AFIT8_ee_iSmoothEdgeThres   [7:0] AFIT8_ee_iMSharpen                   
{{0x0F,0x12},{0x04,0x1C}},	//70000A9A//AFIT8_ee_iWSharpen   [7:0] AFIT8_ee_iMShThresh                         
{{0x0F,0x12},{0x01,0x01}},	//70000A9C//AFIT8_ee_iWShThresh   [7:0] AFIT8_ee_iReduceNegative                   
{{0x0F,0x12},{0x08,0x00}},	//70000A9E//AFIT8_ee_iEmbossCentAdd   [7:0] AFIT8_ee_iShDespeckle                  
{{0x0F,0x12},{0x10,0x04}},	//70000AA0//AFIT8_ee_iReduceEdgeThresh   [7:0] AFIT8_dmsc_iEnhThresh               
{{0x0F,0x12},{0x40,0x08}},	//70000AA2//AFIT8_dmsc_iDesatThresh   [7:0] AFIT8_dmsc_iDemBlurHigh                
{{0x0F,0x12},{0x05,0x40}},	//70000AA4//AFIT8_dmsc_iDemBlurLow   [7:0] AFIT8_dmsc_iDemBlurRange                
{{0x0F,0x12},{0x80,0x06}},	//70000AA6//AFIT8_dmsc_iDecisionThresh   [7:0] AFIT8_dmsc_iCentGrad                
{{0x0F,0x12},{0x00,0x20}},	//70000AA8//AFIT8_dmsc_iMonochrom   [7:0] AFIT8_dmsc_iGBDenoiseVal                 
{{0x0F,0x12},{0x00,0x00}},	//70000AAA//AFIT8_dmsc_iGRDenoiseVal   [7:0] AFIT8_dmsc_iEdgeDesatThrHigh          
{{0x0F,0x12},{0x18,0x00}},	//70000AAC//AFIT8_dmsc_iEdgeDesatThrLow   [7:0] AFIT8_dmsc_iEdgeDesat              
{{0x0F,0x12},{0x00,0x00}},	//70000AAE//AFIT8_dmsc_iNearGrayDesat   [7:0] AFIT8_dmsc_iEdgeDesatLimit           
{{0x0F,0x12},{0x1E,0x10}},	//70000AB0//AFIT8_postdmsc_iBCoeff   [7:0] AFIT8_postdmsc_iGCoeff                  
{{0x0F,0x12},{0x00,0x0B}},	//70000AB2//AFIT8_postdmsc_iWideMult   [7:0] AFIT8_yuvemix_mNegSlopes_0            
{{0x0F,0x12},{0x06,0x07}},	//70000AB4//AFIT8_yuvemix_mNegSlopes_1   [7:0] AFIT8_yuvemix_mNegSlopes_2          
{{0x0F,0x12},{0x00,0x05}},	//70000AB6//AFIT8_yuvemix_mNegSlopes_3   [7:0] AFIT8_yuvemix_mPosSlopes_0          
{{0x0F,0x12},{0x06,0x07}},	//70000AB8//AFIT8_yuvemix_mPosSlopes_1   [7:0] AFIT8_yuvemix_mPosSlopes_2          
{{0x0F,0x12},{0x04,0x05}},	//70000ABA//AFIT8_yuvemix_mPosSlopes_3   [7:0] AFIT8_yuviirnr_iXSupportY           
{{0x0F,0x12},{0x02,0x05}},	//70000ABC//AFIT8_yuviirnr_iXSupportUV   [7:0] AFIT8_yuviirnr_iLowYNorm            
{{0x0F,0x12},{0x03,0x04}},	//70000ABE//AFIT8_yuviirnr_iHighYNorm   [7:0] AFIT8_yuviirnr_iLowUVNorm            
{{0x0F,0x12},{0x04,0x09}},	//70000AC0//AFIT8_yuviirnr_iHighUVNorm   [7:0] AFIT8_yuviirnr_iYNormShift          
{{0x0F,0x12},{0x03,0x06}},	//70000AC2//AFIT8_yuviirnr_iUVNormShift   [7:0] AFIT8_yuviirnr_iVertLength_Y       
{{0x0F,0x12},{0x04,0x07}},	//70000AC4//AFIT8_yuviirnr_iVertLength_UV   [7:0] AFIT8_yuviirnr_iDiffThreshL_Y    
{{0x0F,0x12},{0x1F,0x04}},	//70000AC6//AFIT8_yuviirnr_iDiffThreshH_Y   [7:0] AFIT8_yuviirnr_iDiffThreshL_UV   
{{0x0F,0x12},{0x02,0x18}},	//70000AC8//AFIT8_yuviirnr_iDiffThreshH_UV   [7:0] AFIT8_yuviirnr_iMaxThreshL_Y    
{{0x0F,0x12},{0x11,0x02}},	//70000ACA//AFIT8_yuviirnr_iMaxThreshH_Y   [7:0] AFIT8_yuviirnr_iMaxThreshL_UV     
{{0x0F,0x12},{0x06,0x11}},	//70000ACC//AFIT8_yuviirnr_iMaxThreshH_UV   [7:0] AFIT8_yuviirnr_iYNRStrengthL     
{{0x0F,0x12},{0x1A,0x02}},	//70000ACE//AFIT8_yuviirnr_iYNRStrengthH   [7:0] AFIT8_yuviirnr_iUVNRStrengthL     
{{0x0F,0x12},{0x80,0x18}},	//70000AD0//AFIT8_yuviirnr_iUVNRStrengthH   [7:0] AFIT8_byr_gras_iShadingPower     
{{0x0F,0x12},{0x00,0x80}},	//70000AD2//AFIT8_RGBGamma2_iLinearity   [7:0] AFIT8_RGBGamma2_iDarkReduce         
{{0x0F,0x12},{0x03,0x80}},	//70000AD4//AFIT8_ccm_oscar_iSaturation   [7:0] AFIT8_RGB2YUV_iYOffset             
{{0x0F,0x12},{0x01,0x80}},	//70000AD6//AFIT8_RGB2YUV_iRGBGain   [7:0] AFIT8_bnr_nClustLevel_H                 
{{0x0F,0x12},{0x0A,0x0A}},	//70000AD8//AFIT8_bnr_iClustMulT_H   [7:0] AFIT8_bnr_iClustMulT_C                  
{{0x0F,0x12},{0x01,0x01}},	//70000ADA//AFIT8_bnr_iClustThresh_H   [7:0] AFIT8_bnr_iClustThresh_C              
{{0x0F,0x12},{0x1B,0x24}},	//70000ADC//AFIT8_bnr_iDenThreshLow   [7:0] AFIT8_bnr_iDenThreshHigh               
{{0x0F,0x12},{0x60,0x24}},	//70000ADE//AFIT8_ee_iLowSharpPower   [7:0] AFIT8_ee_iHighSharpPower               
{{0x0F,0x12},{0x1D,0x22}},	//70000AE0//AFIT8_ee_iLowShDenoise   [7:0] AFIT8_ee_iHighShDenoise                 
{{0x0F,0x12},{0xFF,0xFF}},	//70000AE2//AFIT8_ee_iLowSharpClamp   [7:0] AFIT8_ee_iHighSharpClamp               
{{0x0F,0x12},{0x08,0x08}},	//70000AE4//AFIT8_ee_iReduceEdgeMinMult   [7:0] AFIT8_ee_iReduceEdgeSlope          
{{0x0F,0x12},{0x0A,0x01}},	//70000AE6//AFIT8_bnr_nClustLevel_H_Bin   [7:0] AFIT8_bnr_iClustMulT_H_Bin         
{{0x0F,0x12},{0x01,0x0A}},	//70000AE8//AFIT8_bnr_iClustMulT_C_Bin   [7:0] AFIT8_bnr_iClustThresh_H_Bin        
{{0x0F,0x12},{0x24,0x01}},	//70000AEA//AFIT8_bnr_iClustThresh_C_Bin   [7:0] AFIT8_bnr_iDenThreshLow_Bin       
{{0x0F,0x12},{0x24,0x1B}},	//70000AEC//AFIT8_bnr_iDenThreshHigh_Bin   [7:0] AFIT8_ee_iLowSharpPower_Bin       
{{0x0F,0x12},{0x1E,0x60}},	//70000AEE//AFIT8_ee_iHighSharpPower_Bin   [7:0] AFIT8_ee_iLowShDenoise_Bin        
{{0x0F,0x12},{0xFF,0x18}},	//70000AF0//AFIT8_ee_iHighShDenoise_Bin   [7:0] AFIT8_ee_iLowSharpClamp_Bin        
{{0x0F,0x12},{0x08,0xFF}},	//70000AF2//AFIT8_ee_iHighSharpClamp_Bin   [7:0] AFIT8_ee_iReduceEdgeMinMult_Bin   
{{0x0F,0x12},{0x00,0x08}},	//70000AF4//AFIT8_ee_iReduceEdgeSlope_Bin [7:0]                                    
{{0x0F,0x12},{0x00,0x01}},	//70000AF6//AFITB_bnr_nClustLevel_C      [0]                                       	
{{0x0F,0x12},{0x00,0x00}},	//70000AF8//AFIT16_BRIGHTNESS           
{{0x0F,0x12},{0x00,0x00}},	//70000AFA//AFIT16_CONTRAST                                                        
{{0x0F,0x12},{0x00,0x00}},	//70000AFC//AFIT16_SATURATION                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000AFE//AFIT16_SHARP_BLUR                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000B00//AFIT16_GLAMOUR                                                         
{{0x0F,0x12},{0x00,0xC0}},	//70000B02//AFIT16_bnr_edge_high                                                   
{{0x0F,0x12},{0x00,0x64}},	//70000B04//AFIT16_postdmsc_iLowBright                                             
{{0x0F,0x12},{0x03,0x84}},	//70000B06//AFIT16_postdmsc_iHighBright                                            
{{0x0F,0x12},{0x00,0x43}},	//70000B08//AFIT16_postdmsc_iLowSat                                                
{{0x0F,0x12},{0x01,0xF4}},	//70000B0A//AFIT16_postdmsc_iHighSat                                               
{{0x0F,0x12},{0x00,0x70}},	//70000B0C//AFIT16_postdmsc_iTune                                                  
{{0x0F,0x12},{0x00,0x40}},	//70000B0E//AFIT16_yuvemix_mNegRanges_0                                            
{{0x0F,0x12},{0x00,0xA0}},	//70000B10//AFIT16_yuvemix_mNegRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000B12//AFIT16_yuvemix_mNegRanges_2                                            
{{0x0F,0x12},{0x00,0x10}},	//70000B14//AFIT16_yuvemix_mPosRanges_0                                            
{{0x0F,0x12},{0x00,0x60}},	//70000B16//AFIT16_yuvemix_mPosRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000B18//AFIT16_yuvemix_mPosRanges_2                                            
{{0x0F,0x12},{0x14,0x30}},	//70000B1A//AFIT8_bnr_edge_low  [7:0] AFIT8_bnr_repl_thresh                        
{{0x0F,0x12},{0x02,0x01}},	//70000B1C//AFIT8_bnr_repl_force  [7:0] AFIT8_bnr_iHotThreshHigh                   
{{0x0F,0x12},{0x02,0x04}},	//70000B1E//AFIT8_bnr_iHotThreshLow   [7:0] AFIT8_bnr_iColdThreshHigh              
{{0x0F,0x12},{0x1B,0x04}},	//70000B20//AFIT8_bnr_iColdThreshLow   [7:0] AFIT8_bnr_DispTH_Low                  
{{0x0F,0x12},{0x03,0x12}},	//70000B22//AFIT8_bnr_DispTH_High   [7:0] AFIT8_bnr_DISP_Limit_Low                 
{{0x0F,0x12},{0x00,0x03}},	//70000B24//AFIT8_bnr_DISP_Limit_High   [7:0] AFIT8_bnr_iDistSigmaMin              
{{0x0F,0x12},{0x0C,0x03}},	//70000B26//AFIT8_bnr_iDistSigmaMax   [7:0] AFIT8_bnr_iDiffSigmaLow                
{{0x0F,0x12},{0x28,0x06}},	//70000B28//AFIT8_bnr_iDiffSigmaHigh   [7:0] AFIT8_bnr_iNormalizedSTD_TH           
{{0x0F,0x12},{0x00,0x60}},	//70000B2A//AFIT8_bnr_iNormalizedSTD_Limit   [7:0] AFIT8_bnr_iDirNRTune            
{{0x0F,0x12},{0x15,0x80}},	//70000B2C//AFIT8_bnr_iDirMinThres   [7:0] AFIT8_bnr_iDirFltDiffThresHigh          
{{0x0F,0x12},{0x20,0x20}},	//70000B2E//AFIT8_bnr_iDirFltDiffThresLow   [7:0] AFIT8_bnr_iDirSmoothPowerHigh    
{{0x0F,0x12},{0x06,0x20}},	//70000B30//AFIT8_bnr_iDirSmoothPowerLow   [7:0] AFIT8_bnr_iLowMaxSlopeAllowed     
{{0x0F,0x12},{0x03,0x06}},	//70000B32//AFIT8_bnr_iHighMaxSlopeAllowed   [7:0] AFIT8_bnr_iLowSlopeThresh       
{{0x0F,0x12},{0x20,0x03}},	//70000B34//AFIT8_bnr_iHighSlopeThresh   [7:0] AFIT8_bnr_iSlopenessTH              
{{0x0F,0x12},{0xFF,0x01}},	//70000B36//AFIT8_bnr_iSlopeBlurStrength   [7:0] AFIT8_bnr_iSlopenessLimit         
{{0x0F,0x12},{0x04,0x04}},	//70000B38//AFIT8_bnr_AddNoisePower1   [7:0] AFIT8_bnr_AddNoisePower2              
{{0x0F,0x12},{0x03,0x00}},	//70000B3A//AFIT8_bnr_iRadialTune   [7:0] AFIT8_bnr_iRadialPower                   
{{0x0F,0x12},{0x14,0x5A}},	//70000B3C//AFIT8_bnr_iRadialLimit   [7:0] AFIT8_ee_iFSMagThLow                    
{{0x0F,0x12},{0x10,0x10}},	//70000B3E//AFIT8_ee_iFSMagThHigh   [7:0] AFIT8_ee_iFSVarThLow                     
{{0x0F,0x12},{0x00,0x0B}},	//70000B40//AFIT8_ee_iFSVarThHigh   [7:0] AFIT8_ee_iFSThLow                        
{{0x0F,0x12},{0x0E,0x00}},	//70000B42//AFIT8_ee_iFSThHigh   [7:0] AFIT8_ee_iFSmagPower                        
{{0x0F,0x12},{0x5A,0x0F}},	//70000B44//AFIT8_ee_iFSVarCountTh   [7:0] AFIT8_ee_iRadialLimit                   
{{0x0F,0x12},{0x05,0x04}},	//70000B46//AFIT8_ee_iRadialPower   [7:0] AFIT8_ee_iSmoothEdgeSlope                
{{0x0F,0x12},{0x18,0x02}},	//70000B48//AFIT8_ee_iROADThres   [7:0] AFIT8_ee_iROADMaxNR                        
{{0x0F,0x12},{0x00,0x00}},	//70000B4A//AFIT8_ee_iROADSubMaxNR   [7:0] AFIT8_ee_iROADSubThres                  
{{0x0F,0x12},{0x20,0x06}},	//70000B4C//AFIT8_ee_iROADNeiThres   [7:0] AFIT8_ee_iROADNeiMaxNR                  
{{0x0F,0x12},{0x38,0x28}},	//70000B4E//AFIT8_ee_iSmoothEdgeThres   [7:0] AFIT8_ee_iMSharpen                   
{{0x0F,0x12},{0x04,0x28}},	//70000B50//AFIT8_ee_iWSharpen   [7:0] AFIT8_ee_iMShThresh                         
{{0x0F,0x12},{0x01,0x01}},	//70000B52//AFIT8_ee_iWShThresh   [7:0] AFIT8_ee_iReduceNegative                   
{{0x0F,0x12},{0x80,0x00}},	//70000B54//AFIT8_ee_iEmbossCentAdd   [7:0] AFIT8_ee_iShDespeckle                  
{{0x0F,0x12},{0x0A,0x04}},	//70000B56//AFIT8_ee_iReduceEdgeThresh   [7:0] AFIT8_dmsc_iEnhThresh               
{{0x0F,0x12},{0x40,0x08}},	//70000B58//AFIT8_dmsc_iDesatThresh   [7:0] AFIT8_dmsc_iDemBlurHigh                
{{0x0F,0x12},{0x05,0x40}},	//70000B5A//AFIT8_dmsc_iDemBlurLow   [7:0] AFIT8_dmsc_iDemBlurRange                
{{0x0F,0x12},{0x80,0x06}},	//70000B5C//AFIT8_dmsc_iDecisionThresh   [7:0] AFIT8_dmsc_iCentGrad                
{{0x0F,0x12},{0x00,0x20}},	//70000B5E//AFIT8_dmsc_iMonochrom   [7:0] AFIT8_dmsc_iGBDenoiseVal                 
{{0x0F,0x12},{0x00,0x00}},	//70000B60//AFIT8_dmsc_iGRDenoiseVal   [7:0] AFIT8_dmsc_iEdgeDesatThrHigh          
{{0x0F,0x12},{0x18,0x00}},	//70000B62//AFIT8_dmsc_iEdgeDesatThrLow   [7:0] AFIT8_dmsc_iEdgeDesat              
{{0x0F,0x12},{0x00,0x00}},	//70000B64//AFIT8_dmsc_iNearGrayDesat   [7:0] AFIT8_dmsc_iEdgeDesatLimit           
{{0x0F,0x12},{0x1E,0x10}},	//70000B66//AFIT8_postdmsc_iBCoeff   [7:0] AFIT8_postdmsc_iGCoeff                  
{{0x0F,0x12},{0x00,0x0B}},	//70000B68//AFIT8_postdmsc_iWideMult   [7:0] AFIT8_yuvemix_mNegSlopes_0            
{{0x0F,0x12},{0x06,0x07}},	//70000B6A//AFIT8_yuvemix_mNegSlopes_1   [7:0] AFIT8_yuvemix_mNegSlopes_2          
{{0x0F,0x12},{0x00,0x05}},	//70000B6C//AFIT8_yuvemix_mNegSlopes_3   [7:0] AFIT8_yuvemix_mPosSlopes_0          
{{0x0F,0x12},{0x06,0x07}},	//70000B6E//AFIT8_yuvemix_mPosSlopes_1   [7:0] AFIT8_yuvemix_mPosSlopes_2          
{{0x0F,0x12},{0x04,0x05}},	//70000B70//AFIT8_yuvemix_mPosSlopes_3   [7:0] AFIT8_yuviirnr_iXSupportY           
{{0x0F,0x12},{0x02,0x07}},	//70000B72//AFIT8_yuviirnr_iXSupportUV   [7:0] AFIT8_yuviirnr_iLowYNorm            
{{0x0F,0x12},{0x03,0x04}},	//70000B74//AFIT8_yuviirnr_iHighYNorm   [7:0] AFIT8_yuviirnr_iLowUVNorm            
{{0x0F,0x12},{0x04,0x09}},	//70000B76//AFIT8_yuviirnr_iHighUVNorm   [7:0] AFIT8_yuviirnr_iYNormShift          
{{0x0F,0x12},{0x03,0x06}},	//70000B78//AFIT8_yuviirnr_iUVNormShift   [7:0] AFIT8_yuviirnr_iVertLength_Y       
{{0x0F,0x12},{0x04,0x07}},	//70000B7A//AFIT8_yuviirnr_iVertLength_UV   [7:0] AFIT8_yuviirnr_iDiffThreshL_Y    
{{0x0F,0x12},{0x24,0x04}},	//70000B7C//AFIT8_yuviirnr_iDiffThreshH_Y   [7:0] AFIT8_yuviirnr_iDiffThreshL_UV   
{{0x0F,0x12},{0x02,0x21}},	//70000B7E//AFIT8_yuviirnr_iDiffThreshH_UV   [7:0] AFIT8_yuviirnr_iMaxThreshL_Y    
{{0x0F,0x12},{0x12,0x02}},	//70000B80//AFIT8_yuviirnr_iMaxThreshH_Y   [7:0] AFIT8_yuviirnr_iMaxThreshL_UV     
{{0x0F,0x12},{0x06,0x13}},	//70000B82//AFIT8_yuviirnr_iMaxThreshH_UV   [7:0] AFIT8_yuviirnr_iYNRStrengthL     
{{0x0F,0x12},{0x1A,0x02}},	//70000B84//AFIT8_yuviirnr_iYNRStrengthH   [7:0] AFIT8_yuviirnr_iUVNRStrengthL     
{{0x0F,0x12},{0x80,0x18}},	//70000B86//AFIT8_yuviirnr_iUVNRStrengthH   [7:0] AFIT8_byr_gras_iShadingPower     
{{0x0F,0x12},{0x00,0x80}},	//70000B88//AFIT8_RGBGamma2_iLinearity   [7:0] AFIT8_RGBGamma2_iDarkReduce         
{{0x0F,0x12},{0x00,0x80}},	//70000B8A//AFIT8_ccm_oscar_iSaturation   [7:0] AFIT8_RGB2YUV_iYOffset             
{{0x0F,0x12},{0x01,0x80}},	//70000B8C//AFIT8_RGB2YUV_iRGBGain   [7:0] AFIT8_bnr_nClustLevel_H                 
{{0x0F,0x12},{0x0A,0x0A}},	//70000B8E//AFIT8_bnr_iClustMulT_H   [7:0] AFIT8_bnr_iClustMulT_C                  
{{0x0F,0x12},{0x01,0x01}},	//70000B90//AFIT8_bnr_iClustThresh_H   [7:0] AFIT8_bnr_iClustThresh_C              
{{0x0F,0x12},{0x14,0x1D}},	//70000B92//AFIT8_bnr_iDenThreshLow   [7:0] AFIT8_bnr_iDenThreshHigh               
{{0x0F,0x12},{0x60,0x24}},	//70000B94//AFIT8_ee_iLowSharpPower   [7:0] AFIT8_ee_iHighSharpPower               
{{0x0F,0x12},{0x0C,0x0C}},	//70000B96//AFIT8_ee_iLowShDenoise   [7:0] AFIT8_ee_iHighShDenoise                 
{{0x0F,0x12},{0xFF,0xFF}},	//70000B98//AFIT8_ee_iLowSharpClamp   [7:0] AFIT8_ee_iHighSharpClamp               
{{0x0F,0x12},{0x08,0x08}},	//70000B9A//AFIT8_ee_iReduceEdgeMinMult   [7:0] AFIT8_ee_iReduceEdgeSlope          
{{0x0F,0x12},{0x0A,0x01}},	//70000B9C//AFIT8_bnr_nClustLevel_H_Bin   [7:0] AFIT8_bnr_iClustMulT_H_Bin         
{{0x0F,0x12},{0x01,0x0A}},	//70000B9E//AFIT8_bnr_iClustMulT_C_Bin   [7:0] AFIT8_bnr_iClustThresh_H_Bin        
{{0x0F,0x12},{0x1B,0x01}},	//70000BA0//AFIT8_bnr_iClustThresh_C_Bin   [7:0] AFIT8_bnr_iDenThreshLow_Bin       
{{0x0F,0x12},{0x24,0x12}},	//70000BA2//AFIT8_bnr_iDenThreshHigh_Bin   [7:0] AFIT8_ee_iLowSharpPower_Bin       
{{0x0F,0x12},{0x0C,0x60}},	//70000BA4//AFIT8_ee_iHighSharpPower_Bin   [7:0] AFIT8_ee_iLowShDenoise_Bin        
{{0x0F,0x12},{0xFF,0x0C}},	//70000BA6//AFIT8_ee_iHighShDenoise_Bin   [7:0] AFIT8_ee_iLowSharpClamp_Bin        
{{0x0F,0x12},{0x08,0xFF}},	//70000BA8//AFIT8_ee_iHighSharpClamp_Bin   [7:0] AFIT8_ee_iReduceEdgeMinMult_Bin   
{{0x0F,0x12},{0x00,0x08}},	//70000BAA//AFIT8_ee_iReduceEdgeSlope_Bin [7:0]                                    
{{0x0F,0x12},{0x00,0x01}},	//70000BAC//AFITB_bnr_nClustLevel_C      [0]                                       	
{{0x0F,0x12},{0x00,0x00}},	//70000BAE//AFIT16_BRIGHTNESS           
{{0x0F,0x12},{0x00,0x00}},	//70000BB0//AFIT16_CONTRAST                                                        
{{0x0F,0x12},{0x00,0x00}},	//70000BB2//AFIT16_SATURATION                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000BB4//AFIT16_SHARP_BLUR                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000BB6//AFIT16_GLAMOUR                                                         
{{0x0F,0x12},{0x00,0xC0}},	//70000BB8//AFIT16_bnr_edge_high                                                   
{{0x0F,0x12},{0x00,0x64}},	//70000BBA//AFIT16_postdmsc_iLowBright                                             
{{0x0F,0x12},{0x03,0x84}},	//70000BBC//AFIT16_postdmsc_iHighBright                                            
{{0x0F,0x12},{0x00,0x32}},	//70000BBE//AFIT16_postdmsc_iLowSat                                                
{{0x0F,0x12},{0x01,0xF4}},	//70000BC0//AFIT16_postdmsc_iHighSat                                               
{{0x0F,0x12},{0x00,0x70}},	//70000BC2//AFIT16_postdmsc_iTune                                                  
{{0x0F,0x12},{0x00,0x40}},	//70000BC4//AFIT16_yuvemix_mNegRanges_0                                            
{{0x0F,0x12},{0x00,0xA0}},	//70000BC6//AFIT16_yuvemix_mNegRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000BC8//AFIT16_yuvemix_mNegRanges_2                                            
{{0x0F,0x12},{0x00,0x10}},	//70000BCA//AFIT16_yuvemix_mPosRanges_0                                            
{{0x0F,0x12},{0x00,0x60}},	//70000BCC//AFIT16_yuvemix_mPosRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000BCE//AFIT16_yuvemix_mPosRanges_2                                            
{{0x0F,0x12},{0x14,0x30}},	//70000BD0//AFIT8_bnr_edge_low  [7:0] AFIT8_bnr_repl_thresh                        
{{0x0F,0x12},{0x02,0x01}},	//70000BD2//AFIT8_bnr_repl_force  [7:0] AFIT8_bnr_iHotThreshHigh                   
{{0x0F,0x12},{0x02,0x04}},	//70000BD4//AFIT8_bnr_iHotThreshLow   [7:0] AFIT8_bnr_iColdThreshHigh              
{{0x0F,0x12},{0x15,0x04}},	//70000BD6//AFIT8_bnr_iColdThreshLow   [7:0] AFIT8_bnr_DispTH_Low                  
{{0x0F,0x12},{0x03,0x0F}},	//70000BD8//AFIT8_bnr_DispTH_High   [7:0] AFIT8_bnr_DISP_Limit_Low                 
{{0x0F,0x12},{0x00,0x03}},	//70000BDA//AFIT8_bnr_DISP_Limit_High   [7:0] AFIT8_bnr_iDistSigmaMin              
{{0x0F,0x12},{0x09,0x02}},	//70000BDC//AFIT8_bnr_iDistSigmaMax   [7:0] AFIT8_bnr_iDiffSigmaLow                
{{0x0F,0x12},{0x20,0x04}},	//70000BDE//AFIT8_bnr_iDiffSigmaHigh   [7:0] AFIT8_bnr_iNormalizedSTD_TH           
{{0x0F,0x12},{0x00,0x50}},	//70000BE0//AFIT8_bnr_iNormalizedSTD_Limit   [7:0] AFIT8_bnr_iDirNRTune            
{{0x0F,0x12},{0x11,0x40}},	//70000BE2//AFIT8_bnr_iDirMinThres   [7:0] AFIT8_bnr_iDirFltDiffThresHigh          
{{0x0F,0x12},{0x20,0x1C}},	//70000BE4//AFIT8_bnr_iDirFltDiffThresLow   [7:0] AFIT8_bnr_iDirSmoothPowerHigh    
{{0x0F,0x12},{0x06,0x20}},	//70000BE6//AFIT8_bnr_iDirSmoothPowerLow   [7:0] AFIT8_bnr_iLowMaxSlopeAllowed     
{{0x0F,0x12},{0x03,0x06}},	//70000BE8//AFIT8_bnr_iHighMaxSlopeAllowed   [7:0] AFIT8_bnr_iLowSlopeThresh       
{{0x0F,0x12},{0x20,0x03}},	//70000BEA//AFIT8_bnr_iHighSlopeThresh   [7:0] AFIT8_bnr_iSlopenessTH              
{{0x0F,0x12},{0xFF,0x01}},	//70000BEC//AFIT8_bnr_iSlopeBlurStrength   [7:0] AFIT8_bnr_iSlopenessLimit         
{{0x0F,0x12},{0x04,0x04}},	//70000BEE//AFIT8_bnr_AddNoisePower1   [7:0] AFIT8_bnr_AddNoisePower2              
{{0x0F,0x12},{0x03,0x00}},	//70000BF0//AFIT8_bnr_iRadialTune   [7:0] AFIT8_bnr_iRadialPower                   
{{0x0F,0x12},{0x14,0x5A}},	//70000BF2//AFIT8_bnr_iRadialLimit   [7:0] AFIT8_ee_iFSMagThLow                    
{{0x0F,0x12},{0x10,0x10}},	//70000BF4//AFIT8_ee_iFSMagThHigh   [7:0] AFIT8_ee_iFSVarThLow                     
{{0x0F,0x12},{0x00,0x0B}},	//70000BF6//AFIT8_ee_iFSVarThHigh   [7:0] AFIT8_ee_iFSThLow                        
{{0x0F,0x12},{0x10,0x00}},	//70000BF8//AFIT8_ee_iFSThHigh   [7:0] AFIT8_ee_iFSmagPower                        
{{0x0F,0x12},{0x5A,0x0F}},	//70000BFA//AFIT8_ee_iFSVarCountTh   [7:0] AFIT8_ee_iRadialLimit                   
{{0x0F,0x12},{0x05,0x03}},	//70000BFC//AFIT8_ee_iRadialPower   [7:0] AFIT8_ee_iSmoothEdgeSlope                
{{0x0F,0x12},{0x18,0x02}},	//70000BFE//AFIT8_ee_iROADThres   [7:0] AFIT8_ee_iROADMaxNR                        
{{0x0F,0x12},{0x00,0x00}},	//70000C00//AFIT8_ee_iROADSubMaxNR   [7:0] AFIT8_ee_iROADSubThres                  
{{0x0F,0x12},{0x20,0x06}},	//70000C02//AFIT8_ee_iROADNeiThres   [7:0] AFIT8_ee_iROADNeiMaxNR                  
{{0x0F,0x12},{0x3C,0x28}},	//70000C04//AFIT8_ee_iSmoothEdgeThres   [7:0] AFIT8_ee_iMSharpen                   
{{0x0F,0x12},{0x04,0x2C}},	//70000C06//AFIT8_ee_iWSharpen   [7:0] AFIT8_ee_iMShThresh                         
{{0x0F,0x12},{0x01,0x01}},	//70000C08//AFIT8_ee_iWShThresh   [7:0] AFIT8_ee_iReduceNegative                   
{{0x0F,0x12},{0xFF,0x00}},	//70000C0A//AFIT8_ee_iEmbossCentAdd   [7:0] AFIT8_ee_iShDespeckle                  
{{0x0F,0x12},{0x09,0x04}},	//70000C0C//AFIT8_ee_iReduceEdgeThresh   [7:0] AFIT8_dmsc_iEnhThresh               
{{0x0F,0x12},{0x40,0x08}},	//70000C0E//AFIT8_dmsc_iDesatThresh   [7:0] AFIT8_dmsc_iDemBlurHigh                
{{0x0F,0x12},{0x05,0x40}},	//70000C10//AFIT8_dmsc_iDemBlurLow   [7:0] AFIT8_dmsc_iDemBlurRange                
{{0x0F,0x12},{0x80,0x06}},	//70000C12//AFIT8_dmsc_iDecisionThresh   [7:0] AFIT8_dmsc_iCentGrad                
{{0x0F,0x12},{0x00,0x20}},	//70000C14//AFIT8_dmsc_iMonochrom   [7:0] AFIT8_dmsc_iGBDenoiseVal                 
{{0x0F,0x12},{0x00,0x00}},	//70000C16//AFIT8_dmsc_iGRDenoiseVal   [7:0] AFIT8_dmsc_iEdgeDesatThrHigh          
{{0x0F,0x12},{0x18,0x00}},	//70000C18//AFIT8_dmsc_iEdgeDesatThrLow   [7:0] AFIT8_dmsc_iEdgeDesat              
{{0x0F,0x12},{0x00,0x00}},	//70000C1A//AFIT8_dmsc_iNearGrayDesat   [7:0] AFIT8_dmsc_iEdgeDesatLimit           
{{0x0F,0x12},{0x1E,0x10}},	//70000C1C//AFIT8_postdmsc_iBCoeff   [7:0] AFIT8_postdmsc_iGCoeff                  
{{0x0F,0x12},{0x00,0x0B}},	//70000C1E//AFIT8_postdmsc_iWideMult   [7:0] AFIT8_yuvemix_mNegSlopes_0            
{{0x0F,0x12},{0x06,0x07}},	//70000C20//AFIT8_yuvemix_mNegSlopes_1   [7:0] AFIT8_yuvemix_mNegSlopes_2          
{{0x0F,0x12},{0x00,0x05}},	//70000C22//AFIT8_yuvemix_mNegSlopes_3   [7:0] AFIT8_yuvemix_mPosSlopes_0          
{{0x0F,0x12},{0x06,0x07}},	//70000C24//AFIT8_yuvemix_mPosSlopes_1   [7:0] AFIT8_yuvemix_mPosSlopes_2          
{{0x0F,0x12},{0x04,0x05}},	//70000C26//AFIT8_yuvemix_mPosSlopes_3   [7:0] AFIT8_yuviirnr_iXSupportY           
{{0x0F,0x12},{0x02,0x06}},	//70000C28//AFIT8_yuviirnr_iXSupportUV   [7:0] AFIT8_yuviirnr_iLowYNorm            
{{0x0F,0x12},{0x03,0x04}},	//70000C2A//AFIT8_yuviirnr_iHighYNorm   [7:0] AFIT8_yuviirnr_iLowUVNorm            
{{0x0F,0x12},{0x04,0x09}},	//70000C2C//AFIT8_yuviirnr_iHighUVNorm   [7:0] AFIT8_yuviirnr_iYNormShift          
{{0x0F,0x12},{0x03,0x05}},	//70000C2E//AFIT8_yuviirnr_iUVNormShift   [7:0] AFIT8_yuviirnr_iVertLength_Y       
{{0x0F,0x12},{0x04,0x06}},	//70000C30//AFIT8_yuviirnr_iVertLength_UV   [7:0] AFIT8_yuviirnr_iDiffThreshL_Y    
{{0x0F,0x12},{0x28,0x04}},	//70000C32//AFIT8_yuviirnr_iDiffThreshH_Y   [7:0] AFIT8_yuviirnr_iDiffThreshL_UV   
{{0x0F,0x12},{0x02,0x28}},	//70000C34//AFIT8_yuviirnr_iDiffThreshH_UV   [7:0] AFIT8_yuviirnr_iMaxThreshL_Y    
{{0x0F,0x12},{0x14,0x02}},	//70000C36//AFIT8_yuviirnr_iMaxThreshH_Y   [7:0] AFIT8_yuviirnr_iMaxThreshL_UV     
{{0x0F,0x12},{0x06,0x18}},	//70000C38//AFIT8_yuviirnr_iMaxThreshH_UV   [7:0] AFIT8_yuviirnr_iYNRStrengthL     
{{0x0F,0x12},{0x1A,0x02}},	//70000C3A//AFIT8_yuviirnr_iYNRStrengthH   [7:0] AFIT8_yuviirnr_iUVNRStrengthL     
{{0x0F,0x12},{0x80,0x18}},	//70000C3C//AFIT8_yuviirnr_iUVNRStrengthH   [7:0] AFIT8_byr_gras_iShadingPower     
{{0x0F,0x12},{0x00,0x80}},	//70000C3E//AFIT8_RGBGamma2_iLinearity   [7:0] AFIT8_RGBGamma2_iDarkReduce         
{{0x0F,0x12},{0x00,0x80}},	//70000C40//AFIT8_ccm_oscar_iSaturation   [7:0] AFIT8_RGB2YUV_iYOffset             
{{0x0F,0x12},{0x01,0x80}},	//70000C42//AFIT8_RGB2YUV_iRGBGain   [7:0] AFIT8_bnr_nClustLevel_H                 
{{0x0F,0x12},{0x0A,0x0A}},	//70000C44//AFIT8_bnr_iClustMulT_H   [7:0] AFIT8_bnr_iClustMulT_C                  
{{0x0F,0x12},{0x01,0x01}},	//70000C46//AFIT8_bnr_iClustThresh_H   [7:0] AFIT8_bnr_iClustThresh_C              
{{0x0F,0x12},{0x11,0x17}},	//70000C48//AFIT8_bnr_iDenThreshLow   [7:0] AFIT8_bnr_iDenThreshHigh               
{{0x0F,0x12},{0x60,0x24}},	//70000C4A//AFIT8_ee_iLowSharpPower   [7:0] AFIT8_ee_iHighSharpPower               
{{0x0F,0x12},{0x0A,0x0A}},	//70000C4C//AFIT8_ee_iLowShDenoise   [7:0] AFIT8_ee_iHighShDenoise                 
{{0x0F,0x12},{0xFF,0xFF}},	//70000C4E//AFIT8_ee_iLowSharpClamp   [7:0] AFIT8_ee_iHighSharpClamp               
{{0x0F,0x12},{0x08,0x08}},	//70000C50//AFIT8_ee_iReduceEdgeMinMult   [7:0] AFIT8_ee_iReduceEdgeSlope          
{{0x0F,0x12},{0x0A,0x01}},	//70000C52//AFIT8_bnr_nClustLevel_H_Bin   [7:0] AFIT8_bnr_iClustMulT_H_Bin         
{{0x0F,0x12},{0x01,0x0A}},	//70000C54//AFIT8_bnr_iClustMulT_C_Bin   [7:0] AFIT8_bnr_iClustThresh_H_Bin        
{{0x0F,0x12},{0x15,0x01}},	//70000C56//AFIT8_bnr_iClustThresh_C_Bin   [7:0] AFIT8_bnr_iDenThreshLow_Bin       
{{0x0F,0x12},{0x24,0x0F}},	//70000C58//AFIT8_bnr_iDenThreshHigh_Bin   [7:0] AFIT8_ee_iLowSharpPower_Bin       
{{0x0F,0x12},{0x0A,0x60}},	//70000C5A//AFIT8_ee_iHighSharpPower_Bin   [7:0] AFIT8_ee_iLowShDenoise_Bin        
{{0x0F,0x12},{0xFF,0x0A}},	//70000C5C//AFIT8_ee_iHighShDenoise_Bin   [7:0] AFIT8_ee_iLowSharpClamp_Bin        
{{0x0F,0x12},{0x08,0xFF}},	//70000C5E//AFIT8_ee_iHighSharpClamp_Bin   [7:0] AFIT8_ee_iReduceEdgeMinMult_Bin   
{{0x0F,0x12},{0x00,0x08}},	//70000C60//AFIT8_ee_iReduceEdgeSlope_Bin [7:0]                                    
{{0x0F,0x12},{0x00,0x01}},	//70000C62//AFITB_bnr_nClustLevel_C      [0]                                       	
{{0x0F,0x12},{0x00,0x00}},	//70000C64//AFIT16_BRIGHTNESS           
{{0x0F,0x12},{0x00,0x00}},	//70000C66//AFIT16_CONTRAST                                                        
{{0x0F,0x12},{0x00,0x00}},	//70000C68//AFIT16_SATURATION                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000C6A//AFIT16_SHARP_BLUR                                                      
{{0x0F,0x12},{0x00,0x00}},	//70000C6C//AFIT16_GLAMOUR                                                         
{{0x0F,0x12},{0x00,0xC0}},	//70000C6E//AFIT16_bnr_edge_high                                                   
{{0x0F,0x12},{0x00,0x64}},	//70000C70//AFIT16_postdmsc_iLowBright                                             
{{0x0F,0x12},{0x03,0x84}},	//70000C72//AFIT16_postdmsc_iHighBright                                            
{{0x0F,0x12},{0x00,0x32}},	//70000C74//AFIT16_postdmsc_iLowSat                                                
{{0x0F,0x12},{0x01,0xF4}},	//70000C76//AFIT16_postdmsc_iHighSat                                               
{{0x0F,0x12},{0x00,0x70}},	//70000C78//AFIT16_postdmsc_iTune                                                  
{{0x0F,0x12},{0x00,0x40}},	//70000C7A//AFIT16_yuvemix_mNegRanges_0                                            
{{0x0F,0x12},{0x00,0xA0}},	//70000C7C//AFIT16_yuvemix_mNegRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000C7E//AFIT16_yuvemix_mNegRanges_2                                            
{{0x0F,0x12},{0x00,0x10}},	//70000C80//AFIT16_yuvemix_mPosRanges_0                                            
{{0x0F,0x12},{0x00,0x60}},	//70000C82//AFIT16_yuvemix_mPosRanges_1                                            
{{0x0F,0x12},{0x01,0x00}},	//70000C84//AFIT16_yuvemix_mPosRanges_2                                            
{{0x0F,0x12},{0x14,0x30}},	//70000C86//AFIT8_bnr_edge_low  [7:0] AFIT8_bnr_repl_thresh                        
{{0x0F,0x12},{0x02,0x01}},	//70000C88//AFIT8_bnr_repl_force  [7:0] AFIT8_bnr_iHotThreshHigh                   
{{0x0F,0x12},{0x02,0x04}},	//70000C8A//AFIT8_bnr_iHotThreshLow   [7:0] AFIT8_bnr_iColdThreshHigh              
{{0x0F,0x12},{0x0F,0x04}},	//70000C8C//AFIT8_bnr_iColdThreshLow   [7:0] AFIT8_bnr_DispTH_Low                  
{{0x0F,0x12},{0x03,0x0C}},	//70000C8E//AFIT8_bnr_DispTH_High   [7:0] AFIT8_bnr_DISP_Limit_Low                 
{{0x0F,0x12},{0x00,0x03}},	//70000C90//AFIT8_bnr_DISP_Limit_High   [7:0] AFIT8_bnr_iDistSigmaMin              
{{0x0F,0x12},{0x06,0x02}},	//70000C92//AFIT8_bnr_iDistSigmaMax   [7:0] AFIT8_bnr_iDiffSigmaLow                
{{0x0F,0x12},{0x18,0x03}},	//70000C94//AFIT8_bnr_iDiffSigmaHigh   [7:0] AFIT8_bnr_iNormalizedSTD_TH           
{{0x0F,0x12},{0x00,0x40}},	//70000C96//AFIT8_bnr_iNormalizedSTD_Limit   [7:0] AFIT8_bnr_iDirNRTune            
{{0x0F,0x12},{0x0E,0x20}},	//70000C98//AFIT8_bnr_iDirMinThres   [7:0] AFIT8_bnr_iDirFltDiffThresHigh          
{{0x0F,0x12},{0x20,0x18}},	//70000C9A//AFIT8_bnr_iDirFltDiffThresLow   [7:0] AFIT8_bnr_iDirSmoothPowerHigh    
{{0x0F,0x12},{0x06,0x20}},	//70000C9C//AFIT8_bnr_iDirSmoothPowerLow   [7:0] AFIT8_bnr_iLowMaxSlopeAllowed     
{{0x0F,0x12},{0x03,0x06}},	//70000C9E//AFIT8_bnr_iHighMaxSlopeAllowed   [7:0] AFIT8_bnr_iLowSlopeThresh       
{{0x0F,0x12},{0x20,0x03}},	//70000CA0//AFIT8_bnr_iHighSlopeThresh   [7:0] AFIT8_bnr_iSlopenessTH              
{{0x0F,0x12},{0xFF,0x01}},	//70000CA2//AFIT8_bnr_iSlopeBlurStrength   [7:0] AFIT8_bnr_iSlopenessLimit         
{{0x0F,0x12},{0x04,0x04}},	//70000CA4//AFIT8_bnr_AddNoisePower1   [7:0] AFIT8_bnr_AddNoisePower2              
{{0x0F,0x12},{0x02,0x00}},	//70000CA6//AFIT8_bnr_iRadialTune   [7:0] AFIT8_bnr_iRadialPower                   
{{0x0F,0x12},{0x14,0x5A}},	//70000CA8//AFIT8_bnr_iRadialLimit   [7:0] AFIT8_ee_iFSMagThLow                    
{{0x0F,0x12},{0x10,0x10}},	//70000CAA//AFIT8_ee_iFSMagThHigh   [7:0] AFIT8_ee_iFSVarThLow                     
{{0x0F,0x12},{0x00,0x0B}},	//70000CAC//AFIT8_ee_iFSVarThHigh   [7:0] AFIT8_ee_iFSThLow                        
{{0x0F,0x12},{0x12,0x00}},	//70000CAE//AFIT8_ee_iFSThHigh   [7:0] AFIT8_ee_iFSmagPower                        
{{0x0F,0x12},{0x5A,0x0F}},	//70000CB0//AFIT8_ee_iFSVarCountTh   [7:0] AFIT8_ee_iRadialLimit                   
{{0x0F,0x12},{0x05,0x02}},	//70000CB2//AFIT8_ee_iRadialPower   [7:0] AFIT8_ee_iSmoothEdgeSlope                
{{0x0F,0x12},{0x18,0x02}},	//70000CB4//AFIT8_ee_iROADThres   [7:0] AFIT8_ee_iROADMaxNR                        
{{0x0F,0x12},{0x00,0x00}},	//70000CB6//AFIT8_ee_iROADSubMaxNR   [7:0] AFIT8_ee_iROADSubThres                  
{{0x0F,0x12},{0x20,0x06}},	//70000CB8//AFIT8_ee_iROADNeiThres   [7:0] AFIT8_ee_iROADNeiMaxNR                  
{{0x0F,0x12},{0x40,0x28}},	//70000CBA//AFIT8_ee_iSmoothEdgeThres   [7:0] AFIT8_ee_iMSharpen                   
{{0x0F,0x12},{0x04,0x30}},	//70000CBC//AFIT8_ee_iWSharpen   [7:0] AFIT8_ee_iMShThresh                         
{{0x0F,0x12},{0x01,0x01}},	//70000CBE//AFIT8_ee_iWShThresh   [7:0] AFIT8_ee_iReduceNegative                   
{{0x0F,0x12},{0xFF,0x00}},	//70000CC0//AFIT8_ee_iEmbossCentAdd   [7:0] AFIT8_ee_iShDespeckle                  
{{0x0F,0x12},{0x08,0x04}},	//70000CC2//AFIT8_ee_iReduceEdgeThresh   [7:0] AFIT8_dmsc_iEnhThresh               
{{0x0F,0x12},{0x40,0x08}},	//70000CC4//AFIT8_dmsc_iDesatThresh   [7:0] AFIT8_dmsc_iDemBlurHigh                
{{0x0F,0x12},{0x05,0x40}},	//70000CC6//AFIT8_dmsc_iDemBlurLow   [7:0] AFIT8_dmsc_iDemBlurRange                
{{0x0F,0x12},{0x80,0x06}},	//70000CC8//AFIT8_dmsc_iDecisionThresh   [7:0] AFIT8_dmsc_iCentGrad                
{{0x0F,0x12},{0x00,0x20}},	//70000CCA//AFIT8_dmsc_iMonochrom   [7:0] AFIT8_dmsc_iGBDenoiseVal                 
{{0x0F,0x12},{0x00,0x00}},	//70000CCC//AFIT8_dmsc_iGRDenoiseVal   [7:0] AFIT8_dmsc_iEdgeDesatThrHigh          
{{0x0F,0x12},{0x18,0x00}},	//70000CCE//AFIT8_dmsc_iEdgeDesatThrLow   [7:0] AFIT8_dmsc_iEdgeDesat              
{{0x0F,0x12},{0x00,0x00}},	//70000CD0//AFIT8_dmsc_iNearGrayDesat   [7:0] AFIT8_dmsc_iEdgeDesatLimit           
{{0x0F,0x12},{0x1E,0x10}},	//70000CD2//AFIT8_postdmsc_iBCoeff   [7:0] AFIT8_postdmsc_iGCoeff                  
{{0x0F,0x12},{0x00,0x0B}},	//70000CD4//AFIT8_postdmsc_iWideMult   [7:0] AFIT8_yuvemix_mNegSlopes_0            
{{0x0F,0x12},{0x06,0x07}},	//70000CD6//AFIT8_yuvemix_mNegSlopes_1   [7:0] AFIT8_yuvemix_mNegSlopes_2          
{{0x0F,0x12},{0x00,0x05}},	//70000CD8//AFIT8_yuvemix_mNegSlopes_3   [7:0] AFIT8_yuvemix_mPosSlopes_0          
{{0x0F,0x12},{0x06,0x07}},	//70000CDA//AFIT8_yuvemix_mPosSlopes_1   [7:0] AFIT8_yuvemix_mPosSlopes_2          
{{0x0F,0x12},{0x04,0x05}},	//70000CDC//AFIT8_yuvemix_mPosSlopes_3   [7:0] AFIT8_yuviirnr_iXSupportY           
{{0x0F,0x12},{0x02,0x05}},	//70000CDE//AFIT8_yuviirnr_iXSupportUV   [7:0] AFIT8_yuviirnr_iLowYNorm            
{{0x0F,0x12},{0x03,0x04}},	//70000CE0//AFIT8_yuviirnr_iHighYNorm   [7:0] AFIT8_yuviirnr_iLowUVNorm            
{{0x0F,0x12},{0x04,0x09}},	//70000CE2//AFIT8_yuviirnr_iHighUVNorm   [7:0] AFIT8_yuviirnr_iYNormShift          
{{0x0F,0x12},{0x03,0x06}},	//70000CE4//AFIT8_yuviirnr_iUVNormShift   [7:0] AFIT8_yuviirnr_iVertLength_Y       
{{0x0F,0x12},{0x04,0x07}},	//70000CE6//AFIT8_yuviirnr_iVertLength_UV   [7:0] AFIT8_yuviirnr_iDiffThreshL_Y    
{{0x0F,0x12},{0x2C,0x04}},	//70000CE8//AFIT8_yuviirnr_iDiffThreshH_Y   [7:0] AFIT8_yuviirnr_iDiffThreshL_UV   
{{0x0F,0x12},{0x02,0x2C}},	//70000CEA//AFIT8_yuviirnr_iDiffThreshH_UV   [7:0] AFIT8_yuviirnr_iMaxThreshL_Y    
{{0x0F,0x12},{0x14,0x02}},	//70000CEC//AFIT8_yuviirnr_iMaxThreshH_Y   [7:0] AFIT8_yuviirnr_iMaxThreshL_UV     
{{0x0F,0x12},{0x06,0x18}},	//70000CEE//AFIT8_yuviirnr_iMaxThreshH_UV   [7:0] AFIT8_yuviirnr_iYNRStrengthL     
{{0x0F,0x12},{0x1A,0x02}},	//70000CF0//AFIT8_yuviirnr_iYNRStrengthH   [7:0] AFIT8_yuviirnr_iUVNRStrengthL     
{{0x0F,0x12},{0x80,0x18}},	//70000CF2//AFIT8_yuviirnr_iUVNRStrengthH   [7:0] AFIT8_byr_gras_iShadingPower     
{{0x0F,0x12},{0x00,0x80}},	//70000CF4//AFIT8_RGBGamma2_iLinearity   [7:0] AFIT8_RGBGamma2_iDarkReduce         
{{0x0F,0x12},{0x00,0x80}},	//70000CF6//AFIT8_ccm_oscar_iSaturation   [7:0] AFIT8_RGB2YUV_iYOffset             
{{0x0F,0x12},{0x01,0x80}},	//70000CF8//AFIT8_RGB2YUV_iRGBGain   [7:0] AFIT8_bnr_nClustLevel_H                 
{{0x0F,0x12},{0x0A,0x0A}},	//70000CFA//AFIT8_bnr_iClustMulT_H   [7:0] AFIT8_bnr_iClustMulT_C                  
{{0x0F,0x12},{0x01,0x01}},	//70000CFC//AFIT8_bnr_iClustThresh_H   [7:0] AFIT8_bnr_iClustThresh_C              
{{0x0F,0x12},{0x0C,0x0F}},	//70000CFE//AFIT8_bnr_iDenThreshLow   [7:0] AFIT8_bnr_iDenThreshHigh               
{{0x0F,0x12},{0x60,0x24}},	//70000D00//AFIT8_ee_iLowSharpPower   [7:0] AFIT8_ee_iHighSharpPower               
{{0x0F,0x12},{0x08,0x08}},	//70000D02//AFIT8_ee_iLowShDenoise   [7:0] AFIT8_ee_iHighShDenoise                 
{{0x0F,0x12},{0xFF,0xFF}},	//70000D04//AFIT8_ee_iLowSharpClamp   [7:0] AFIT8_ee_iHighSharpClamp               
{{0x0F,0x12},{0x08,0x08}},	//70000D06//AFIT8_ee_iReduceEdgeMinMult   [7:0] AFIT8_ee_iReduceEdgeSlope          
{{0x0F,0x12},{0x0A,0x01}},	//70000D08//AFIT8_bnr_nClustLevel_H_Bin   [7:0] AFIT8_bnr_iClustMulT_H_Bin         
{{0x0F,0x12},{0x01,0x0A}},	//70000D0A//AFIT8_bnr_iClustMulT_C_Bin   [7:0] AFIT8_bnr_iClustThresh_H_Bin        
{{0x0F,0x12},{0x0F,0x01}},	//70000D0C//AFIT8_bnr_iClustThresh_C_Bin   [7:0] AFIT8_bnr_iDenThreshLow_Bin       
{{0x0F,0x12},{0x24,0x0C}},	//70000D0E//AFIT8_bnr_iDenThreshHigh_Bin   [7:0] AFIT8_ee_iLowSharpPower_Bin       
{{0x0F,0x12},{0x08,0x60}},	//70000D10//AFIT8_ee_iHighSharpPower_Bin   [7:0] AFIT8_ee_iLowShDenoise_Bin        
{{0x0F,0x12},{0xFF,0x08}},	//70000D12//AFIT8_ee_iHighShDenoise_Bin   [7:0] AFIT8_ee_iLowSharpClamp_Bin        
{{0x0F,0x12},{0x08,0xFF}},	//70000D14//AFIT8_ee_iHighSharpClamp_Bin   [7:0] AFIT8_ee_iReduceEdgeMinMult_Bin   
{{0x0F,0x12},{0x00,0x08}},	//70000D16//AFIT8_ee_iReduceEdgeSlope_Bin [7:0]                                    
{{0x0F,0x12},{0x00,0x01}},	//70000D18//AFITB_bnr_nClustLevel_C      [0]                                       
	
{{0x0F,0x12},{0x23,0xCE}},	//70000D19//ConstAfitBaseVals
{{0x0F,0x12},{0xFD,0xC8}},	//70000D1A//ConstAfitBaseVals
{{0x0F,0x12},{0x11,0x2E}},	//70000D1B//ConstAfitBaseVals
{{0x0F,0x12},{0x93,0xA5}},	//70000D1C//ConstAfitBaseVals
{{0x0F,0x12},{0xFE,0x67}},	//70000D1D//ConstAfitBaseVals
{{0x0F,0x12},{0x00,0x00}},	//70000D1E//ConstAfitBaseVals
	
//==================================================================================
// 18.JPEG Thumnail Setting
//==================================================================================
	
//s002A0478	
//s0F12005F	//REG_TC_BRC_usPrevQuality    
//s0F12005F	//REG_TC_BRC_usCaptureQuality 
//s0F120001	//REG_TC_THUMB_Thumb_bActive  
//s0F120280	//REG_TC_THUMB_Thumb_uWidth   
//s0F1201E0	//REG_TC_THUMB_Thumb_uHeight  
//s0F120005	//REG_TC_THUMB_Thumb_Format   

//s002A17DC	
//s0F120054	//jpeg_ManualMBCV                
//s002A1AE4	                                 
//s0F12001C	//senHal_bExtraAddLine           
//s002A0284	                                 
//s0F120001	//REG_TC_GP_bBypassScalerJpg     
//s002A028A	                                 
//s0F120000	//REG_TC_GP_bUse1FrameCaptureMode

//s002A1CC2	//DRx_uDRxWeight for AutoCont function  
//s0F120100                                           
//s0F120100                                           
//s0F120100                                           
//s0F120100   
                                        
//==================================================================================
// 19.Input Size Setting
//==================================================================================
//Input Size	
{{0x00,0x2A},{0x02,0x50}},                            	
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_PrevReqInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_PrevReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_PrevInputWidthOfs
{{0x0F,0x12},{0x00,0x0c}},	//REG_TC_GP_PrevInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_CapReqInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_CapReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_CapInputWidthOfs
{{0x0F,0x12},{0x00,0x0c}},	//REG_TC_GP_CapInputHeightOfs

{{0x00,0x2A},{0x04,0x94}}, 
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_PZOOM_ZoomInputWidth    
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_PZOOM_ZoomInputHeight   
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputWidthOfs 
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_CZOOM_ZoomInputWidth    
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_CZOOM_ZoomInputHeight   
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputWidthOfs 
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputHeightOfs

{{0x00,0x2A},{0x02,0x62}},                                
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInPre  
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInCap 
                       
//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
//Preview config[0] 640x480  7.5~30fps	
{{0x00,0x2A},{0x02,0xA6}},	
{{0x0F,0x12},{0x02,0x80}},	//REG_0TC_PCFG_usWidth              	
{{0x0F,0x12},{0x01,0xE0}},	//REG_0TC_PCFG_usHeight             	
{{0x0F,0x12},{0x00,0x05}},	//REG_0TC_PCFG_Format	                
{{0x0F,0x12},{0x4F,0x1A}},	//REG_0TC_PCFG_usMaxOut4KHzRate     	
{{0x0F,0x12},{0x4F,0x1A}},	//REG_0TC_PCFG_usMinOut4KHzRate     	
{{0x0F,0x12},{0x01,0x00}},	//REG_0TC_PCFG_OutClkPerPix88       	
{{0x0F,0x12},{0x03,0x00}},	//REG_0TC_PCFG_uBpp88               	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_PVIMask              	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_OIFMask              	
{{0x0F,0x12},{0x01,0xE0}},	//REG_0TC_PCFG_usJpegPacketSize     	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_usJpegTotalPackets   	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uClockInd            	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_usFrTimeType         	
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_PCFG_FrRateQualityType    	
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_PCFG_usMaxFrTimeMsecMult10	
{{0x0F,0x12},{0x01,0x4D}},	//REG_0TC_PCFG_usMinFrTimeMsecMult10	
{{0x00,0x2A},{0x02,0xD0}},	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uPrevMirror
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uCaptureMirror
	
//Capture Config[0] 2560x1920   7.5fps                           	
{{0x00,0x2A},{0x03,0x96}},                                     	
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_CCFG_uCaptureMode         
{{0x0F,0x12},{0x0A,0x00}},	//REG_0TC_CCFG_usWidth              
{{0x0F,0x12},{0x07,0x80}},	//REG_0TC_CCFG_usHeight             
{{0x0F,0x12},{0x00,0x05}},	//REG_0TC_CCFG_Format               
{{0x0F,0x12},{0x4F,0x1A}},	//REG_0TC_CCFG_usMaxOut4KHzRate     
{{0x0F,0x12},{0x4F,0x1A}},	//REG_0TC_CCFG_usMinOut4KHzRate     
{{0x0F,0x12},{0x01,0x00}},	//REG_0TC_CCFG_OutClkPerPix88       
{{0x0F,0x12},{0x03,0x00}},	//REG_0TC_CCFG_uBpp88               
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_PVIMask              
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_OIFMask              
{{0x0F,0x12},{0x01,0xE0}},	//REG_0TC_CCFG_usJpegPacketSize     
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_usJpegTotalPackets   
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_uClockInd            
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_usFrTimeType         
{{0x0F,0x12},{0x00,0x02}},	//REG_0TC_CCFG_FrRateQualityType    
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_CCFG_usMaxFrTimeMsecMult10
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_CCFG_usMinFrTimeMsecMult10

{{0x00,0x2A},{0x02,0x2C}},	
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_IPRM_InitParamsUpdated

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW                                                                        		

{{0x00,0x28},{0x70,0x00}},                                		                               		
{{0x00,0x2A},{0x02,0x66}},   
{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig        		
{{0x00,0x2A},{0x02,0x6A}},	                                   
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange     		
{{0x00,0x2A},{0x02,0x68}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged                             
{{0x00,0x2A},{0x02,0x6E}},                                          
{{0x0f,0x12},{0x00,0x00}},       //REG_TC_GP_ActiveCapConfig          
{{0x00,0x2A},{0x02,0x6A}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapOpenAfterChange           
{{0x00,0x2A},{0x02,0x70}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapConfigChanged   
    		
{{0x00,0x2A},{0x02,0x4E}},	                                   
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync           		
{{0x00,0x2A},{0x02,0x3E}},	                                   
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview           	
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged    	

//===================================================================================
// 22. ESD Check
//===================================================================================

{{0x00,0x28},{0x70,0x00}},	                                                                                    
{{0x00,0x2A},{0x01,0xA8}},	//ESD Check                                                                         
{{0x0f,0x12},{0xAA,0xAA}},

//===================================================================================
// 23. Brightness min/Max
//===================================================================================
{{0x00,0x28},{0x14,0x7C}},
{{0x0F,0x12},{0x01,0x80}}, //bp_uMaxBrightnessFactor	
{{0x00,0x28},{0x14,0x82}},
{{0x0F,0x12},{0x01,0x80}}, //bp_uMinBrightnessFactor	

//===================================================================================
// 24.ISSUE
//===================================================================================
//20110728 : Sequence Changed by image dev.(by J.M.Ahn)
//20110728 : ESD Check Register Address Change
//20110829 : TnP Changed by S.Y.Lee
//20120104 : init Parm Update sequence changed(by J.M.Ahn)
//20120201 : Flash ֺ Green Noise  setting (by J.M.Ahn)
//20120228 : Add Brightness Block

};                         	                         

#if 0
//for capture                                                                    	    
static struct regval_list sensor_qsxga_regs[] = { //qsxga: 2560*1920  5fps
/* capture setting */

//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0x00,0x2A},{0x03,0x96}},                                     	
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_CCFG_uCaptureMode         
{{0x0F,0x12},{0x0A,0x00}},	//REG_0TC_CCFG_usWidth              
{{0x0F,0x12},{0x07,0x80}},	//REG_0TC_CCFG_usHeight
{{0x00,0x2A},{0x03,0xAE}},                                     	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_uClockInd 
{{0x00,0x2A},{0x03,0xB4}},                                     	
{{0x0F,0x12},{0x07,0xD0}},	//REG_0TC_CCFG_usMaxFrTimeMsecMult10
{{0x0F,0x12},{0x07,0xD0}},	//REG_0TC_CCFG_usMinFrTimeMsecMult10

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//CAPTURE
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x6E}},
{{0x0F,0x12},{0x00,0x00}}, //REG_TC_GP_ActiveCapConfig
{{0x0F,0x12},{0x00,0x01}}, // /
{{0x00,0x2A},{0x02,0x42}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_EnableCapture
{{0x00,0x2A},{0x02,0x4E}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_NewConfigSync
{{0x00,0x2A},{0x02,0x44}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_EnableCaptureChanged

{{0xff,0xff},{0x00,0xc8}},  //delay 200ms
};
#else
//for capture                                                                    	    
static struct regval_list sensor_qsxga_regs[] = { //qsxga: 2560*1920  7fps
/* capture setting */

//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0x00,0x2A},{0x03,0x96}},                                     	
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_CCFG_uCaptureMode         
{{0x0F,0x12},{0x0A,0x00}},	//REG_0TC_CCFG_usWidth              
{{0x0F,0x12},{0x07,0x80}},	//REG_0TC_CCFG_usHeight
{{0x00,0x2A},{0x03,0xAE}},                                     	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_uClockInd            
{{0x00,0x2A},{0x03,0xB4}},      
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_CCFG_usMaxFrTimeMsecMult10
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_CCFG_usMinFrTimeMsecMult10

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//CAPTURE
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x6E}},
{{0x0F,0x12},{0x00,0x00}}, //REG_TC_GP_ActiveCapConfig
{{0x0F,0x12},{0x00,0x01}}, // /
{{0x00,0x2A},{0x02,0x42}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_EnableCapture
{{0x00,0x2A},{0x02,0x4E}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_NewConfigSync
{{0x00,0x2A},{0x02,0x44}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_EnableCaptureChanged

{{0xff,0xff},{0x00,0xc8}},  //delay 200ms
};
#endif

static struct regval_list sensor_qxga_regs[] = { //qxga: 2048*1536
/* capture setting */

//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0x00,0x2A},{0x03,0x96}},                                     	
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_CCFG_uCaptureMode         
{{0x0F,0x12},{0x08,0x00}},	//REG_0TC_CCFG_usWidth              
{{0x0F,0x12},{0x06,0x00}},	//REG_0TC_CCFG_usHeight
{{0x00,0x2A},{0x03,0xAE}},                                     	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_CCFG_uClockInd    
{{0x00,0x2A},{0x03,0xB4}},                                     	
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_CCFG_usMaxFrTimeMsecMult10
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_CCFG_usMinFrTimeMsecMult10

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//CAPTURE
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x6E}},
{{0x0F,0x12},{0x00,0x00}}, //REG_TC_GP_ActiveCapConfig
{{0x0F,0x12},{0x00,0x01}}, // /
{{0x00,0x2A},{0x02,0x42}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_EnableCapture
{{0x00,0x2A},{0x02,0x4E}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_NewConfigSync
{{0x00,0x2A},{0x02,0x44}}, // 
{{0x0F,0x12},{0x00,0x01}}, // /#REG_TC_GP_EnableCaptureChanged

{{0xff,0xff},{0x00,0xc8}},  //delay 200ms
};                                      


//for video
static struct regval_list sensor_1080p_regs[] = { //1080: 1920*1080 
/* preview setting */

{{0x00,0x2A},{0x18,0xAC}},
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsBin
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsNoBin
{{0x0F,0x12},{0x05,0xC0}}, //senHal_uMinColsBin
{{0x0F,0x12},{0x0A,0x96}}, //05C0 //senHal_uMinColsNoBin

//==================================================================================
// 19.Input Size Setting
//==================================================================================
//Input Size
{{0x00,0x2A},{0x02,0x50}},
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_PrevReqInputWidth
{{0x0F,0x12},{0x04,0x38}},	//REG_TC_GP_PrevReqInputHeight
{{0x0F,0x12},{0x01,0x4E}},	//REG_TC_GP_PrevInputWidthOfs
{{0x0F,0x12},{0x01,0xB0}},	//REG_TC_GP_PrevInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_CapReqInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_CapReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_CapInputWidthOfs
{{0x0F,0x12},{0x00,0x0C}},	//REG_TC_GP_CapInputHeightOfs

{{0x00,0x2A},{0x04,0x94}},
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_PZOOM_ZoomInputWidth
{{0x0F,0x12},{0x04,0x38}},	//REG_TC_PZOOM_ZoomInputHeight
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputWidthOfs
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_CZOOM_ZoomInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_CZOOM_ZoomInputHeight
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputWidthOfs
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputHeightOfs

{{0x00,0x2A},{0x02,0x62}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInPre
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInCap

//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xA6}},
{{0x0F,0x12},{0x07,0x80}},	//280 REG_0TC_PCFG_usWidth
{{0x0F,0x12},{0x04,0x38}},	//1E0 REG_0TC_PCFG_usHeight
{{0x00,0x2A},{0x02,0xBC}},
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_PCFG_uClockInd
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_PCFG_usFrTimeType
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_FrRateQualityType
{{0x0F,0x12},{0x02,0x9A}},	//REG_0TC_PCFG_usMaxFrTimeMsecMult10
{{0x0F,0x12},{0x02,0x9A}},	//REG_0TC_PCFG_usMinFrTimeMsecMult10

{{0x00,0x2A},{0x02,0x2C}},	
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_IPRM_InitParamsUpdated

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x66}},
{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
{{0x00,0x2A},{0x02,0x6A}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
{{0x00,0x2A},{0x02,0x68}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
{{0x00,0x2A},{0x02,0x6E}},                                          
{{0x0f,0x12},{0x00,0x00}},       //REG_TC_GP_ActiveCapConfig          
{{0x00,0x2A},{0x02,0x6A}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapOpenAfterChange           
{{0x00,0x2A},{0x02,0x70}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapConfigChanged   
    		
{{0x00,0x2A},{0x02,0x4E}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
{{0x00,0x2A},{0x02,0x3E}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged


{{0xff,0xff},{0x00,0xc8}},  //delay 200ms
};

static struct regval_list sensor_720p_regs[] = { //1280*720
/* preview setting */

{{0x00,0x2A},{0x18,0xAC}},
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsBin
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsNoBin
{{0x0F,0x12},{0x05,0xC0}}, //senHal_uMinColsBin
{{0x0F,0x12},{0x0A,0x96}}, //05C0 //senHal_uMinColsNoBin

//==================================================================================
// 19.Input Size Setting
//==================================================================================
//Input Size
{{0x00,0x2A},{0x02,0x50}},
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_PrevReqInputWidth 
{{0x0F,0x12},{0x05,0xA0}},	//REG_TC_GP_PrevReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_PrevInputWidthOfs 
{{0x0F,0x12},{0x00,0xFC}},	//REG_TC_GP_PrevInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_CapReqInputWidth  
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_CapReqInputHeight 
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_CapInputWidthOfs  
{{0x0F,0x12},{0x00,0x0C}},	//REG_TC_GP_CapInputHeightOfs 

{{0x00,0x2A},{0x04,0x94}},
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_PZOOM_ZoomInputWidth
{{0x0F,0x12},{0x05,0xA0}},	//REG_TC_PZOOM_ZoomInputHeight
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputWidthOfs    
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputHeightOfs   
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_CZOOM_ZoomInputWidth       
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_CZOOM_ZoomInputHeight      
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputWidthOfs    
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputHeightOfs  

{{0x00,0x2A},{0x02,0x62}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInPre
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInCap

//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xA6}},
{{0x0F,0x12},{0x05,0x00}},	//280 REG_0TC_PCFG_usWidth
{{0x0F,0x12},{0x02,0xD0}},	//1E0 REG_0TC_PCFG_usHeight
{{0x00,0x2A},{0x02,0xBC}},
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uClockInd
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_PCFG_usFrTimeType
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_PCFG_FrRateQualityType
{{0x0F,0x12},{0x01,0x4D}},	//REG_0TC_PCFG_usMaxFrTimeMsecMult10
{{0x0F,0x12},{0x01,0x4D}},	//REG_0TC_PCFG_usMinFrTimeMsecMult10

{{0x00,0x2A},{0x02,0x2C}},	
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_IPRM_InitParamsUpdated

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x66}},
{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
{{0x00,0x2A},{0x02,0x6A}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
{{0x00,0x2A},{0x02,0x68}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged                             
{{0x00,0x2A},{0x02,0x6E}},                                          
{{0x0f,0x12},{0x00,0x00}},       //REG_TC_GP_ActiveCapConfig          
{{0x00,0x2A},{0x02,0x6A}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapOpenAfterChange           
{{0x00,0x2A},{0x02,0x70}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapConfigChanged   
    		
{{0x00,0x2A},{0x02,0x4E}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
{{0x00,0x2A},{0x02,0x3E}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged

{{0xff,0xff},{0x00,0xc8}},  //delay 200ms
};

static struct regval_list sensor_vga_regs[] = { //VGA:  640*480
#if 0  //vga 7.5~30fps
/* preview setting */

{{0x00,0x2A},{0x18,0xAC}},
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsBin
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsNoBin
{{0x0F,0x12},{0x05,0xC0}}, //senHal_uMinColsBin
{{0x0F,0x12},{0x05,0xC0}}, //senHal_uMinColsNoBin

//==================================================================================
// 19.Input Size Setting
//==================================================================================
//Input Size
{{0x00,0x2A},{0x02,0x50}},
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_PrevReqInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_PrevReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_PrevInputWidthOfs
{{0x0F,0x12},{0x00,0x0C}},	//REG_TC_GP_PrevInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_CapReqInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_CapReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_CapInputWidthOfs
{{0x0F,0x12},{0x00,0x0C}},	//REG_TC_GP_CapInputHeightOfs

{{0x00,0x2A},{0x04,0x94}},
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_PZOOM_ZoomInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_PZOOM_ZoomInputHeight
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputWidthOfs
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_CZOOM_ZoomInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_CZOOM_ZoomInputHeight
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputWidthOfs
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputHeightOfs

{{0x00,0x2A},{0x02,0x62}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInPre
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInCap

//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xA6}},	
{{0x0F,0x12},{0x02,0x80}},	//REG_0TC_PCFG_usWidth              	
{{0x0F,0x12},{0x01,0xE0}},	//REG_0TC_PCFG_usHeight              	
{{0x00,0x2A},{0x02,0xBC}},	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uClockInd            	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_usFrTimeType         	
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_PCFG_FrRateQualityType    	
{{0x0F,0x12},{0x05,0x35}},	//REG_0TC_PCFG_usMaxFrTimeMsecMult10	
{{0x0F,0x12},{0x01,0x4D}},	//REG_0TC_PCFG_usMinFrTimeMsecMult10	

{{0x00,0x2A},{0x02,0x2C}},	
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_IPRM_InitParamsUpdated

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},                                		                               		
{{0x00,0x2A},{0x02,0x66}},   
{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig        		
{{0x00,0x2A},{0x02,0x6A}},	                                   
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange     		
{{0x00,0x2A},{0x02,0x68}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged                             
{{0x00,0x2A},{0x02,0x6E}},                                          
{{0x0f,0x12},{0x00,0x00}},       //REG_TC_GP_ActiveCapConfig          
{{0x00,0x2A},{0x02,0x6A}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapOpenAfterChange           
{{0x00,0x2A},{0x02,0x70}},	                                         
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapConfigChanged   
    		
{{0x00,0x2A},{0x02,0x4E}},	                                   
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync           		
{{0x00,0x2A},{0x02,0x3E}},	                                   
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview           	
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged   

{{0xff,0xff},{0x00,0xc8}},  //delay 200ms
#else			//vga fix 30fps
/* preview setting */

{{0x00,0x2A},{0x18,0xAC}},
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsBin
{{0x0F,0x12},{0x00,0x60}}, //senHal_uAddColsNoBin
{{0x0F,0x12},{0x05,0xC0}}, //senHal_uMinColsBin
{{0x0F,0x12},{0x05,0xC0}}, //05C0 //senHal_uMinColsNoBin

//==================================================================================
// 19.Input Size Setting
//==================================================================================
//Input Size
{{0x00,0x2A},{0x02,0x50}},
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_PrevReqInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_PrevReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_PrevInputWidthOfs
{{0x0F,0x12},{0x00,0x0C}},	//REG_TC_GP_PrevInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_GP_CapReqInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_GP_CapReqInputHeight
{{0x0F,0x12},{0x00,0x10}},	//REG_TC_GP_CapInputWidthOfs
{{0x0F,0x12},{0x00,0x0C}},	//REG_TC_GP_CapInputHeightOfs

{{0x00,0x2A},{0x04,0x94}},
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_PZOOM_ZoomInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_PZOOM_ZoomInputHeight
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputWidthOfs
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_PZOOM_ZoomInputHeightOfs
{{0x0F,0x12},{0x0A,0x00}},	//REG_TC_CZOOM_ZoomInputWidth
{{0x0F,0x12},{0x07,0x80}},	//REG_TC_CZOOM_ZoomInputHeight
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputWidthOfs
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_CZOOM_ZoomInputHeightOfs

{{0x00,0x2A},{0x02,0x62}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInPre
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_bUseReqInputInCap

//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xA6}},	
{{0x0F,0x12},{0x02,0x80}},	//280 REG_0TC_PCFG_usWidth
{{0x0F,0x12},{0x01,0xE0}},	//1E0 REG_0TC_PCFG_usHeight
{{0x00,0x2A},{0x02,0xBC}},	
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uClockInd
{{0x0F,0x12},{0x00,0x02}},	//REG_0TC_PCFG_usFrTimeType
{{0x0F,0x12},{0x00,0x01}},	//REG_0TC_PCFG_FrRateQualityType
{{0x0F,0x12},{0x01,0x4D}},	//REG_0TC_PCFG_usMaxFrTimeMsecMult10
{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_usMinFrTimeMsecMult10

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x66}},
{{0x0f,0x12},{0x00,0x00}}, //REG_TC_GP_ActivePrevConfig
{{0x00,0x2A},{0x02,0x6A}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
{{0x00,0x2A},{0x02,0x68}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
{{0x00,0x2A},{0x02,0x6E}},
{{0x0f,0x12},{0x00,0x00}}, //REG_TC_GP_ActiveCapConfig
{{0x00,0x2A},{0x02,0x6A}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapOpenAfterChange
{{0x00,0x2A},{0x02,0x70}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_CapConfigChanged

{{0x00,0x2A},{0x02,0x4E}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
{{0x00,0x2A},{0x02,0x3E}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged

//{{0xff,0xff},{0x00,150}},
#endif

//{{0xFC,0xFC},{0xD0,0x00}},
//{{0x00,0x28},{0x70,0x00}},
//{{0x00,0x2A},{0x02,0x36}},
//{{0x0F,0x12},{0x00,0x40}},   //Control value
};

#if 0
//misc
static struct regval_list sensor_oe_disable_regs[] = {
};
#endif

#if 0
//stuff about auto focus
static struct regval_list sensor_af_fw_regs[] = {
};
#endif

static struct regval_list sensor_ae_awb_lockon_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x2C,0x5E}}, // /* AE Lock On */
{{0x0F,0x12},{0x00,0x00}},
};

static struct regval_list sensor_ae_awb_lockoff_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x2C,0x5E}}, // /* AE Lock Off */
{{0x0F,0x12},{0x00,0x01}},
};

static struct regval_list sensor_af_single_trig_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x8E}},
{{0x0F,0x12},{0x00,0x00}},
{{0x00,0x2A},{0x02,0x8C}},
{{0x0F,0x12},{0x00,0x05}},
};

static struct regval_list sensor_af_continueous_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x8E}},
{{0x0F,0x12},{0x00,0x00}},
{{0x00,0x2A},{0x02,0x8C}},
{{0x0F,0x12},{0x00,0x06}},
};

static struct regval_list sensor_af_infinity_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x8E}}, // write [7000 028E, REG_TC_AF_AfCmdParam]
{{0x0F,0x12},{0x00,0x00}}, // write lens position from 0000 to 00FF. 0000 means infinity and 00FF means macro.
{{0x00,0x2A},{0x02,0x8C}},
{{0x0F,0x12},{0x00,0x04}}, // write [7000 028C, REG_TC_AF_AfCmd] = 0004 , manual AF.

{{0xff,0xff},{0x00,150}},	// delay 150ms

{{0x00,0x2A},{0x16,0x48}},
{{0x0F,0x12},{0x90,0x02}},
};

static struct regval_list sensor_af_macro_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x8E}}, // write [7000 028E, REG_TC_AF_AfCmdParam]
{{0x0F,0x12},{0x00,0xD0}}, // write lens position from 0000 to 00FF. 0000 means infinity and 00FF means macro.
{{0x00,0x2A},{0x02,0x8C}},
{{0x0F,0x12},{0x00,0x04}}, // write [7000 028C, REG_TC_AF_AfCmd] = 0004 , manual AF.

{{0xff,0xff},{0x00,150}},	// delay 150ms

{{0x00,0x2A},{0x16,0x48}},
{{0x0F,0x12},{0x90,0x02}},
};



/*
 * The white balance settings
 * Here only tune the R G B channel gain. 
 * The white balance enalbe bit is modified in sensor_s_autowb and sensor_s_wb
 */
static struct regval_list sensor_wb_auto_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x04,0xE6}},
{{0x0F,0x12},{0x07,0x7F}},
};

static struct regval_list sensor_wb_cloud_regs[] = {	
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x04,0xE6}},
{{0x0F,0x12},{0x07,0x77}},
{{0x00,0x2A},{0x04,0xBA}},
{{0x0F,0x12},{0x06,0xD0}},
{{0x00,0x2A},{0x04,0xBE}},
{{0x0F,0x12},{0x03,0xD0}},
{{0x00,0x2A},{0x04,0xC2}},
{{0x0F,0x12},{0x05,0x20}},
{{0x00,0x2A},{0x04,0xC6}},
{{0x0F,0x12},{0x00,0x01}},
};

static struct regval_list sensor_wb_daylight_regs[] = {
	//tai yang guang
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x04,0xE6}},
{{0x0F,0x12},{0x07,0x77}},
{{0x00,0x2A},{0x04,0xBA}},
{{0x0F,0x12},{0x06,0x20}},
{{0x00,0x2A},{0x04,0xBE}},
{{0x0F,0x12},{0x03,0xD0}},
{{0x00,0x2A},{0x04,0xC2}},
{{0x0F,0x12},{0x05,0x80}},
{{0x00,0x2A},{0x04,0xC6}},
{{0x0F,0x12},{0x00,0x01}},
};

static struct regval_list sensor_wb_incandescence_regs[] = {
	//bai re guang
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x04,0xE6}},
{{0x0F,0x12},{0x07,0x77}},
{{0x00,0x2A},{0x04,0xBA}},
{{0x0F,0x12},{0x04,0x30}},
{{0x00,0x2A},{0x04,0xBE}},
{{0x0F,0x12},{0x03,0xD0}},
{{0x00,0x2A},{0x04,0xC2}},
{{0x0F,0x12},{0x09,0xF0}},
{{0x00,0x2A},{0x04,0xC6}},
{{0x0F,0x12},{0x00,0x01}},
};

static struct regval_list sensor_wb_fluorescent_regs[] = {
	//ri guang deng
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x04,0xE6}}, 
{{0x0F,0x12},{0x07,0x77}},
{{0x00,0x2A},{0x04,0xBA}},
{{0x0F,0x12},{0x06,0xC0}},
{{0x00,0x2A},{0x04,0xBE}},
{{0x0F,0x12},{0x04,0x40}},
{{0x00,0x2A},{0x04,0xC2}},
{{0x0F,0x12},{0x07,0xA0}},
{{0x00,0x2A},{0x04,0xC6}},
{{0x0F,0x12},{0x00,0x01}},
};

static struct regval_list sensor_wb_tungsten_regs[] = {
	//wu si deng
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x04,0xE6}},
{{0x0F,0x12},{0x07,0x77}},
{{0x00,0x2A},{0x04,0xBA}},
{{0x0F,0x12},{0x05,0x40}},
{{0x00,0x2A},{0x04,0xBE}},
{{0x0F,0x12},{0x03,0xD0}},
{{0x00,0x2A},{0x04,0xC2}},
{{0x0F,0x12},{0x08,0xF0}},
{{0x00,0x2A},{0x04,0xC6}},
{{0x0F,0x12},{0x00,0x01}},
};

/*
 * The color effect settings
 */
static struct regval_list sensor_colorfx_none_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3C}},
{{0x0F,0x12},{0x00,0x00}},	//REG_TC_GP_SpecialEffects
};

static struct regval_list sensor_colorfx_bw_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3C}},
{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_SpecialEffects
};

static struct regval_list sensor_colorfx_sepia_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3C}},
{{0x0F,0x12},{0x00,0x04}},	//REG_TC_GP_SpecialEffects
};

static struct regval_list sensor_colorfx_negative_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3C}},
{{0x0F,0x12},{0x00,0x03}},	//REG_TC_GP_SpecialEffects
};

static struct regval_list sensor_colorfx_emboss_regs[] = {

};

static struct regval_list sensor_colorfx_sketch_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3C}},
{{0x0F,0x12},{0x00,0x09}},	//REG_TC_GP_SpecialEffects
};

static struct regval_list sensor_colorfx_sky_blue_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3C}},
{{0x0F,0x12},{0x00,0x07}},	//REG_TC_GP_SpecialEffects
};

static struct regval_list sensor_colorfx_grass_green_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3C}},
{{0x0F,0x12},{0x00,0x09}},	//REG_TC_GP_SpecialEffects
};

static struct regval_list sensor_colorfx_skin_whiten_regs[] = {
//NULL
};

static struct regval_list sensor_colorfx_vivid_regs[] = {
//NULL
};

/*
 * The power frequency
 */
static struct regval_list sensor_flicker_50hz_regs[] = {
{{0x00,0x2A},{0x04,0xE6}},
{{0x0F,0x12},{0x07,0x5F}},	//REG_TC_DBG 7F: 60Hz  5F:50Hz				
{{0x00,0x2a},{0x04,0xd6}},
{{0x0F,0x12},{0x00,0x01}},	//REG_SF_USER_FlickerQuant 01h : 50hz  02h : 60hz				
{{0x0F,0x12},{0x00,0x01}},	//REG_SF_USER_FlickerQuantChanged
};

static struct regval_list sensor_flicker_60hz_regs[] = {
{{0x00,0x2A},{0x04,0xE6}},														
{{0x0F,0x12},{0x07,0x5F}},	//REG_TC_DBG 7F: 60Hz  5F:50Hz				
{{0x00,0x2a},{0x04,0xd6}},
{{0x0f,0x12},{0x00,0x02}},	//REG_SF_USER_FlickerQuant 01h : 50hz  02h : 60hz				
{{0x0F,0x12},{0x00,0x01}},	//REG_SF_USER_FlickerQuantChanged
};

/*
 * The brightness setttings
 */
static struct regval_list sensor_brightness_neg4_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x24}},
};

static struct regval_list sensor_brightness_neg3_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x2A}},
};

static struct regval_list sensor_brightness_neg2_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x30}},
};

static struct regval_list sensor_brightness_neg1_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x36}},
};

static struct regval_list sensor_brightness_zero_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x3C}},
};

static struct regval_list sensor_brightness_pos1_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x47}},
};

static struct regval_list sensor_brightness_pos2_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x52}},
};

static struct regval_list sensor_brightness_pos3_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x57}},
};

static struct regval_list sensor_brightness_pos4_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x14,0x84}},
{{0x0F,0x12},{0x00,0x68}},
};

/*
 * The contrast setttings
 */
static struct regval_list sensor_contrast_neg4_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0xFF,0x80}},   //Control value
};

static struct regval_list sensor_contrast_neg3_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0xFF,0xA0}},   //Control value
};

static struct regval_list sensor_contrast_neg2_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0xFF,0xC0}},   //Control value
};

static struct regval_list sensor_contrast_neg1_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0xFF,0xE0}},   //Control value
};

static struct regval_list sensor_contrast_zero_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0x00,0x00}},   //Control value
};

static struct regval_list sensor_contrast_pos1_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0x00,0x20}},   //Control value
};

static struct regval_list sensor_contrast_pos2_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0x00,0x40}},   //Control value
};

static struct regval_list sensor_contrast_pos3_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0x00,0x60}},   //Control value
};

static struct regval_list sensor_contrast_pos4_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x32}},
{{0x0F,0x12},{0x00,0x80}},   //Control value
};

/*
 * The saturation setttings
 */
static struct regval_list sensor_saturation_neg4_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0xFF,0x80}},   //Control value
};

static struct regval_list sensor_saturation_neg3_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0xFF,0xA0}},   //Control value
};

static struct regval_list sensor_saturation_neg2_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0xFF,0xC0}},   //Control value
};

static struct regval_list sensor_saturation_neg1_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0xFF,0xE0}},   //Control value
};

static struct regval_list sensor_saturation_zero_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0x00,0x00}},   //Control value
};

static struct regval_list sensor_saturation_pos1_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0x00,0x20}},   //Control value
};

static struct regval_list sensor_saturation_pos2_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0x00,0x40}},   //Control value
};

static struct regval_list sensor_saturation_pos3_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0x00,0x60}},   //Control value
};

static struct regval_list sensor_saturation_pos4_regs[] = {
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x34}},
{{0x0F,0x12},{0x00,0x80}},   //Control value
};

/*
 * The exposure target setttings
 */
static struct regval_list sensor_ev_neg4_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x00,0x80}},   //Control value
};

static struct regval_list sensor_ev_neg3_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x00,0xA0}},   //Control value
};

static struct regval_list sensor_ev_neg2_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x00,0xC0}},   //Control value
};

static struct regval_list sensor_ev_neg1_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x00,0xE0}},   //Control value
};                     

static struct regval_list sensor_ev_zero_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x01,0x00}},   //Control value
};

static struct regval_list sensor_ev_pos1_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x01,0x20}},   //Control value
};

static struct regval_list sensor_ev_pos2_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x01,0x40}},   //Control value
};

static struct regval_list sensor_ev_pos3_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x01,0x60}},   //Control value
};

static struct regval_list sensor_ev_pos4_regs[] = {
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0x3A}},
{{0x0F,0x12},{0x01,0x80}},   //Control value
};


/*
 * Here we'll try to encapsulate the changes for just the output
 * video format.
 * 
 */


static struct regval_list sensor_fmt_yuv422_yuyv[] = {	
	//YUYV
//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFF,0xFF},{0x00,0x00}},
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xB4}},	
{{0x0F,0x12},{0x00,0x40}},	//REG_0TC_PCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11        	
{{0x00,0x2A},{0x03,0xA6}},	
{{0x0F,0x12},{0x00,0x40}},	//REG_0TC_CCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11  

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
//{{0xFC,0xFC},{0xD0,0x00}},
//{{0x00,0x28},{0x70,0x00}},
//{{0x00,0x2A},{0x02,0x66}},
//{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
//{{0x00,0x2A},{0x02,0x6A}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
//{{0x00,0x2A},{0x02,0x4E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
//{{0x00,0x2A},{0x02,0x68}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
//{{0x00,0x2A},{0x02,0x3E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged	
//
//{{0xFF,0xFF},{0x00,0x20}},	
   
};

static struct regval_list sensor_fmt_yuv422_yvyu[] = {
//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFF,0xFF},{0x00,0x00}},
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xB4}},	
{{0x0F,0x12},{0x00,0x60}},	//REG_0TC_PCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11        	
{{0x00,0x2A},{0x03,0xA6}},	
{{0x0F,0x12},{0x00,0x60}},	//REG_0TC_CCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11  

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
//{{0xFC,0xFC},{0xD0,0x00}},
//{{0x00,0x28},{0x70,0x00}},
//{{0x00,0x2A},{0x02,0x66}},
//{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
//{{0x00,0x2A},{0x02,0x6A}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
//{{0x00,0x2A},{0x02,0x4E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
//{{0x00,0x2A},{0x02,0x68}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
//{{0x00,0x2A},{0x02,0x3E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged	  
//
//{{0xFF,0xFF},{0x00,0x20}},	
};

static struct regval_list sensor_fmt_yuv422_vyuy[] = {
//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFF,0xFF},{0x00,0x00}},
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xB4}},	
{{0x0F,0x12},{0x00,0x70}},	//REG_0TC_PCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11        	
{{0x00,0x2A},{0x03,0xA6}},	
{{0x0F,0x12},{0x00,0x70}},	//REG_0TC_CCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11  

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
//{{0xFC,0xFC},{0xD0,0x00}},
//{{0x00,0x28},{0x70,0x00}},
//{{0x00,0x2A},{0x02,0x66}},
//{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
//{{0x00,0x2A},{0x02,0x6A}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
//{{0x00,0x2A},{0x02,0x4E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
//{{0x00,0x2A},{0x02,0x68}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
//{{0x00,0x2A},{0x02,0x3E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged	
// 
//{{0xFF,0xFF},{0x00,0x20}},	  
};

static struct regval_list sensor_fmt_yuv422_uyvy[] = {
//==================================================================================
// 20.Preview & Capture Configration Setting
//==================================================================================
{{0xFF,0xFF},{0x00,0x00}},
{{0xFC,0xFC},{0xD0,0x00}},
{{0x00,0x28},{0x70,0x00}},
{{0x00,0x2A},{0x02,0xB4}},	
{{0x0F,0x12},{0x00,0x50}},	//REG_0TC_PCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11        	
{{0x00,0x2A},{0x03,0xA6}},	
{{0x0F,0x12},{0x00,0x50}},	//REG_0TC_CCFG_PVIMask    [5:4] YUYV 00 / UYVY 01 / YVYU 10 / VYUV 11

//==================================================================================
// 21.Select Cofigration Display
//==================================================================================
//PREVIEW
//{{0xFC,0xFC},{0xD0,0x00}},
//{{0x00,0x28},{0x70,0x00}},
//{{0x00,0x2A},{0x02,0x66}},
//{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
//{{0x00,0x2A},{0x02,0x6A}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
//{{0x00,0x2A},{0x02,0x4E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
//{{0x00,0x2A},{0x02,0x68}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
//{{0x00,0x2A},{0x02,0x3E}},
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
//{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged	
//
//{{0xFF,0xFF},{0x00,0x20}},	
};

//static struct regval_list sensor_fmt_raw[] = {
//	
//};



/*
 * Low-level register I/O.
 *
 */


/*
 * On most platforms, we'd rather do straight i2c I/O.
 */
#if 0
static int sensor_read(struct v4l2_subdev *sd, unsigned char *reg,
		unsigned char *value)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	u8 data[REG_STEP];
	struct i2c_msg msg;
	int ret,i;
	
	for(i = 0; i < REG_ADDR_STEP; i++)
		data[i] = reg[i];
	
	for(i = REG_ADDR_STEP; i < REG_STEP; i++)
		data[i] = 0xff;
	/*
	 * Send out the register address...
	 */
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_ADDR_STEP;
	msg.buf = data;
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret < 0) {
		csi_dev_err("Error %d on register write\n", ret);
		return ret;
	}
	/*
	 * ...then read back the result.
	 */
	
	msg.flags = I2C_M_RD;
	msg.len = REG_DATA_STEP;
	msg.buf = &data[REG_ADDR_STEP];
	
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret >= 0) {
		for(i = 0; i < REG_DATA_STEP; i++)
			value[i] = data[i+REG_ADDR_STEP];
		ret = 0;
	}
	else {
		csi_dev_err("Error %d on register read\n", ret);
	}
	return ret;
}
#endif

static int sensor_read_im(struct v4l2_subdev *sd, unsigned int addr,
		unsigned int *value)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	u8 data[REG_STEP];
	struct i2c_msg msg;
	int ret,i,j;
	
	for(i = 0, j = REG_ADDR_STEP-1; i < REG_ADDR_STEP; i++,j--)
		data[i] = (addr&(0xff<<(j*8)))>>(j*8);
	
	for(i = REG_ADDR_STEP; i < REG_STEP; i++)
		data[i] = 0xff;
	/*
	 * Send out the register address...
	 */
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_ADDR_STEP;
	msg.buf = data;
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret < 0) {
		csi_dev_err("Error %d on register write\n", ret);
		return ret;
	}
	/*
	 * ...then read back the result.
	 */
	
	msg.flags = I2C_M_RD;
	msg.len = REG_DATA_STEP;
	msg.buf = &data[REG_ADDR_STEP];
	
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret >= 0) {
		for(i = 0,j = REG_DATA_STEP-1; i < REG_DATA_STEP; i++,j--)
			*((unsigned char*)(value)+j) = data[i+REG_ADDR_STEP];
//		*value = data[2]*256+data[3];
		ret = 0;
	}
	else {
		csi_dev_err("Error %d on register read\n", ret);
	}
	return ret;
}


static int sensor_write(struct v4l2_subdev *sd, unsigned char *reg,
		unsigned char *value)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct i2c_msg msg;
	unsigned char data[REG_STEP];
	int ret,i;
	
	for(i = 0; i < REG_ADDR_STEP; i++)
			data[i] = reg[i];
	for(i = REG_ADDR_STEP; i < REG_STEP; i++)
			data[i] = value[i-REG_ADDR_STEP];
	
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_STEP;
	msg.buf = data;

	
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret > 0) {
		ret = 0;
	}
	else if (ret < 0) {
		csi_dev_err("addr = 0x%4x, value = 0x%4x\n ",reg[0]*256+reg[1],value[0]*256+value[1]);
		csi_dev_err("sensor_write error!\n");
	}
	return ret;
}

static int sensor_write_im(struct v4l2_subdev *sd, unsigned int addr,
		unsigned int value)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct i2c_msg msg;
	unsigned char data[REG_STEP];
	int ret;
	unsigned i,j;
	
	for(i = 0, j = REG_ADDR_STEP-1; i < REG_ADDR_STEP; i++,j--) {
			data[i] = (addr&(0xff<<(j*8)))>>(j*8);
	}
			
	for(i = REG_ADDR_STEP,j = REG_DATA_STEP-1; i < REG_STEP; i++,j--)
			data[i] = *((unsigned char*)(&value)+j);
	
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_STEP;
	msg.buf = data;

	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret > 0) {
		ret = 0;
	}
	else if (ret < 0) {
		csi_dev_err("addr = 0x%4x, value = 0x%4x\n ",addr,value);
		csi_dev_err("sensor_write error!\n");
	}
	return ret;
}


/*
 * Write a list of register settings;
 */
static int sensor_write_array(struct v4l2_subdev *sd, struct regval_list *vals , uint size)
{
	int i,ret;
//	unsigned char rd;
	if (size == 0)
		return -EINVAL;

	for(i = 0; i < size ; i++)
	{
		if(vals->reg_num[0] == 0xff && vals->reg_num[1] == 0xff) {
			mdelay(vals->value[0]*256+vals->value[1]);
		}	
		else {	
			ret = sensor_write(sd, vals->reg_num, vals->value);
			if (ret < 0)
			{
				csi_dev_err("sensor_write_err!\n");
				return ret;
			}
		}
		vals++;
	}
	
	return 0;
}

#if 0
static int sensor_write_continuous(struct v4l2_subdev *sd, int addr, char vals[] , uint size)
{
	int i,ret;
	struct regval_list reg_addr;
	
	if (size == 0)
		return -EINVAL;
	
	for(i = 0; i < size ; i++)
	{
		reg_addr.reg_num[0] = (addr&0xff00)>>8;
		reg_addr.reg_num[1] = (addr&0x00ff);
		
		ret = sensor_write(sd, reg_addr.reg_num, &vals[i]);
		if (ret < 0)
		{
			csi_dev_err("sensor_write_err!\n");
			return ret;
		}
		addr++;
	}
	
	return 0;
}
#endif

/* stuff about exposure when capturing image */

static int sensor_set_exposure(struct v4l2_subdev *sd)
{	
	return 0;
}

/* stuff about auto focus */



static int sensor_download_af_fw(struct v4l2_subdev *sd)
{
  return 0;
}

static int sensor_ae_awb_lockon(struct v4l2_subdev *sd)
{
	int ret;
	
	ret = sensor_write_array(sd, sensor_ae_awb_lockon_regs , ARRAY_SIZE(sensor_ae_awb_lockon_regs));
	if(ret < 0)
		csi_dev_err("sensor_ae_awb_lockon error!\n"); 
	
	return ret;
};

static int sensor_ae_awb_lockoff(struct v4l2_subdev *sd)
{
	int ret;
	
	ret = sensor_write_array(sd, sensor_ae_awb_lockoff_regs , ARRAY_SIZE(sensor_ae_awb_lockoff_regs));
	if(ret < 0)
		csi_dev_err("sensor_ae_awb_lockoff error!\n"); 
	
	return ret;
};

static int sensor_g_single_af(struct v4l2_subdev *sd)
{
	unsigned int rdval;
	int ret;
	struct sensor_info *info = to_state(sd);
	
	csi_dev_dbg("sensor_g_single_af\n");
	csi_dev_dbg("info->coarse_af_pd = %d\n",info->coarse_af_pd);
	if(info->coarse_af_pd != 1) {
		//wait for 1st af complete
	  rdval = 0xffff;
	  sensor_write_im(sd, 0x002c , 0x7000);
		sensor_write_im(sd, 0x002e , 0x2eee);
	  ret =  sensor_read_im(sd, 0x0f12 , &rdval);
		if (ret < 0)
		{
			csi_dev_err("sensor_g_single_af read error !\n");
			ret = -EAGAIN;
			goto af_out;
		}
	
	  if(rdval == 0x0001 ) {
			csi_dev_dbg("Single AF 1st is busy,value = 0x%4x\n",rdval);
			msleep(50);
			info->coarse_af_pd = 0;
			return EBUSY;
		} else if (rdval == 0x0002) {
			//focus ok
			info->coarse_af_pd = 1;
			csi_dev_print("Single AF 1st is complete,value = 0x%4x\n",rdval);
		} else {
			csi_dev_print("Single AF 1st is failed,value = 0x%4x\n",rdval);
			info->focus_status = 0;	//idle
			info->coarse_af_pd = 2;
			ret = EFAULT;
			goto af_out;
		}
	}
	
	//wait for 2nd af complete
	rdval = 0xffff;
	sensor_write_im(sd, 0x002c , 0x7000);
	sensor_write_im(sd, 0x002e , 0x2207);
	ret =  sensor_read_im(sd, 0x0f12 , &rdval);
	if (ret < 0)
	{
		csi_dev_err("sensor_g_single_af read error !\n");
		goto af_out;
	}
	
//	while((rdval&0xff00)!=0x0000){
//		sensor_read_im(sd, 0x0f12 , &rdval);
//		msleep(50);
//	}

	if((rdval&0xff00)!=0x0000) {
		csi_dev_dbg("Single AF 2nd is busy,value = 0x%4x\n",rdval);
		return EBUSY;
	}
	
	csi_dev_print("Single AF 2nd is complete,value = 0x%4x\n",rdval);
	info->focus_status = 0;	//idle
	
	ret = 0;
af_out:	
	
	//ae lock off
	sensor_ae_awb_lockoff(sd);
	
	return ret;
}

static int sensor_s_single_af(struct v4l2_subdev *sd)
{
  int ret;
	struct sensor_info *info = to_state(sd);
#if 0
	unsigned int rdval,cnt;	
#endif

	csi_dev_print("sensor_s_single_af\n");
	
	info->focus_status = 0;	//idle
	info->coarse_af_pd = 0;
	//ae lock on
	sensor_ae_awb_lockon(sd);
	
	//single trigger	
	ret = sensor_write_array(sd, sensor_af_single_trig_regs , ARRAY_SIZE(sensor_af_single_trig_regs));
	if(ret < 0) {
		csi_dev_err("sensor_s_single_af error!\n"); 
    //ae lock off
		sensor_ae_awb_lockoff(sd);
  } else if(ret == 0) {
  	info->focus_status = 1;	//busy
  }
	
	//msleep(100);
	
#if 0	
	//wait for 1st af complete

  rdval = 0xffff;
	cnt = 0;
	while(rdval!=0x0002)
	{
		ret = sensor_write_im(sd, 0x002c , 0x7000);
		ret = sensor_write_im(sd, 0x002e , 0x2eee);
		if(ret < 0)
		{
			csi_dev_err("sensor_s_single_af error!\n"); 
	    goto af_out;
	  }
		ret =  sensor_read_im(sd, 0x0f12 , &rdval);
  	if (ret < 0)
  	{
  		csi_dev_err("sensor_s_single_af read error !\n");
  		goto af_out;
  	}
  	csi_dev_dbg("Single AF 1st ,value = 0x%4x\n",rdval);
  	
  	if(rdval!=0x0001 && rdval!=0x0002) {
  		csi_dev_err("Single AF 1st is failed,value = 0x%4x\n",rdval);
			ret = -EFAULT;
  		goto af_out;
  	} else {
			mdelay(15);
			cnt++;
			if(cnt>60) {
				csi_dev_err("Single AF 1st is timeout,value = 0x%4x\n",rdval);
				ret = -EFAULT;
	  		goto af_out;
			}
		}
	}
	csi_dev_print("Single AF 1st is complete,value = 0x%4x\n",rdval);
	
	//wait for 2nd af complete
	rdval = 0xffff;
  cnt = 0;
	while((rdval&0xff00)!=0x0000)
	{
		ret = sensor_write_im(sd, 0x002c , 0x7000);
		ret = sensor_write_im(sd, 0x002e , 0x2207);
		if(ret < 0)
		{
			csi_dev_err("sensor_s_single_af error!\n"); 
	    goto af_out;
	  }
	  
		ret =  sensor_read_im(sd, 0x0f12 , &rdval);
  	if (ret < 0)
  	{
  		csi_dev_err("sensor_s_single_af read error !\n");
  		goto af_out;
  	}
  	mdelay(5);
  	
		cnt++;
		if(cnt>300) {
			csi_dev_err("Single AF 2nd is timeout,value = 0x%4x\n",rdval);
			ret = -EFAULT;
  		goto af_out;
		}
	}
	csi_dev_print("Single AF 2nd is complete,value = 0x%4x\n",rdval);
	
af_out:	
	//ae lock off
	sensor_ae_awb_lockoff(sd);
#endif	

	return ret;
}

static int sensor_s_continueous_af(struct v4l2_subdev *sd)
{
  int ret;

  csi_dev_print("sensor_s_continueous_af\n");
  ret = sensor_write_array(sd,sensor_af_continueous_regs,ARRAY_SIZE(sensor_af_continueous_regs));
  if(ret < 0)
    csi_dev_err("sensor_s_continueous_af error\n");
   
  return ret;
}

static int sensor_s_infinity_af(struct v4l2_subdev *sd)
{
  int ret;
  
  csi_dev_print("sensor_s_infinity_af\n");
  ret = sensor_write_array(sd,sensor_af_infinity_regs,ARRAY_SIZE(sensor_af_infinity_regs));
  if(ret < 0)
    csi_dev_err("sensor_s_infinity_af error\n");
   
  return ret;
}

static int sensor_s_macro_af(struct v4l2_subdev *sd)
{
  int ret;
  
  csi_dev_print("sensor_s_macro_af\n");
  ret = sensor_write_array(sd,sensor_af_macro_regs,ARRAY_SIZE(sensor_af_macro_regs));
  if(ret < 0)
    csi_dev_err("sensor_af_macro_regs error\n");
   
  return ret;
}

static int sensor_s_pause_af(struct v4l2_subdev *sd)
{
	return 0;
}

static int sensor_s_release_af(struct v4l2_subdev *sd)
{
	return 0;
}

static int sensor_s_af_zone(struct v4l2_subdev *sd, int xc, int yc)
{
	struct sensor_info *info = to_state(sd);
	int fst_win_start_x,fst_win_start_y;
	int fst_disp_start_x,fst_disp_start_y;
	int fst_win_size_x = 512,fst_win_size_y = 568;
	int fst_disp_size_x,fst_disp_size_y;
	int scnd_win_start_x,scnd_win_start_y;
	int scnd_disp_start_x,scnd_disp_start_y;
	int scnd_win_size_x = 116,scnd_win_size_y = 306;
	int scnd_disp_size_x,scnd_disp_size_y;
	
	if(info->width == 0 || info->height == 0) {
		csi_dev_err("current width or height is zero!\n");
		return -EINVAL;
	}
	
	if(info->focus_status == 1)	//can not set af zone when focus is busy
		return 0;
	
	csi_dev_print("sensor_s_af_zone\n");
	csi_dev_dbg("af zone input xc=%d,yc=%d\n",xc,yc);	
	//first window
	
	fst_disp_size_x = fst_win_size_x * info->width /1024;
	
	if(xc + fst_disp_size_x/2 > info->width) {
		fst_disp_start_x = info->width - fst_disp_size_x;
	} else if(xc - (fst_disp_size_x/2) < 0) {
		fst_disp_start_x = 0;
	} else {
		fst_disp_start_x = xc - (fst_disp_size_x/2);
	}
	
	fst_disp_size_y = fst_win_size_y * info->height /1024;
	if(yc + fst_disp_size_y/2 > info->height) {
		fst_disp_start_y = info->height - fst_disp_size_y;
	} else if(yc - fst_disp_size_y/2 < 0) {
		fst_disp_start_y = 0;
	} else {
		fst_disp_start_y = yc - fst_disp_size_y/2;
	}
	
	fst_win_start_x = fst_disp_start_x * 1024 / info->width;
	fst_win_start_y = fst_disp_start_y * 1024 / info->height;
	
	//second window
	
	scnd_disp_size_x = scnd_win_size_x * info->width /1024;
	if(xc + scnd_disp_size_x/2 > info->width) {
		scnd_disp_start_x = info->width - scnd_disp_size_x;
	} else if(xc - scnd_disp_size_x/2 < 0) {
		scnd_disp_start_x = 0;
	} else {
		scnd_disp_start_x = xc - scnd_disp_size_x/2;
	}
	
	scnd_disp_size_y = scnd_win_size_y * info->height /1024;
	if(yc + scnd_disp_size_y/2 > info->height) {
		scnd_disp_start_y = info->height - scnd_disp_size_y;
	} else if(yc - scnd_disp_size_y/2 < 0) {
		scnd_disp_start_y = 0;
	} else {
		scnd_disp_start_y = yc - scnd_disp_size_y/2;
	}
	
	scnd_win_start_x = scnd_disp_start_x * 1024 / info->width;
	scnd_win_start_y = scnd_disp_start_y * 1024 / info->height;
	
	sensor_write_im(sd,0x0028,0x7000);
	sensor_write_im(sd,0x002a,0x0294);
	sensor_write_im(sd,0x0f12,fst_win_start_x);	//REG_TC_AF_FstWinStartX 
	sensor_write_im(sd,0x0f12,fst_win_start_y);	//REG_TC_AF_FstWinStartY 
	sensor_write_im(sd,0x0f12,fst_win_size_x); 	//REG_TC_AF_FstWinSizeX  
	sensor_write_im(sd,0x0f12,fst_win_size_y); 	//REG_TC_AF_FstWinSizeY  
	sensor_write_im(sd,0x0f12,scnd_win_start_x); //REG_TC_AF_ScndWinStartX
	sensor_write_im(sd,0x0f12,scnd_win_start_y);	//REG_TC_AF_ScndWinStartY
	sensor_write_im(sd,0x0f12,scnd_win_size_x);	//REG_TC_AF_ScndWinSizeX 
	sensor_write_im(sd,0x0f12,scnd_win_size_y);	//REG_TC_AF_ScndWinSizeY 
	sensor_write_im(sd,0x0f12,0x0001);						//REG_TC_AF_WinSizesUpdated;
	
	csi_dev_dbg("af zone 1st window stx=%d,sty=%d,width=%d,height=%d\n",fst_win_start_x,fst_win_start_y,fst_win_size_x,fst_win_size_y);
	csi_dev_dbg("af zone 2nd window stx=%d,sty=%d,width=%d,height=%d\n",scnd_win_start_x,scnd_win_start_y,scnd_win_size_x,scnd_win_size_y);
	
	mdelay(30);
	return 0;
}

#if 0
static int sensor_s_relaunch_af_zone(struct v4l2_subdev *sd)
{
	return 0;
}
#endif

/*
 * CSI GPIO control
 */
static void csi_gpio_write(struct v4l2_subdev *sd, struct gpio_config *gpio, int level)
{
//	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
  if(gpio->gpio==GPIO_INDEX_INVALID)
  {
    csi_dev_dbg("invalid gpio\n");
    return;
  }
  
	if(gpio->mul_sel==1)
	{
	  gpio_direction_output(gpio->gpio, level);
	  gpio->data=level;
	} else {
	  csi_dev_dbg("gpio is not in output function\n");
	}
}

static void csi_gpio_set_status(struct v4l2_subdev *sd, struct gpio_config *gpio, int status)
{
//	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
	
	if(1 == status) {  /* output */
		if(0 != gpio_direction_output(gpio->gpio, gpio->data))
			csi_dev_dbg("gpio_direction_output failed\n");
	} else if(0 == status) {  /* input */
	  if(0 != gpio_direction_input(gpio->gpio) )
	    csi_dev_dbg("gpio_direction_input failed\n");
	}
	gpio->mul_sel=status;
}



/*
 * Stuff that knows about the sensor.
 */
 
static int sensor_power(struct v4l2_subdev *sd, int on)
{
	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
	struct i2c_client *client = v4l2_get_subdevdata(sd);
//	int ret;
  
  //insure that clk_disable() and clk_enable() are called in pair 
  //when calling CSI_SUBDEV_STBY_ON/OFF and CSI_SUBDEV_PWR_ON/OFF
  switch(on)
	{
		case CSI_SUBDEV_STBY_ON:
			csi_dev_dbg("CSI_SUBDEV_STBY_ON\n");
			//disable io oe
//			csi_dev_print("disalbe oe!\n");
//			ret = sensor_write_array(sd, sensor_oe_disable_regs , ARRAY_SIZE(sensor_oe_disable_regs));
//			if(ret < 0)
//				csi_dev_err("disalbe oe falied!\n");
			//make sure that no device can access i2c bus during sensor initial or power down
			//when using i2c_lock_adpater function, the following codes must not access i2c bus before calling i2c_unlock_adapter
			i2c_lock_adapter(client->adapter);
			//reset on io
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			//standby on io
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_ON);
			mdelay(1);
			//inactive mclk after stadby in
			clk_disable(dev->csi_module_clk);
			//remember to unlock i2c adapter, so the device can access the i2c bus again
			i2c_unlock_adapter(client->adapter);	
			break;
		case CSI_SUBDEV_STBY_OFF:
			csi_dev_dbg("CSI_SUBDEV_STBY_OFF\n");
			//make sure that no device can access i2c bus during sensor initial or power down
			//when using i2c_lock_adpater function, the following codes must not access i2c bus before calling i2c_unlock_adapter
			i2c_lock_adapter(client->adapter);
			//active mclk before stadby out
			clk_enable(dev->csi_module_clk);
			mdelay(1);
			//standby off io
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_OFF);
			mdelay(1);
			//reset on io
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			//remember to unlock i2c adapter, so the device can access the i2c bus again
			i2c_unlock_adapter(client->adapter);	
			break;
		case CSI_SUBDEV_PWR_ON:
			csi_dev_dbg("CSI_SUBDEV_PWR_ON\n");
			//make sure that no device can access i2c bus during sensor initial or power down
			//when using i2c_lock_adpater function, the following codes must not access i2c bus before calling i2c_unlock_adapter
			i2c_lock_adapter(client->adapter);
			csi_gpio_set_status(sd,&dev->standby_io,1);//set the gpio to output
			csi_gpio_set_status(sd,&dev->reset_io,1);//set the gpio to output
			//power supply
			csi_gpio_write(sd,&dev->power_io,CSI_PWR_ON);
			if(dev->iovdd) {
				regulator_enable(dev->iovdd);
			}
			if(dev->dvdd) {
				regulator_enable(dev->dvdd);
			}
			if(dev->avdd) {
				regulator_enable(dev->avdd);
			}
			csi_gpio_write(sd,&dev->af_power_io,CSI_AF_PWR_ON);
			mdelay(10);
			//active mclk power on reset
			clk_enable(dev->csi_module_clk);
			//power on reset
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_ON);
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_OFF);
			udelay(20);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			udelay(100);
			//remember to unlock i2c adapter, so the device can access the i2c bus again
			i2c_unlock_adapter(client->adapter);	
			break;
		case CSI_SUBDEV_PWR_OFF:
			csi_dev_dbg("CSI_SUBDEV_PWR_OFF\n");
			//make sure that no device can access i2c bus during sensor initial or power down
			//when using i2c_lock_adpater function, the following codes must not access i2c bus before calling i2c_unlock_adapter
			i2c_lock_adapter(client->adapter);
			//reset on
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			udelay(100);
			//inactive mclk before power off
			clk_disable(dev->csi_module_clk);
			udelay(10);
			//standy on
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_ON);		
			//power supply off
			csi_gpio_write(sd,&dev->af_power_io,CSI_AF_PWR_OFF);
			if(dev->dvdd) {
				regulator_disable(dev->dvdd);
			}
			if(dev->avdd) {
				regulator_disable(dev->avdd);
			}
			if(dev->iovdd) {
				regulator_disable(dev->iovdd);
			}
			csi_gpio_write(sd,&dev->power_io,CSI_PWR_OFF);
			//set the io to hi-z
			csi_gpio_set_status(sd,&dev->reset_io,0);//set the gpio to input
			csi_gpio_set_status(sd,&dev->standby_io,0);//set the gpio to input
			//remember to unlock i2c adapter, so the device can access the i2c bus again
			i2c_unlock_adapter(client->adapter);	
			break;
		default:
			return -EINVAL;
	}		

	return 0;
}
 
static int sensor_reset(struct v4l2_subdev *sd, u32 val)
{
	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);

	switch(val)
	{
		case CSI_SUBDEV_RST_OFF:
			csi_dev_dbg("CSI_SUBDEV_RST_OFF\n");
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			break;
		case CSI_SUBDEV_RST_ON:
			csi_dev_dbg("CSI_SUBDEV_RST_ON\n");
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			break;
		case CSI_SUBDEV_RST_PUL:
			csi_dev_dbg("CSI_SUBDEV_RST_PUL\n");
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			break;
		default:
			return -EINVAL;
	}
		
	return 0;
}

static int sensor_detect(struct v4l2_subdev *sd)
{
	int ret;
	unsigned int rdval=0xffff;
	
	ret = sensor_write_im(sd, 0x002c, 0x7000);
	ret = sensor_write_im(sd, 0x002e, 0x01a4);
	if (ret < 0)
	{
		csi_dev_err("sensor_write err at sensor_detect!\n");
		return ret;
	}

	ret = sensor_read_im(sd, 0x0f12, &rdval);

	if(rdval != 0x4ec0 )
	{
		csi_dev_err("id read from sensor is 0x%4x,not 0x4ec0\n",rdval);
		return -ENODEV;
	}
	
	return 0;
}

static int sensor_init(struct v4l2_subdev *sd, u32 val)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	csi_dev_dbg("sensor_init\n");
	
	/*Make sure it is a target sensor*/
	ret = sensor_detect(sd);
	if (ret) {
		csi_dev_err("chip found is not an target chip.\n");
		return ret;
	}
	
	ret = sensor_write_array(sd, sensor_default_regs , ARRAY_SIZE(sensor_default_regs));	
	if(ret < 0) {
		csi_dev_err("write sensor_default_regs error\n");
		return ret;
	}
			
	sensor_s_band_filter(sd, V4L2_CID_POWER_LINE_FREQUENCY_50HZ);
	
	info->focus_status = 0;
	info->width = 0;
	info->height = 0;
	info->brightness = 0;
	info->contrast = 0;
	info->saturation = 0;
	info->hue = 0;
	info->hflip = 0;
	info->vflip = 0;
	info->gain = 0;
	info->autogain = 1;
	info->exp = 0;
	info->autoexp = 0;
	info->autowb = 1;
	info->wb = V4L2_WB_AUTO;
	info->clrfx = V4L2_COLORFX_NONE;
	info->band_filter = V4L2_CID_POWER_LINE_FREQUENCY_50HZ;
	info->af_mode = V4L2_AF_FIXED;
	info->af_ctrl = V4L2_AF_RELEASE;
	info->tpf.numerator = 1;            
	info->tpf.denominator = 30;    /* 30fps */
	info->width = 0;
	info->height = 0;
	
	return 0;
}

static long sensor_ioctl(struct v4l2_subdev *sd, unsigned int cmd, void *arg)
{
	int ret=0;
	
	switch(cmd){
		case CSI_SUBDEV_CMD_GET_INFO: 
		{
			struct sensor_info *info = to_state(sd);
			__csi_subdev_info_t *ccm_info = arg;
			
			csi_dev_dbg("CSI_SUBDEV_CMD_GET_INFO\n");
			
			ccm_info->mclk 	=	info->ccm_info->mclk ;
			ccm_info->vref 	=	info->ccm_info->vref ;
			ccm_info->href 	=	info->ccm_info->href ;
			ccm_info->clock	=	info->ccm_info->clock;
			ccm_info->iocfg	=	info->ccm_info->iocfg;
	
			csi_dev_dbg("ccm_info.mclk=%d\n ",info->ccm_info->mclk);
			csi_dev_dbg("ccm_info.vref=%x\n ",info->ccm_info->vref);
			csi_dev_dbg("ccm_info.href=%x\n ",info->ccm_info->href);
			csi_dev_dbg("ccm_info.clock=%x\n ",info->ccm_info->clock);
			csi_dev_dbg("ccm_info.iocfg=%x\n ",info->ccm_info->iocfg);
			break;
		}
		case CSI_SUBDEV_CMD_SET_INFO:
		{
			struct sensor_info *info = to_state(sd);
			__csi_subdev_info_t *ccm_info = arg;
			
			csi_dev_dbg("CSI_SUBDEV_CMD_SET_INFO\n");
			
			info->ccm_info->mclk 	=	ccm_info->mclk 	;
			info->ccm_info->vref 	=	ccm_info->vref 	;
			info->ccm_info->href 	=	ccm_info->href 	;
			info->ccm_info->clock	=	ccm_info->clock	;
			info->ccm_info->iocfg	=	ccm_info->iocfg	;
			
			csi_dev_dbg("ccm_info.mclk=%d\n ",info->ccm_info->mclk);
			csi_dev_dbg("ccm_info.vref=%x\n ",info->ccm_info->vref);
			csi_dev_dbg("ccm_info.href=%x\n ",info->ccm_info->href);
			csi_dev_dbg("ccm_info.clock=%x\n ",info->ccm_info->clock);
			csi_dev_dbg("ccm_info.iocfg=%x\n ",info->ccm_info->iocfg);
			
			break;
		}
		case CSI_SUBDEV_CMD_DETECT:
		{
			ret=sensor_detect(sd);
			break;
		}
		default:
			return -EINVAL;
	}		
		return ret;
}


/*
 * Store information about the video data format. 
 */
static struct sensor_format_struct {
	__u8 *desc;
	//__u32 pixelformat;
	enum v4l2_mbus_pixelcode mbus_code;//linux-3.0
	struct regval_list *regs;
	int	regs_size;
	int bpp;   /* Bytes per pixel */
} sensor_formats[] = {
	{
		.desc		= "YUYV 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_YUYV8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_yuyv,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_yuyv),
		.bpp		= 2,
	},
	{
		.desc		= "YVYU 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_YVYU8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_yvyu,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_yvyu),
		.bpp		= 2,
	},
	{
		.desc		= "UYVY 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_UYVY8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_uyvy,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_uyvy),
		.bpp		= 2,
	},
	{
		.desc		= "VYUY 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_VYUY8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_vyuy,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_vyuy),
		.bpp		= 2,
	},
//	{
//		.desc		= "Raw RGB Bayer",
//		.mbus_code	= V4L2_MBUS_FMT_SBGGR8_1X8,//linux-3.0
//		.regs 		= sensor_fmt_raw,
//		.regs_size = ARRAY_SIZE(sensor_fmt_raw),
//		.bpp		= 1
//	},
};
#define N_FMTS ARRAY_SIZE(sensor_formats)

	

/*
 * Then there is the issue of window sizes.  Try to capture the info here.
 */


static struct sensor_win_size {
	int	width;
	int	height;
	int	hstart;		/* Start/stop values for the camera.  Note */
	int	hstop;		/* that they do not always make complete */
	int	vstart;		/* sense to humans, but evidently the sensor */
	int	vstop;		/* will do the right thing... */
	struct regval_list *regs; /* Regs to tweak */
	int regs_size;
	int (*set_size) (struct v4l2_subdev *sd);
/* h/vref stuff */
} sensor_win_sizes[] = {
	/* qsxga: 2592*1936 */
	{
		.width			= QSXGA_WIDTH,
		.height 		= QSXGA_HEIGHT,
		.regs			  = sensor_qsxga_regs,
		.regs_size	= ARRAY_SIZE(sensor_qsxga_regs),
		.set_size		= NULL,
	},
	/* qxga: 2048*1536 */
	{
		.width			= QXGA_WIDTH,
		.height 		= QXGA_HEIGHT,
		.regs			  = sensor_qxga_regs,
		.regs_size	= ARRAY_SIZE(sensor_qxga_regs),
		.set_size		= NULL,
	},
	/* 1080P */
	{
		.width			= HD1080_WIDTH,
		.height			= HD1080_HEIGHT,
		.regs 			= sensor_1080p_regs,
		.regs_size	= ARRAY_SIZE(sensor_1080p_regs),
		.set_size		= NULL,
	},

	/* 720p */
	{
		.width			= HD720_WIDTH,
		.height			= HD720_HEIGHT,
		.regs 			= sensor_720p_regs,
		.regs_size	= ARRAY_SIZE(sensor_720p_regs),
		.set_size		= NULL,
	},
	/* VGA */
	{
		.width			= VGA_WIDTH,
		.height			= VGA_HEIGHT,
		.regs				= sensor_vga_regs,
		.regs_size	= ARRAY_SIZE(sensor_vga_regs),
		.set_size		= NULL,
	},
};

#define N_WIN_SIZES (ARRAY_SIZE(sensor_win_sizes))




static int sensor_enum_fmt(struct v4l2_subdev *sd, unsigned index,
                 enum v4l2_mbus_pixelcode *code)//linux-3.0
{
//	struct sensor_format_struct *ofmt;

	if (index >= N_FMTS)//linux-3.0
		return -EINVAL;

	*code = sensor_formats[index].mbus_code;//linux-3.0
//	ofmt = sensor_formats + fmt->index;
//	fmt->flags = 0;
//	strcpy(fmt->description, ofmt->desc);
//	fmt->pixelformat = ofmt->pixelformat;
	return 0;
}


static int sensor_try_fmt_internal(struct v4l2_subdev *sd,
		//struct v4l2_format *fmt,
		struct v4l2_mbus_framefmt *fmt,//linux-3.0
		struct sensor_format_struct **ret_fmt,
		struct sensor_win_size **ret_wsize)
{
	int index;
	struct sensor_win_size *wsize;
//	struct v4l2_pix_format *pix = &fmt->fmt.pix;//linux-3.0

	for (index = 0; index < N_FMTS; index++)
		if (sensor_formats[index].mbus_code == fmt->code)//linux-3.0
			break;
	
	if (index >= N_FMTS) {
		/* default to first format */
		index = 0;
		fmt->code = sensor_formats[0].mbus_code;//linux-3.0
	}
	
	if (ret_fmt != NULL)
		*ret_fmt = sensor_formats + index;
		
	/*
	 * Fields: the sensor devices claim to be progressive.
	 */
	fmt->field = V4L2_FIELD_NONE;//linux-3.0
	
	
	/*
	 * Round requested image size down to the nearest
	 * we support, but not below the smallest.
	 */
	for (wsize = sensor_win_sizes; wsize < sensor_win_sizes + N_WIN_SIZES;
	     wsize++)
		if (fmt->width >= wsize->width && fmt->height >= wsize->height)//linux-3.0
			break;
		
	if (wsize >= sensor_win_sizes + N_WIN_SIZES)
		wsize--;   /* Take the smallest one */
	if (ret_wsize != NULL)
		*ret_wsize = wsize;
	/*
	 * Note the size we'll actually handle.
	 */
	fmt->width = wsize->width;//linux-3.0
	fmt->height = wsize->height;//linux-3.0
	//pix->bytesperline = pix->width*sensor_formats[index].bpp;//linux-3.0
	//pix->sizeimage = pix->height*pix->bytesperline;//linux-3.0
	
	return 0;
}

static int sensor_try_fmt(struct v4l2_subdev *sd, 
             struct v4l2_mbus_framefmt *fmt)//linux-3.0
{
	return sensor_try_fmt_internal(sd, fmt, NULL, NULL);
}


/*
 * Set a format.
 */
static int sensor_s_fmt(struct v4l2_subdev *sd, 
             struct v4l2_mbus_framefmt *fmt)//linux-3.0
{
	int ret;
	struct sensor_format_struct *sensor_fmt;
	struct sensor_win_size *wsize;
	struct sensor_info *info = to_state(sd);
	
	csi_dev_dbg("sensor_s_fmt\n");
	
	ret = sensor_try_fmt_internal(sd, fmt, &sensor_fmt, &wsize);
	if (ret)
		return ret;
	
	if(info->fmt == sensor_fmt && info->width == wsize->width && info->height == wsize->height)
	{
		csi_dev_print("format and size remain the same\n");
		goto update;
	}
		
	if(info->capture_mode == V4L2_MODE_VIDEO || info->capture_mode == V4L2_MODE_PREVIEW)
	{
		//video
		if(info->af_mode != V4L2_AF_FIXED) {
			ret = sensor_s_release_af(sd);
			if (ret < 0)
			{
				csi_dev_err("sensor_s_release_af err !\n");
				return ret;
			}	
		}
	}
	else if(info->capture_mode == V4L2_MODE_IMAGE)
	{
		//capture
		ret = sensor_set_exposure(sd);
		if (ret < 0)
		{
			csi_dev_err("sensor_set_exposure err !\n");
			return ret;
		}	
	}
	
	ret = sensor_write_array(sd, sensor_fmt->regs , sensor_fmt->regs_size);
	if (ret < 0)		
		return ret;
	
	ret = 0;
	if (wsize->regs)
	{
		ret = sensor_write_array(sd, wsize->regs , wsize->regs_size);
		if (ret < 0)
			return ret;
	}
	
	if (wsize->set_size)
	{
		ret = wsize->set_size(sd);
		if (ret < 0)
			return ret;
	}
  
	if(info->capture_mode == V4L2_MODE_VIDEO || info->capture_mode == V4L2_MODE_PREVIEW)
	{
		//video
		if(info->af_mode != V4L2_AF_FIXED) {
			
#if 0			
			if(info->af_mode != V4L2_AF_TOUCH && info->af_mode != V4L2_AF_FACE) {				
				ret = sensor_s_relaunch_af_zone(sd);	//set af zone to default zone
				if (ret < 0) {
					csi_dev_err("sensor_s_relaunch_af_zone err !\n");
					return ret;
				}	
			}
#endif

#ifdef CONTINUEOUS_AF				
			if(info->af_mode != V4L2_AF_INFINITY) {
				ret = sensor_s_continueous_af(sd);		//set continueous af
				if (ret < 0) {
					csi_dev_err("sensor_s_continueous_af err !\n");
					return ret;
				}
			}
		}
#endif
	}

update:
	info->fmt = sensor_fmt;
	info->width = wsize->width;
	info->height = wsize->height;
	
	csi_dev_print("s_fmt set width = %d, height = %d\n",wsize->width,wsize->height);
		
	return 0;
}

/*
 * Implement G/S_PARM.  There is a "high quality" mode we could try
 * to do someday; for now, we just do the frame rate tweak.
 */
static int sensor_g_parm(struct v4l2_subdev *sd, struct v4l2_streamparm *parms)
{
	struct v4l2_captureparm *cp = &parms->parm.capture;
	struct sensor_info *info = to_state(sd);

	if (parms->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	
	memset(cp, 0, sizeof(struct v4l2_captureparm));
	cp->capability = V4L2_CAP_TIMEPERFRAME;
	cp->capturemode = info->capture_mode;
	
	cp->timeperframe.numerator = info->tpf.numerator;
	cp->timeperframe.denominator = info->tpf.denominator;
	 
	return 0;
}

static int sensor_s_parm(struct v4l2_subdev *sd, struct v4l2_streamparm *parms)
{
	return 0;
}


/* 
 * Code for dealing with controls.
 * fill with different sensor module
 * different sensor module has different settings here
 * if not support the follow function ,retrun -EINVAL
 */

/* *********************************************begin of ******************************************** */
static int sensor_queryctrl(struct v4l2_subdev *sd,
		struct v4l2_queryctrl *qc)
{
	/* Fill in min, max, step and default value for these controls. */
	/* see include/linux/videodev2.h for details */
	/* see sensor_s_parm and sensor_g_parm for the meaning of value */
	
	switch (qc->id) {
//	case V4L2_CID_BRIGHTNESS:
//		return v4l2_ctrl_query_fill(qc, -4, 4, 1, 1);
//	case V4L2_CID_CONTRAST:
//		return v4l2_ctrl_query_fill(qc, -4, 4, 1, 1);
//	case V4L2_CID_SATURATION:
//		return v4l2_ctrl_query_fill(qc, -4, 4, 1, 1);
//	case V4L2_CID_HUE:
//		return v4l2_ctrl_query_fill(qc, -180, 180, 5, 0);
	case V4L2_CID_VFLIP:
	case V4L2_CID_HFLIP:
		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 0);
//	case V4L2_CID_GAIN:
//		return v4l2_ctrl_query_fill(qc, 0, 255, 1, 128);
//	case V4L2_CID_AUTOGAIN:
//		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 1);
	case V4L2_CID_EXPOSURE:
		return v4l2_ctrl_query_fill(qc, -4, 4, 1, 0);
	case V4L2_CID_EXPOSURE_AUTO:
		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 0);
	case V4L2_CID_DO_WHITE_BALANCE:
		return v4l2_ctrl_query_fill(qc, 0, 5, 1, 0);
	case V4L2_CID_AUTO_WHITE_BALANCE:
		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 1);
	case V4L2_CID_COLORFX:
		return v4l2_ctrl_query_fill(qc, 0, 9, 1, 0);
	case V4L2_CID_CAMERA_FLASH_MODE:
	  return v4l2_ctrl_query_fill(qc, 0, 4, 1, 0);	
	case V4L2_CID_CAMERA_AF_MODE:
	  return v4l2_ctrl_query_fill(qc, 0, 5, 1, 0);
	case V4L2_CID_CAMERA_AF_CTRL:
	  return v4l2_ctrl_query_fill(qc, 0, 5, 1, 0);
	}
	return -EINVAL;
}

static int sensor_g_hflip(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);

	*value = info->hflip;
	return 0;
}

static int sensor_s_hflip(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	unsigned int pre_val,cap_val;
	struct regval_list regs[] = {
		////==================================================================================
		//// 20.Preview & Capture Configration Setting
		////==================================================================================
		//{{0xFC,0xFC},{0xD0,0x00}},
		//{{0x00,0x28},{0x70,0x00}},
		//{{0x00,0x2A},{0x02,0xD0}},	
		//{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uPrevMirror     Original 00 / H Mirror 05 / V Mirror 0A / HV Mirror 0F
		//{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uCaptureMirror  Original 00 / H Mirror 05 / V Mirror 0A / HV Mirror 0F
		
		//==================================================================================
		// 21.Select Cofigration Display
		//==================================================================================
		//PREVIEW
		{{0xFC,0xFC},{0xD0,0x00}},
		{{0x00,0x28},{0x70,0x00}},
		{{0x00,0x2A},{0x02,0x66}},
		{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
		{{0x00,0x2A},{0x02,0x6A}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
		{{0x00,0x2A},{0x02,0x4E}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
		{{0x00,0x2A},{0x02,0x68}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
		{{0x00,0x2A},{0x02,0x3E}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged;
  };

	if(info->hflip == value)
		return 0;

	sensor_write_im(sd, 0x002c , 0x7000);
	sensor_write_im(sd, 0x002e , 0x02d0);
	sensor_read_im(sd, 0x0f12, &pre_val);
	sensor_read_im(sd, 0x0f12, &cap_val);
	sensor_write_im(sd, 0x0028, 0x7000);
	sensor_write_im(sd, 0x002a, 0x02d0);
	
	switch (value) {
		case 0:
		  pre_val &= 0xfa;
		  cap_val &= 0xfa;
			break;
		case 1:
		  pre_val |= 0x05;
		  cap_val |= 0x05;
			break;
		default:
			return -EINVAL;
	}
	
	sensor_write_im(sd, 0x0f12, pre_val);
  sensor_write_im(sd, 0x0f12, cap_val);
  
	ret = sensor_write_array(sd, regs, ARRAY_SIZE(regs));
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_hflip!\n");
		return ret;
	}

	info->hflip = value;
	
	return 0;
}

static int sensor_g_vflip(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);

	*value = info->vflip;
	return 0;
}

static int sensor_s_vflip(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	unsigned int pre_val,cap_val;
	struct regval_list regs[] = {
		////==================================================================================
		//// 20.Preview & Capture Configration Setting
		////==================================================================================
		//{{0xFC,0xFC},{0xD0,0x00}},
		//{{0x00,0x28},{0x70,0x00}},
		//{{0x00,0x2A},{0x02,0xD0}},	
		//{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uPrevMirror     Original 00 / H Mirror 05 / V Mirror 0A / HV Mirror 0F
		//{{0x0F,0x12},{0x00,0x00}},	//REG_0TC_PCFG_uCaptureMirror  Original 00 / H Mirror 05 / V Mirror 0A / HV Mirror 0F
		
		//==================================================================================
		// 21.Select Cofigration Display
		//==================================================================================
		//PREVIEW
		{{0xFC,0xFC},{0xD0,0x00}},
		{{0x00,0x28},{0x70,0x00}},
		{{0x00,0x2A},{0x02,0x66}},
		{{0x0f,0x12},{0x00,0x00}},	//REG_TC_GP_ActivePrevConfig
		{{0x00,0x2A},{0x02,0x6A}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevOpenAfterChange
		{{0x00,0x2A},{0x02,0x4E}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_NewConfigSync
		{{0x00,0x2A},{0x02,0x68}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_PrevConfigChanged
		{{0x00,0x2A},{0x02,0x3E}},
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreview
		{{0x0F,0x12},{0x00,0x01}},	//REG_TC_GP_EnablePreviewChanged;
  };

	if(info->vflip == value)
		return 0;
	
	sensor_write_im(sd, 0x0028, 0x7000);
	sensor_write_im(sd, 0x002a, 0x02d0);
	sensor_write_im(sd, 0x002c , 0x7000);
	sensor_write_im(sd, 0x002e , 0x02d0);
	sensor_read_im(sd, 0x0f12, &pre_val);
	sensor_read_im(sd, 0x0f12, &cap_val);
	
	switch (value) {
		case 0:
		  pre_val &= 0xf5;
		  cap_val &= 0xf5;
			break;
		case 1:
		  pre_val |= 0x0A;
		  cap_val |= 0x0A;
			break;
		default:
			return -EINVAL;
	}

	sensor_write_im(sd, 0x0f12, pre_val);
  sensor_write_im(sd, 0x0f12, cap_val);

	ret = sensor_write_array(sd, regs, ARRAY_SIZE(regs));
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_vflip!\n");
		return ret;
	}

	info->vflip = value;
	
	return 0;
}

static int sensor_g_autogain(struct v4l2_subdev *sd, __s32 *value)
{
	return -EINVAL;
}

static int sensor_s_autogain(struct v4l2_subdev *sd, int value)
{
	return -EINVAL;
}

static int sensor_g_autoexp(struct v4l2_subdev *sd, __s32 *value)
{
	return -EINVAL;
}

static int sensor_s_autoexp(struct v4l2_subdev *sd,
		enum v4l2_exposure_auto_type value)
{
	return -EINVAL;
}

static int sensor_g_autowb(struct v4l2_subdev *sd, int *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->autowb;
	return 0;
}

static int sensor_s_autowb(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	ret = sensor_write_array(sd, sensor_wb_auto_regs, ARRAY_SIZE(sensor_wb_auto_regs));
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_autowb!\n");
		return ret;
	}
	
	mdelay(10);
	info->autowb = value;
	
	return 0;
}

static int sensor_g_hue(struct v4l2_subdev *sd, __s32 *value)
{
	return -EINVAL;
}

static int sensor_s_hue(struct v4l2_subdev *sd, int value)
{
	return -EINVAL;
}

static int sensor_g_gain(struct v4l2_subdev *sd, __s32 *value)
{
	return -EINVAL;
}

static int sensor_s_gain(struct v4l2_subdev *sd, int value)
{
	return -EINVAL;
}

static int sensor_g_band_filter(struct v4l2_subdev *sd, 
		__s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->band_filter;
	return 0;
}

static int sensor_s_band_filter(struct v4l2_subdev *sd, 
		enum v4l2_power_line_frequency value)
{
	struct sensor_info *info = to_state(sd);

	int ret = 0;
	
	switch(value) {
		case V4L2_CID_POWER_LINE_FREQUENCY_DISABLED:
			break;			
		case V4L2_CID_POWER_LINE_FREQUENCY_50HZ:
			ret = sensor_write_array(sd, sensor_flicker_50hz_regs, ARRAY_SIZE(sensor_flicker_50hz_regs));
			if (ret < 0)
				csi_dev_err("sensor_write_array err at sensor_s_band_filter!\n");
			break;
		case V4L2_CID_POWER_LINE_FREQUENCY_60HZ:
			ret = sensor_write_array(sd, sensor_flicker_60hz_regs, ARRAY_SIZE(sensor_flicker_60hz_regs));
			if (ret < 0)
				csi_dev_err("sensor_write_array err at sensor_s_band_filter!\n");
		  break;
		case V4L2_CID_POWER_LINE_FREQUENCY_AUTO:
		default:
		  break;
	}
	mdelay(10);
	info->band_filter = value;
	return ret;
}
/* *********************************************end of ******************************************** */

static int sensor_g_brightness(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->brightness;
	return 0;
}

static int sensor_s_brightness(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	switch (value) {
		case -4:
		  ret = sensor_write_array(sd, sensor_brightness_neg4_regs, ARRAY_SIZE(sensor_brightness_neg4_regs));
			break;
		case -3:
			ret = sensor_write_array(sd, sensor_brightness_neg3_regs, ARRAY_SIZE(sensor_brightness_neg3_regs));
			break;
		case -2:
			ret = sensor_write_array(sd, sensor_brightness_neg2_regs, ARRAY_SIZE(sensor_brightness_neg2_regs));
			break;   
		case -1:
			ret = sensor_write_array(sd, sensor_brightness_neg1_regs, ARRAY_SIZE(sensor_brightness_neg1_regs));
			break;
		case 0:   
			ret = sensor_write_array(sd, sensor_brightness_zero_regs, ARRAY_SIZE(sensor_brightness_zero_regs));
			break;
		case 1:
			ret = sensor_write_array(sd, sensor_brightness_pos1_regs, ARRAY_SIZE(sensor_brightness_pos1_regs));
			break;
		case 2:
			ret = sensor_write_array(sd, sensor_brightness_pos2_regs, ARRAY_SIZE(sensor_brightness_pos2_regs));
			break;	
		case 3:
			ret = sensor_write_array(sd, sensor_brightness_pos3_regs, ARRAY_SIZE(sensor_brightness_pos3_regs));
			break;
		case 4:
			ret = sensor_write_array(sd, sensor_brightness_pos4_regs, ARRAY_SIZE(sensor_brightness_pos4_regs));
			break;
		default:
			return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_brightness!\n");
		return ret;
	}
	mdelay(10);
	info->brightness = value;
	return 0;
}

static int sensor_g_contrast(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->contrast;
	return 0;
}

static int sensor_s_contrast(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	switch (value) {
		case -4:
		  ret = sensor_write_array(sd, sensor_contrast_neg4_regs, ARRAY_SIZE(sensor_contrast_neg4_regs));
			break;
		case -3:
			ret = sensor_write_array(sd, sensor_contrast_neg3_regs, ARRAY_SIZE(sensor_contrast_neg3_regs));
			break;
		case -2:
			ret = sensor_write_array(sd, sensor_contrast_neg2_regs, ARRAY_SIZE(sensor_contrast_neg2_regs));
			break;   
		case -1:
			ret = sensor_write_array(sd, sensor_contrast_neg1_regs, ARRAY_SIZE(sensor_contrast_neg1_regs));
			break;
		case 0:   
			ret = sensor_write_array(sd, sensor_contrast_zero_regs, ARRAY_SIZE(sensor_contrast_zero_regs));
			break;
		case 1:
			ret = sensor_write_array(sd, sensor_contrast_pos1_regs, ARRAY_SIZE(sensor_contrast_pos1_regs));
			break;
		case 2:
			ret = sensor_write_array(sd, sensor_contrast_pos2_regs, ARRAY_SIZE(sensor_contrast_pos2_regs));
			break;	
		case 3:
			ret = sensor_write_array(sd, sensor_contrast_pos3_regs, ARRAY_SIZE(sensor_contrast_pos3_regs));
			break;
		case 4:
			ret = sensor_write_array(sd, sensor_contrast_pos4_regs, ARRAY_SIZE(sensor_contrast_pos4_regs));
			break;
		default:
			return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_contrast!\n");
		return ret;
	}
	mdelay(10);
	info->contrast = value;
	return 0;
}

static int sensor_g_saturation(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->saturation;
	return 0;
}

static int sensor_s_saturation(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	switch (value) {
		case -4:
		  ret = sensor_write_array(sd, sensor_saturation_neg4_regs, ARRAY_SIZE(sensor_saturation_neg4_regs));
			break;
		case -3:
			ret = sensor_write_array(sd, sensor_saturation_neg3_regs, ARRAY_SIZE(sensor_saturation_neg3_regs));
			break;
		case -2:
			ret = sensor_write_array(sd, sensor_saturation_neg2_regs, ARRAY_SIZE(sensor_saturation_neg2_regs));
			break;   
		case -1:
			ret = sensor_write_array(sd, sensor_saturation_neg1_regs, ARRAY_SIZE(sensor_saturation_neg1_regs));
			break;
		case 0:   
			ret = sensor_write_array(sd, sensor_saturation_zero_regs, ARRAY_SIZE(sensor_saturation_zero_regs));
			break;
		case 1:
			ret = sensor_write_array(sd, sensor_saturation_pos1_regs, ARRAY_SIZE(sensor_saturation_pos1_regs));
			break;
		case 2:
			ret = sensor_write_array(sd, sensor_saturation_pos2_regs, ARRAY_SIZE(sensor_saturation_pos2_regs));
			break;	
		case 3:
			ret = sensor_write_array(sd, sensor_saturation_pos3_regs, ARRAY_SIZE(sensor_saturation_pos3_regs));
			break;
		case 4:
			ret = sensor_write_array(sd, sensor_saturation_pos4_regs, ARRAY_SIZE(sensor_saturation_pos4_regs));
			break;
		default:
			return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_saturation!\n");
		return ret;
	}
	mdelay(10);
	info->saturation = value;
	return 0;
}

static int sensor_g_exp(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->exp;
	return 0;
}

static int sensor_s_exp(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	switch (value) {
		case -4:
		  ret = sensor_write_array(sd, sensor_ev_neg4_regs, ARRAY_SIZE(sensor_ev_neg4_regs));
			break;
		case -3:
			ret = sensor_write_array(sd, sensor_ev_neg3_regs, ARRAY_SIZE(sensor_ev_neg3_regs));
			break;
		case -2:
			ret = sensor_write_array(sd, sensor_ev_neg2_regs, ARRAY_SIZE(sensor_ev_neg2_regs));
			break;   
		case -1:
			ret = sensor_write_array(sd, sensor_ev_neg1_regs, ARRAY_SIZE(sensor_ev_neg1_regs));
			break;
		case 0:   
			ret = sensor_write_array(sd, sensor_ev_zero_regs, ARRAY_SIZE(sensor_ev_zero_regs));
			break;
		case 1:
			ret = sensor_write_array(sd, sensor_ev_pos1_regs, ARRAY_SIZE(sensor_ev_pos1_regs));
			break;
		case 2:
			ret = sensor_write_array(sd, sensor_ev_pos2_regs, ARRAY_SIZE(sensor_ev_pos2_regs));
			break;	
		case 3:
			ret = sensor_write_array(sd, sensor_ev_pos3_regs, ARRAY_SIZE(sensor_ev_pos3_regs));
			break;
		case 4:
			ret = sensor_write_array(sd, sensor_ev_pos4_regs, ARRAY_SIZE(sensor_ev_pos4_regs));
			break;
		default:
			return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_exp!\n");
		return ret;
	}
	mdelay(10);
	info->exp = value;
	return 0;
}

static int sensor_g_wb(struct v4l2_subdev *sd, int *value)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_whiteblance *wb_type = (enum v4l2_whiteblance*)value;
	
	*wb_type = info->wb;
	
	return 0;
}

static int sensor_s_wb(struct v4l2_subdev *sd,
		enum v4l2_whiteblance value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	if (value == V4L2_WB_AUTO) {
		ret = sensor_s_autowb(sd, 1);
		return ret;
	} 
	else {
		ret = sensor_s_autowb(sd, 0);
		if(ret < 0) {
			csi_dev_err("sensor_s_autowb error, return %x!\n",ret);
			return ret;
		}
		
		switch (value) {
			case V4L2_WB_CLOUD:
			  ret = sensor_write_array(sd, sensor_wb_cloud_regs, ARRAY_SIZE(sensor_wb_cloud_regs));
				break;
			case V4L2_WB_DAYLIGHT:
				ret = sensor_write_array(sd, sensor_wb_daylight_regs, ARRAY_SIZE(sensor_wb_daylight_regs));
				break;
			case V4L2_WB_INCANDESCENCE:
				ret = sensor_write_array(sd, sensor_wb_incandescence_regs, ARRAY_SIZE(sensor_wb_incandescence_regs));
				break;    
			case V4L2_WB_FLUORESCENT:
				ret = sensor_write_array(sd, sensor_wb_fluorescent_regs, ARRAY_SIZE(sensor_wb_fluorescent_regs));
				break;
			case V4L2_WB_TUNGSTEN:   
				ret = sensor_write_array(sd, sensor_wb_tungsten_regs, ARRAY_SIZE(sensor_wb_tungsten_regs));
				break;
			default:
				return -EINVAL;
		} 
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_s_wb error, return %x!\n",ret);
		return ret;
	}
	
	mdelay(10);
	info->wb = value;
	return 0;
}

static int sensor_g_colorfx(struct v4l2_subdev *sd,
		__s32 *value)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_colorfx *clrfx_type = (enum v4l2_colorfx*)value;
	
	*clrfx_type = info->clrfx;
	return 0;
}

static int sensor_s_colorfx(struct v4l2_subdev *sd,
		enum v4l2_colorfx value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	switch (value) {
	case V4L2_COLORFX_NONE:
	  ret = sensor_write_array(sd, sensor_colorfx_none_regs, ARRAY_SIZE(sensor_colorfx_none_regs));
		break;
	case V4L2_COLORFX_BW:
		ret = sensor_write_array(sd, sensor_colorfx_bw_regs, ARRAY_SIZE(sensor_colorfx_bw_regs));
		break;  
	case V4L2_COLORFX_SEPIA:
		ret = sensor_write_array(sd, sensor_colorfx_sepia_regs, ARRAY_SIZE(sensor_colorfx_sepia_regs));
		break;   
	case V4L2_COLORFX_NEGATIVE:
		ret = sensor_write_array(sd, sensor_colorfx_negative_regs, ARRAY_SIZE(sensor_colorfx_negative_regs));
		break;
	case V4L2_COLORFX_EMBOSS:   
		ret = sensor_write_array(sd, sensor_colorfx_emboss_regs, ARRAY_SIZE(sensor_colorfx_emboss_regs));
		break;
	case V4L2_COLORFX_SKETCH:     
		ret = sensor_write_array(sd, sensor_colorfx_sketch_regs, ARRAY_SIZE(sensor_colorfx_sketch_regs));
		break;
	case V4L2_COLORFX_SKY_BLUE:
		ret = sensor_write_array(sd, sensor_colorfx_sky_blue_regs, ARRAY_SIZE(sensor_colorfx_sky_blue_regs));
		break;
	case V4L2_COLORFX_GRASS_GREEN:
		ret = sensor_write_array(sd, sensor_colorfx_grass_green_regs, ARRAY_SIZE(sensor_colorfx_grass_green_regs));
		break;
	case V4L2_COLORFX_SKIN_WHITEN:
		ret = sensor_write_array(sd, sensor_colorfx_skin_whiten_regs, ARRAY_SIZE(sensor_colorfx_skin_whiten_regs));
		break;
	case V4L2_COLORFX_VIVID:
		ret = sensor_write_array(sd, sensor_colorfx_vivid_regs, ARRAY_SIZE(sensor_colorfx_vivid_regs));
		break;
	default:
		return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_s_colorfx error, return %x!\n",ret);
		return ret;
	}
	mdelay(10);
	info->clrfx = value;
	
	return 0;
}

static int sensor_g_flash_mode(struct v4l2_subdev *sd,
    __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_flash_mode *flash_mode = (enum v4l2_flash_mode*)value;
	
	*flash_mode = info->flash_mode;
	return 0;
}

static int sensor_s_flash_mode(struct v4l2_subdev *sd,
    enum v4l2_flash_mode value)
{
	struct sensor_info *info = to_state(sd);
	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
	int flash_on,flash_off;
	
	flash_on = (dev->flash_pol!=0)?1:0;
	flash_off = (flash_on==1)?0:1;
	
	switch (value) {
	case V4L2_FLASH_MODE_OFF:
		csi_gpio_write(sd,&dev->flash_io,flash_off);
		break;
	case V4L2_FLASH_MODE_AUTO:
		return -EINVAL;
		break;  
	case V4L2_FLASH_MODE_ON:
		csi_gpio_write(sd,&dev->flash_io,flash_on);
		break;   
	case V4L2_FLASH_MODE_TORCH:
		return -EINVAL;
		break;
	case V4L2_FLASH_MODE_RED_EYE:   
		return -EINVAL;
		break;
	default:
		return -EINVAL;
	}
	
	info->flash_mode = value;
	return 0;
}

static int sensor_g_autofocus_mode(struct v4l2_subdev *sd,
    __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->af_mode;
	
	return 0;
}

static int sensor_s_autofocus_mode(struct v4l2_subdev *sd,
    enum v4l2_autofocus_mode value)
{
	struct sensor_info *info = to_state(sd);
	int ret;
	
	csi_dev_dbg("sensor_s_autofocus_mode = %d\n",value);
	
	switch(value) {
		case V4L2_AF_FIXED:
			break;
		case V4L2_AF_INFINITY:
			ret = sensor_s_infinity_af(sd);
			if (ret < 0)
			{
				csi_dev_err("sensor_af_infinity_regs err when sensor_s_autofocus_mode !\n");
				return ret;
			}
			break;
		case V4L2_AF_MACRO:
		  ret = sensor_s_macro_af(sd);
			if (ret < 0)
			{
				csi_dev_err("sensor_af_macro_regs err when sensor_s_autofocus_mode !\n");
				return ret;
			}
		  break;
		case V4L2_AF_AUTO:
		case V4L2_AF_TOUCH:
		case V4L2_AF_FACE:
//			ret = sensor_s_continueous_af(sd);
//			if (ret < 0)
//			{
//				csi_dev_err("sensor_s_continueous_af err when sensor_s_autofocus_mode!\n");
//				return ret;
//			}
			break;
	}
	
	info->af_mode = value;
	
	return 0;
}

static int sensor_g_autofocus_ctrl(struct v4l2_subdev *sd,
		struct v4l2_control *ctrl)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_autofocus_ctrl af_ctrl = ctrl->value;

	switch(af_ctrl) {
		case V4L2_AF_INIT:
			return sensor_g_single_af(sd);
			break;
		case V4L2_AF_RELEASE:
			break;
		case V4L2_AF_TRIG_SINGLE:
			return sensor_g_single_af(sd);
		case V4L2_AF_TRIG_CONTINUEOUS:
			break;
		case V4L2_AF_LOCK:
			break;
		case V4L2_AF_WIN_XY:
			break;
		case V4L2_AF_WIN_NUM:
			break;
	}
	
	ctrl->value = info->af_ctrl;
	return 0;	
}

static int sensor_s_autofocus_ctrl(struct v4l2_subdev *sd,
		struct v4l2_control *ctrl)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_autofocus_ctrl af_ctrl = ctrl->value;
	struct v4l2_pix_size *pix;
	
	csi_dev_dbg("sensor_s_autofocus_ctrl = %d\n",ctrl->value);
	
	switch(af_ctrl) {
		case V4L2_AF_INIT:
			return sensor_download_af_fw(sd);
		case V4L2_AF_RELEASE:
			return sensor_s_release_af(sd);
		case V4L2_AF_TRIG_SINGLE:
			return sensor_s_single_af(sd);
		case V4L2_AF_TRIG_CONTINUEOUS:
			return sensor_s_continueous_af(sd);
		case V4L2_AF_LOCK:
			return sensor_s_pause_af(sd);
		case V4L2_AF_WIN_XY:
			pix = (struct v4l2_pix_size*)ctrl->user_pt;
			return sensor_s_af_zone(sd,pix->width,pix->height);
			break;
		case V4L2_AF_WIN_NUM:
			break;
	}
	
	info->af_ctrl = ctrl->value;
	return 0;
}

static int sensor_g_ctrl(struct v4l2_subdev *sd, struct v4l2_control *ctrl)
{
	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		return sensor_g_brightness(sd, &ctrl->value);
	case V4L2_CID_CONTRAST:
		return sensor_g_contrast(sd, &ctrl->value);
	case V4L2_CID_SATURATION:
		return sensor_g_saturation(sd, &ctrl->value);
	case V4L2_CID_HUE:
		return sensor_g_hue(sd, &ctrl->value);	
	case V4L2_CID_VFLIP:
		return sensor_g_vflip(sd, &ctrl->value);
	case V4L2_CID_HFLIP:
		return sensor_g_hflip(sd, &ctrl->value);
	case V4L2_CID_GAIN:
		return sensor_g_gain(sd, &ctrl->value);
	case V4L2_CID_AUTOGAIN:
		return sensor_g_autogain(sd, &ctrl->value);
	case V4L2_CID_EXPOSURE:
		return sensor_g_exp(sd, &ctrl->value);
	case V4L2_CID_EXPOSURE_AUTO:
		return sensor_g_autoexp(sd, &ctrl->value);
	case V4L2_CID_DO_WHITE_BALANCE:
		return sensor_g_wb(sd, &ctrl->value);
	case V4L2_CID_AUTO_WHITE_BALANCE:
		return sensor_g_autowb(sd, &ctrl->value);
	case V4L2_CID_COLORFX:
		return sensor_g_colorfx(sd,	&ctrl->value);
	case V4L2_CID_CAMERA_FLASH_MODE:
		return sensor_g_flash_mode(sd, &ctrl->value);
	case V4L2_CID_POWER_LINE_FREQUENCY:
		return sensor_g_band_filter(sd, &ctrl->value);
	case V4L2_CID_CAMERA_AF_MODE:
		return sensor_g_autofocus_mode(sd, &ctrl->value);
	case V4L2_CID_CAMERA_AF_CTRL:
		return sensor_g_autofocus_ctrl(sd, ctrl);
	}
	return -EINVAL;
}

static int sensor_s_ctrl(struct v4l2_subdev *sd, struct v4l2_control *ctrl)
{
	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		return sensor_s_brightness(sd, ctrl->value);
	case V4L2_CID_CONTRAST:
		return sensor_s_contrast(sd, ctrl->value);
	case V4L2_CID_SATURATION:
		return sensor_s_saturation(sd, ctrl->value);
	case V4L2_CID_HUE:
		return sensor_s_hue(sd, ctrl->value);		
	case V4L2_CID_VFLIP:
		return sensor_s_vflip(sd, ctrl->value);
	case V4L2_CID_HFLIP:
		return sensor_s_hflip(sd, ctrl->value);
	case V4L2_CID_GAIN:
		return sensor_s_gain(sd, ctrl->value);
	case V4L2_CID_AUTOGAIN:
		return sensor_s_autogain(sd, ctrl->value);
	case V4L2_CID_EXPOSURE:
		return sensor_s_exp(sd, ctrl->value);
	case V4L2_CID_EXPOSURE_AUTO:
		return sensor_s_autoexp(sd,
				(enum v4l2_exposure_auto_type) ctrl->value);
	case V4L2_CID_DO_WHITE_BALANCE:
		return sensor_s_wb(sd,
				(enum v4l2_whiteblance) ctrl->value);	
	case V4L2_CID_AUTO_WHITE_BALANCE:
		return sensor_s_autowb(sd, ctrl->value);
	case V4L2_CID_COLORFX:
		return sensor_s_colorfx(sd,
				(enum v4l2_colorfx) ctrl->value);
	case V4L2_CID_CAMERA_FLASH_MODE:
	  return sensor_s_flash_mode(sd,
	      (enum v4l2_flash_mode) ctrl->value);
	case V4L2_CID_POWER_LINE_FREQUENCY:
		return sensor_s_band_filter(sd,
	      (enum v4l2_power_line_frequency) ctrl->value);
	case V4L2_CID_CAMERA_AF_MODE:
		return sensor_s_autofocus_mode(sd,
	      (enum v4l2_autofocus_mode) ctrl->value);
	case V4L2_CID_CAMERA_AF_CTRL:
		return sensor_s_autofocus_ctrl(sd, ctrl);
	}
	return -EINVAL;
}


static int sensor_g_chip_ident(struct v4l2_subdev *sd,
		struct v4l2_dbg_chip_ident *chip)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	return v4l2_chip_ident_i2c_client(client, chip, V4L2_IDENT_SENSOR, 0);
}


/* ----------------------------------------------------------------------- */

static const struct v4l2_subdev_core_ops sensor_core_ops = {
	.g_chip_ident = sensor_g_chip_ident,
	.g_ctrl = sensor_g_ctrl,
	.s_ctrl = sensor_s_ctrl,
	.queryctrl = sensor_queryctrl,
	.reset = sensor_reset,
	.init = sensor_init,
	.s_power = sensor_power,
	.ioctl = sensor_ioctl,
};

static const struct v4l2_subdev_video_ops sensor_video_ops = {
	.enum_mbus_fmt = sensor_enum_fmt,//linux-3.0
	.try_mbus_fmt = sensor_try_fmt,//linux-3.0
	.s_mbus_fmt = sensor_s_fmt,//linux-3.0
	.s_parm = sensor_s_parm,//linux-3.0
	.g_parm = sensor_g_parm,//linux-3.0
};

static const struct v4l2_subdev_ops sensor_ops = {
	.core = &sensor_core_ops,
	.video = &sensor_video_ops,
};

/* ----------------------------------------------------------------------- */

static int sensor_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	struct v4l2_subdev *sd;
	struct sensor_info *info;
//	int ret;

	info = kzalloc(sizeof(struct sensor_info), GFP_KERNEL);
	if (info == NULL)
		return -ENOMEM;
	sd = &info->sd;
	v4l2_i2c_subdev_init(sd, client, &sensor_ops);

	info->fmt = &sensor_formats[0];
	info->ccm_info = &ccm_info_con;
	
	return 0;
}


static int sensor_remove(struct i2c_client *client)
{
	struct v4l2_subdev *sd = i2c_get_clientdata(client);

	v4l2_device_unregister_subdev(sd);
	kfree(to_state(sd));
	return 0;
}

static const struct i2c_device_id sensor_id[] = {
	{ "s5k4ec", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, sensor_id);

//linux-3.0
static struct i2c_driver sensor_driver = {
	.driver = {
		.owner = THIS_MODULE,
	.name = "s5k4ec",
	},
	.probe = sensor_probe,
	.remove = sensor_remove,
	.id_table = sensor_id,
};
static __init int init_sensor(void)
{
	return i2c_add_driver(&sensor_driver);
}

static __exit void exit_sensor(void)
{
  i2c_del_driver(&sensor_driver);
}

module_init(init_sensor);
module_exit(exit_sensor);


